Package: arcgisutils
Title: R-ArcGIS Bridge Utility Functions
Version: 0.6.0
Authors@R: c(
    person("Josiah", "Parry", , "josiah.parry@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-9910-865X")),
    person("Kenneth", "Vernon", , "kenneth.b.vernon@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0003-0098-5092")),
    person("Martha", "Bass", role = "ctb",
           comment = c(ORCID = "0009-0004-0268-5426")),
    person("Eli", "Pousson", , "eli.pousson@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0001-8280-1706")),
    person("Ryan", "Zomorrodi", , "ryanzomorrodi@gmail.com", role = "ctb",
           comment = c(ORCID = "0009-0003-6417-5985"))
  )
Description: Developer oriented utility functions designed to be used as
    the building blocks of R packages that work with ArcGIS Location
    Services. It provides functionality for authorization, Esri JSON
    construction and parsing, as well as other utilities pertaining to
    geometry and Esri type conversions. To support 'ArcGIS Pro' users,
    authorization can be done via 'arcgisbinding'. Installation
    instructions for 'arcgisbinding' can be found at
    <https://developers.arcgis.com/r-bridge/installation/>.
License: Apache License (>= 2)
URL: https://github.com/R-ArcGIS/arcgisutils,
        https://developers.arcgis.com/r-bridge/api-reference/arcgisutils/
BugReports: https://github.com/R-ArcGIS/arcgisutils/issues
Depends: R (>= 4.2)
Imports: cli, httr2 (>= 1.0.5), R6, RcppSimdJson, rlang, S7, sf, utils,
        yyjsonr, lifecycle
Suggests: arcgisbinding, collapse (>= 2.0.0), data.table, jsonify,
        testthat (>= 3.0.0), vctrs, curl, shinyOAuth
Config/rextendr/version: 0.4.2
Config/testthat/edition: 3
Encoding: UTF-8
Language: en
RoxygenNote: 7.3.3
SystemRequirements: Cargo (Rust's package manager), rustc >= 1.67, xz
NeedsCompilation: yes
Packaged: 2026-07-10 18:46:24 UTC; josiahparry
Author: Josiah Parry [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9910-865X>),
  Kenneth Vernon [ctb] (ORCID: <https://orcid.org/0000-0003-0098-5092>),
  Martha Bass [ctb] (ORCID: <https://orcid.org/0009-0004-0268-5426>),
  Eli Pousson [ctb] (ORCID: <https://orcid.org/0000-0001-8280-1706>),
  Ryan Zomorrodi [ctb] (ORCID: <https://orcid.org/0009-0003-6417-5985>)
Maintainer: Josiah Parry <josiah.parry@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-10 23:20:08 UTC

# Generated by roxygen2: do not edit by hand

S3method(determine_dims,sfc)
S3method(determine_dims,sfg)
S3method(has_m,sfc)
S3method(has_m,sfg)
S3method(has_z,sfc)
S3method(has_z,sfg)
S3method(print,PortalGroup)
S3method(print,PortalItem)
S3method(print,PortalUser)
S3method(print,arc_gp_job)
export("%||%")
export(arc_agent)
export(arc_base_req)
export(arc_form_params)
export(arc_gp_job)
export(arc_group)
export(arc_group_content)
export(arc_group_users)
export(arc_host)
export(arc_item)
export(arc_item_data)
export(arc_job_status)
export(arc_paginate_req)
export(arc_portal_resources)
export(arc_portal_self)
export(arc_portal_servers)
export(arc_portal_urls)
export(arc_portal_users)
export(arc_self_meta)
export(arc_token)
export(arc_url_parse)
export(arc_url_type)
export(arc_user)
export(arc_user_content)
export(arc_user_self)
export(as_esri_features)
export(as_esri_featureset)
export(as_esri_geometry)
export(as_extent)
export(as_feature_collection)
export(as_features)
export(as_featureset)
export(as_fields)
export(as_form_params)
export(as_gp_areal_unit)
export(as_gp_date)
export(as_gp_feature_record_set)
export(as_gp_linear_unit)
export(as_gp_raster_layer)
export(as_layer)
export(as_layer_definition)
export(as_record_set)
export(as_spatial_reference)
export(auth_binding)
export(auth_client)
export(auth_code)
export(auth_key)
export(auth_shiny)
export(auth_user)
export(catch_error)
export(check_dots_named)
export(check_token_has_user)
export(compact)
export(data_frame)
export(date_to_ms)
export(detect_errors)
export(determine_dims)
export(determine_esri_geo_type)
export(fetch_layer_metadata)
export(fields_as_ptype_df)
export(from_envelope)
export(from_esri_date)
export(from_spatial_reference)
export(gp_areal_unit)
export(gp_job_from_url)
export(gp_linear_unit)
export(has_m)
export(has_z)
export(infer_esri_type)
export(is_date)
export(is_url)
export(item_keyword)
export(item_type)
export(new_gp_job)
export(oauth_provider_arcgis)
export(obj_check_token)
export(parse_esri_json)
export(parse_gp_areal_unit)
export(parse_gp_date)
export(parse_gp_feature_record_set)
export(parse_gp_linear_unit)
export(parse_gp_record_set)
export(parse_spatial_reference)
export(portal_item_keywords)
export(portal_item_types)
export(ptype_tbl)
export(rbind_results)
export(refresh_token)
export(search_items)
export(set_arc_token)
export(unset_arc_token)
export(validate_crs)
export(validate_or_refresh_token)
if (getRversion() < "4.3.0") importFrom("S7", "@")
importFrom(R6,R6Class)
importFrom(lifecycle,deprecated)
useDynLib(arcgisutils, .registration = TRUE)

# arcgisutils <img src="man/figures/logo.svg" align="right" height="139" alt="" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/R-ArcGIS/arcgisutils/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/R-ArcGIS/arcgisutils/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/arcgisutils)](https://CRAN.R-project.org/package=arcgisutils)
[![extendr](https://img.shields.io/badge/extendr-%5E0.8.0-276DC2)](https://extendr.rs/extendr/extendr_api/)
<!-- badges: end -->

arcgisutils is the foundational infrastructure package that powers the
R-ArcGIS Bridge Data and Location Service ecosystem. It provides
sophisticated, production-ready tools for interacting with ArcGIS
Online, ArcGIS Enterprise, and ArcGIS Platform via their REST APIs.

<div class="callout-tip">

An [`llms.txt`](https://r.esri.com/arcgisutils/llms.txt) file is
available to provide context for LLMs when working with this package.

</div>

## Key Capabilities

**🔐 Comprehensive Authentication**:

- Multiple OAuth2 workflows (`auth_code()`, `auth_client()`)
- API key and legacy token support
- Automatic token refresh and validation
- Integration with ArcGIS Pro via `arcgisbinding`

**🌐 Portal Integration**:

- Advanced content search with filtering, sorting, and pagination
- User, group, and organization metadata management
- Portal item discovery and content management workflows

**⚙️ Geoprocessing Services**:

- Support for the geoprocessing service framework built upon R6 and S7
- Enables users to call their own custom geoprocessing services or build
  on top of existing services
- R6-based job management (`arc_gp_job`) with real-time status tracking
  and built-in result parsing

**📄 Esri JSON Ecosystem**:

- Bidirectional conversion between R spatial data and Esri JSON formats
- Support for FeatureSets, geometry objects, field definitions, and
  spatial reference systems
- Optimized parsing with automatic `sf` integration

**🛠️ Developer Utilities**:

- Standardized HTTP client (`arc_base_req()`, `arc_paginate_req()`)
- Robust error detection and user-friendly error messages
- URL parsing, service introspection, and metadata extraction

## Installation

`{arcgisutils}` is part of the `{arcgis}` metapackage, which provides
the complete R-ArcGIS Bridge toolkit. For most users, installing the
metapackage is recommended:

``` r
install.packages("arcgis")
```

You can also install `{arcgisutils}` individually from CRAN:

``` r
install.packages("arcgisutils")
```

To install the development version:

``` r
pak::pak("r-arcgis/arcgisutils")
```

### Authentication

Authorization tokens are provided through the functions `auth_code()`,
`auth_client()`, `auth_user()`, `auth_key()`, and `auth_binding()`.
Additional token validation functions are provided via `refresh_token()`
and `validate_or_refresh_token()`.

`auth_code()` can be used for integrating into Shiny applications, for
example, to have individual users log in. We recommend using
`auth_key()` for authenticating in non-interactive environments (for
example scheduled scripts or deployments).

Tokens are managed using `set_arc_token()` and `unset_arc_token()`. They
are fetched using `arc_token()`. `set_arc_token()` can set the token
globally or set multiple named environments. Here is a minimal example:

``` r
library(arcgisutils)
#> 
#> Attaching package: 'arcgisutils'
#> The following object is masked from 'package:base':
#> 
#>     %||%
```

``` r
key <- auth_key()
set_arc_token(key)
```

Alternatively, tokens can be set based on a key-value pair for multiple
environments:

``` r
set_arc_token("production" = prod_token, "development" = dev_token)
```

And fetched based on their name via

``` r
arc_token("production")
```

### Portal Integration

Search and discover content across your ArcGIS organization:

``` r
# Search for feature services containing "crime" data
crime_items <- search_items(
  query = "crime", 
  item_type = "Feature Service",
  max_pages = 1
)

crime_items
#> # A data frame: 50 × 46
#>    id      owner created             modified            guid  name  title type 
#>  * <chr>   <chr> <dttm>              <dttm>              <lgl> <chr> <chr> <chr>
#>  1 ea0cfe… Toro… 2023-03-28 15:02:39 2026-02-02 17:45:21 NA    Neig… Neig… Feat…
#>  2 64691a… Temp… 2024-01-17 20:01:43 2024-01-17 20:04:45 NA    hate… Hate… Feat…
#>  3 5e055d… JASo… 2023-04-04 17:36:59 2023-09-07 19:05:06 NA    <NA>  Sher… Feat…
#>  4 30644d… MyCi… 2025-03-14 14:55:06 2025-08-20 13:55:24 NA    HPD_… HPD … Feat…
#>  5 c749e3… open… 2024-02-23 19:36:34 2026-01-09 18:36:21 NA    <NA>  Crim… Feat…
#>  6 e0992d… balt… 2023-07-31 20:27:01 2025-01-22 21:21:01 NA    Part… Part… Feat…
#>  7 2cb53d… KASU… 2019-12-10 19:06:39 2019-12-10 19:14:27 NA    Viol… Viol… Feat…
#>  8 5dc4e6… iwat… 2023-06-23 22:07:21 2023-08-09 15:33:46 NA    <NA>  Prop… Feat…
#>  9 ab92f5… KASU… 2019-12-09 16:16:13 2019-12-10 18:55:20 NA    Prop… Prop… Feat…
#> 10 94bc33… admi… 2023-08-11 20:09:35 2026-01-09 17:03:25 NA    <NA>  Crim… Feat…
#> # ℹ 40 more rows
#> # ℹ 38 more variables: typeKeywords <list>, description <chr>, tags <list>,
#> #   snippet <chr>, thumbnail <chr>, documentation <lgl>, extent <list>,
#> #   categories <list>, spatialReference <chr>, accessInformation <chr>,
#> #   classification <lgl>, licenseInfo <chr>, culture <chr>, properties <list>,
#> #   advancedSettings <lgl>, url <chr>, proxyFilter <lgl>, access <chr>,
#> #   size <int>, subInfo <int>, appCategories <list>, industries <list>, …
```

``` r
# Get detailed item information for a portal item
arc_item(crime_items$id[1])
#> <PortalItem<Feature Service>>
#> id: ea0cfecdb1de416884e6b0bf08a9e195
#> title: Neighbourhood Crime Rates Open Data
#> owner: TorontoPoliceService
```

### Developer Utilities

Always use `arc_base_req()` as this will handle setting the user agent
and authorization token. The function creates a standardized `httr2`
request object:

``` r
# defaults to arcgis.com
host <- arc_host() 

req <- arc_base_req(host)
req
#> <httr2_request>
#> GET https://www.arcgis.com
#> Body: empty
#> Options:
#> * useragent: "arcgisutils v0.4.0.9001"
```

To handle paginated services and requests use `arc_paginate_req()` to
automatically handle fetching pages.

### Esri JSON

There are also a number of utility functions for creating and parsing
Esri JSON. For example we can create an Esri `FeatureSet` json string
using `as_esri_featureset()` directly from an `sf` object.

``` r
library(sf)

# load the NC SIDS dataset and extract centroids
# of the first few rows
nc <- system.file("shape/nc.shp", package = "sf") |> 
  st_read(quiet = TRUE) |> 
  st_centroid() 

# convert to json
nc_json <- as_esri_featureset(nc[1:2, 1:3])

jsonify::pretty_json(nc_json)
#> {
#>     "geometryType": "esriGeometryPoint",
#>     "spatialReference": {
#>         "wkid": 4267
#>     },
#>     "features": [
#>         {
#>             "geometry": {
#>                 "x": -81.4982290095261,
#>                 "y": 36.43139560823758
#>             },
#>             "attributes": {
#>                 "AREA": 0.114,
#>                 "CNTY_": 1825.0,
#>                 "PERIMETER": 1.442
#>             }
#>         },
#>         {
#>             "geometry": {
#>                 "x": -81.12512977849917,
#>                 "y": 36.49110847237506
#>             },
#>             "attributes": {
#>                 "AREA": 0.061,
#>                 "CNTY_": 1827.0,
#>                 "PERIMETER": 1.231
#>             }
#>         }
#>     ]
#> }
```

Feature set json can also be parsed using `parse_esri_json()`.

``` r
parse_esri_json(nc_json)
#> Simple feature collection with 2 features and 3 fields
#> Geometry type: POINT
#> Dimension:     XY
#> Bounding box:  xmin: -81.49823 ymin: 36.4314 xmax: -81.12513 ymax: 36.49111
#> Geodetic CRS:  NAD27
#>    AREA CNTY_ PERIMETER                   geometry
#> 1 0.114  1825     1.442  POINT (-81.49823 36.4314)
#> 2 0.061  1827     1.231 POINT (-81.12513 36.49111)
```

Additionally, sf’s `crs` object can be converted to a
[`spatialReference`](https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm#GUID-DFF0E738-5A42-40BC-A811-ACCB5814BABC)
JSON object using `validate_crs()`. Convert these to json with `yyjsonr`
or `jsonify`.

``` r
crs <- validate_crs(27700)
jsonify::pretty_json(crs, unbox = TRUE)
#> {
#>     "spatialReference": {
#>         "wkid": 27700
#>     }
#> }
```

### Geoprocessing Services

The geoprocessing service framework is completely supported in
`{arcgisutils}`. Here we combine the functionality of the geoprocessing
job framework with utilities such as `as_esri_featureset()` to call the
[Trace DownStream Elevation
Service](https://developers.arcgis.com/rest/elevation-analysis/trace-downstream/)

``` r
trace_downstream <- function(
  input_points,
  point_id_field = NULL,
  resolution = NULL,
  generalize = FALSE,
  token = arc_token()
) {

  # create a list of parameters
  params <- compact(list(
    InputPoints = as_esri_featureset(input_points),
    PointIdField = point_id_field,
    DataSourceResolution = resolution,
    Generalize = as.character(generalize),
    f = "json"
  ))

  service_url <- "https://hydro.arcgis.com/arcgis/rest/services/Tools/Hydrology/GPServer/TraceDownstream"
  arc_gp_job$new(
    base_url = service_url,
    params = params,
    result_fn = parse_gp_feature_record_set,
    token
  )
}
```

This new function can be called to start a new job:

``` r
# create input points
input_points <- st_sfc(
  st_point(c(-159.548936, 21.955888)),
  crs = 4326
)

# initialze an empty job
job <- trace_downstream(
  input_points,
  token = auth_user()
)

# start the job
job$start()
#> <arc_gp_job>
#> Job ID: j89b6513d04f44037b39748e8d931cb31
#> Status: esriJobExecuting
#> Resource: /TraceDownstream
#> Params:
#> • InputPoints
#> • Generalize
#> • f
```

Jobs run asynchronously so we can check the status with `job$status`

``` r
job$status
#> <arcgisutils::arc_job_status>
#>  @ status: chr "esriJobExecuting"
```

Then, when the job is complete, we can fetch the results applying the
result function which is `parse_gp_feature_record_set()` in this case.

``` r
job$results
#> $param_name
#> [1] "OutputTraceLine"
#> 
#> $data_type
#> [1] "GPFeatureRecordSetLayer"
#> 
#> $geometry
#> Simple feature collection with 1 feature and 6 fields
#> Geometry type: MULTILINESTRING
#> Dimension:     XY
#> Bounding box:  xmin: 438895 ymin: 2422310 xmax: 443325 ymax: 2428045
#> Projected CRS: NAD83 / UTM zone 4N + Unknown VCS
#>   OBJECTID PourPtID               Description DataResolution LengthKm
#> 1        1        1 NED 10m processed by Esri           10.0 9.489823
#>   Shape_Length                       geometry
#> 1     9489.823 MULTILINESTRING ((443325 24...

# store and view the results
res <- job$results
res
#> $param_name
#> [1] "OutputTraceLine"
#> 
#> $data_type
#> [1] "GPFeatureRecordSetLayer"
#> 
#> $geometry
#> Simple feature collection with 1 feature and 6 fields
#> Geometry type: MULTILINESTRING
#> Dimension:     XY
#> Bounding box:  xmin: 438895 ymin: 2422310 xmax: 443325 ymax: 2428045
#> Projected CRS: NAD83 / UTM zone 4N + Unknown VCS
#>   OBJECTID PourPtID               Description DataResolution LengthKm
#> 1        1        1 NED 10m processed by Esri           10.0 9.489823
#>   Shape_Length                       geometry
#> 1     9489.823 MULTILINESTRING ((443325 24...

# plot the resultant geometry
plot(st_geometry(res$geometry))
```

<img src="man/figures/README-unnamed-chunk-17-1.png" width="100%" />

## Learn More

To learn more about the R-ArcGIS Bridge project visit the [developer
documentation](https://developers.arcgis.com/r-bridge/).

library(shiny)
library(calcite) # requires calcite 0.1.2
library(shinyOAuth)
library(arcgisutils)

# create the shiny AGOL login provider
client <- auth_shiny()

ui <- calcite_shell(
  # Include the shinyOAuth module
  use_shinyOAuth(),
  uiOutput("navigation")
)

server <- function(input, output, session) {
  auth <- shinyOAuth::oauth_module_server(
    "auth",
    client,
    auto_redirect = FALSE
  )

  # Render entire navigation based on authentication status
  output$navigation <- renderUI({
    calcite_navigation(
      slot = "header",
      calcite_menu(
        slot = "content-end",
        calcite_menu_item(
          text = "Drivers",
          `icon-start` = "license"
        ),
        calcite_menu_item(
          text = "Routes",
          `icon-start` = "road-sign"
        ),
        calcite_menu_item(
          text = "Forecast",
          `icon-start` = "snow"
        )
      ),
      # Conditionally render user slot
      if (auth$authenticated) {
        user_info <- auth$token@userinfo
        calcite_navigation_user(
          slot = "user",
          username = user_info$user$username,
          `full-name` = user_info$user$fullName,
          `user-id` = user_info$user$id
        )
      } else {
        calcite_button(
          id = "login_btn",
          slot = "user",
          "Login"
        )
      }
    )
  })

  # Trigger login when button is clicked
  observeEvent(input$login_btn$clicked, {
    auth$request_login()
  })
}

runApp(
  shinyApp(ui, server),
  port = 8100,
  launch.browser = FALSE
)

library(shiny)
library(shinyOAuth)
library(arcgisutils)

# Simple UI
ui <- fluidPage(
  use_shinyOAuth(),
  uiOutput("login_information")
)

client <- auth_shiny()

server <- function(input, output, session) {
  auth <- shinyOAuth::oauth_module_server(
    "auth",
    client,
    auto_redirect = FALSE
  )

  output$login_information <- renderUI({
    if (auth$authenticated) {
      user_info <- auth$token@userinfo
      tagList(
        tags$p("You are logged in! Your details:"),
        tags$pre(paste(capture.output(str(user_info)), collapse = "\n"))
      )
    } else {
      tagList(
        tags$p("You are not logged in."),
        actionButton("login_btn", "Login with ArcGIS")
      )
    }
  })

  # Trigger login when button is clicked
  observeEvent(input$login_btn, {
    auth$request_login()
  })
}

runApp(
  shinyApp(ui, server),
  port = 8100,
  launch.browser = FALSE
)

# Open the app in your regular browser at http://127.0.01:8100

#' Portal Item Metadata
#'
#' Given the unique ID of a content item, fetches the item's metadata from a portal.
#'
#' @param item_id the ID of the item to fetch. A scalar character.
#' @inheritParams auth_user
#' @details
#'
#' See [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/item/) for more information.
#'
#' `r lifecycle::badge("experimental")`
#' @export
#' @family portal item
#' @examplesIf curl::has_internet()
#' arc_item("9df5e769bfe8412b8de36a2e618c7672")
#' @returns an object of class `PortalItem` which is a list with the item's metadata.
arc_item <- function(item_id, host = arc_host(), token = arc_token()) {
  check_string(item_id, allow_empty = FALSE)
  resp <- arc_base_req(
    host,
    path = paste0("sharing/rest/content/items/", item_id),
    token,
    query = c("f" = "json")
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    RcppSimdJson::fparse() |>
    detect_errors()

  for (field in c("created", "modified", "lastViewed")) {
    resp[[field]] <- from_esri_date(resp[[field]])
  }

  structure(resp, class = c("PortalItem", "list"))
}

#' @export
print.PortalItem <- function(x, ...) {
  cat(sprintf("<PortalItem<%s>>\n", x$type))
  for (field in c("id", "title", "owner")) {
    cat(field, ": ", x[[field]], "\n", sep = "")
  }
  invisible(x)
}

#' Organization's URLs
#'
#' Returns the URLs of an organizations services.
#'
#' See [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/urls/) for more information.
#' `r lifecycle::badge("experimental")`
#' @export
#' @family portal
#' @inheritParams auth_user
#' @examplesIf curl::has_internet()
#' arc_portal_urls()
#' @returns a data.frame
arc_portal_urls <- function(host = arc_host(), token = arc_token()) {
  arc_base_req(
    host,
    token,
    "sharing/rest/portals/self/urls",
    query = c("f" = "json")
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    RcppSimdJson::fparse()
}


#' Download an Item's Data
#'
#' Download the data backing a portal item. This function always returns
#' a raw vector as the type of the data that is downloaded cannot always be known.
#'
#' `r lifecycle::badge("experimental")`
#' @param item the item ID or the result of `arc_item()`.
#' @export
#' @inheritParams auth_user
#' @family portal item
#' @examplesIf curl::has_internet()
#' arc_item_data("9df5e769bfe8412b8de36a2e618c7672")
#' @returns a raw vector containing the bytes of the data associated with the item. If the response is `application/json` then the json string is returned without parsing.
arc_item_data <- function(
  item,
  host = arc_host(),
  token = arc_token()
) {
  e_msg <- "Expected a content ID or {.cls PortalItem<_>} created with {.fn arc_item}"

  if (rlang::is_string(item)) {
    item <- rlang::try_fetch(
      arc_item(item, host, token),
      error = function(cnd) cli::cli_abort(e_msg)
    )
  }

  if (!inherits(item, "PortalItem")) {
    cli::cli_abort(e_msg)
  }

  resp <- arc_base_req(
    host,
    path = c("sharing/rest/content/items/", item[["id"]], "data"),
    token
  ) |>
    httr2::req_perform()

  resp_type <- httr2::resp_content_type(resp)

  if (resp_type == "application/json") {
    resp_str <- httr2::resp_body_string(resp)
    catch_error(resp_str)
    resp_str
  } else {
    httr2::resp_body_raw(resp)
  }
}


#' Fetch Group Information
#'
#' Fetches metadata about a group based on a provided `group_id`.
#'
#' `r lifecycle::badge("experimental")`
#' @param group_id the unique group identifier. A scalar character.
#' @inheritParams arc_item
#' @export
#' @family portal organization
#' @examplesIf curl::has_internet()
#' arc_group("2f0ec8cb03574128bd673cefab106f39")
#' @returns a list with group metadata
arc_group <- function(
  group_id,
  host = arc_host(),
  token = arc_token()
) {
  check_string(group_id)
  group <- arc_base_req(
    host,
    token,
    c("sharing/rest/community/groups", group_id),
    query = c("f" = "json")
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    RcppSimdJson::fparse() |>
    detect_errors()

  for (field in c("created", "modified")) {
    group[[field]] <- from_esri_date(group[[field]])
  }

  structure(group, class = c("PortalGroup", "list"))
}

#' @export
print.PortalGroup <- function(x, ...) {
  fields <- c("id", "owner", "created")
  cat(sprintf("<PortalGroup<%s>>\n", x[["title"]]))
  for (field in fields) {
    v <- x[[field]]

    if (is.null(v)) {
      next
    }

    if (!nzchar(v)) {
      next
    }

    cat(field, ": ", format(x[[field]]), "\n", sep = "")
  }
  invisible(x)
}

#' User Information
#'
#' Fetch a user's metadata based on username.
#'
#' `r lifecycle::badge("experimental")`
#'
#' @param username the username to fetch. A scalar character.
#' @inheritParams arc_item
#' @export
#' @family portal organization
#' @returns a list of class `PortalUser`
#' @examplesIf curl::has_internet()
#' arc_user("esri_en")
arc_user <- function(username, host = arc_host(), token = arc_token()) {
  user <- arc_base_req(
    host,
    token,
    c("sharing/rest/community/users", username),
    query = c("f" = "json")
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    RcppSimdJson::fparse() |>
    detect_errors()

  for (field in c("created", "modified")) {
    user[[field]] <- from_esri_date(user[[field]])
  }

  structure(user, class = c("PortalUser", "list"))
}

#' @export
print.PortalUser <- function(x, ...) {
  cat(sprintf("<PortalUser<%s>>\n", x$username))
  for (field in c("id", "fullName", "created")) {
    v <- x[[field]]

    if (is.null(v)) {
      next
    }

    if (!nzchar(v)) {
      next
    }

    cat(field, ": ", format(x[[field]]), "\n", sep = "")
  }
  invisible(x)
}

test_that("arc_item() works", {
  skip_if_offline()
  unset_arc_token()
  test_cases <- c(
    webmap_app = "80eb92ffc89b4086abe8cedd58ab160c",
    storymap = "ad791fda858c46fdbe79636aa5f35dd8",
    instant_app = "f2dfd67d29ed4cabbb91e742e0297955",
    experience = "6e360741bfd84db79d5db774a1147815",
    webapp = "950b4eec577a4dc5b298a61adab41c06",
    notebook_item = "9a9fca3f09bb41dd856c9cd4239b8519",
    notebook = "9a9fca3f09bb41dd856c9cd4239b8519",
    webscene = "7b506043536246faa4194d4c3d4c921b",
    item_db = "84ba9c03786e462d960e3172bc1b2204",
    item_mapserver = "1d150c40d9f642cb8bd691017bf22cee",
    feature_collection = "24aa36ce1d7747c2b5a6aa57711d03fb"
  )

  expected_types <- c(
    "Web Mapping Application",
    "StoryMap",
    "Web Mapping Application",
    "Web Experience",
    "Web Mapping Application",
    "Notebook",
    "Notebook",
    "Web Scene",
    "Dashboard",
    "Map Service",
    "Feature Collection"
  )

  for (i in seq_along(test_cases)) {
    expect_equal(
      arc_item(test_cases[i])[["type"]],
      expected_types[i]
    )
  }
})

#' Portal Content Items
#'
#' For a given user or group, returns a data.frame of all content items owned by them.
#'
#' @param group a scalar character of the group ID or a `PortalGroup` object created using [`arc_group()`]
#' @param user a scalar character of the username or a `PortalUser` object created using [`arc_user()`]
#' @inheritParams arc_item
#' @keywords content portal
#' @export
#' @examples
#' \dontrun{
#' library(arcgis)
#'
#' # authenticate
#' set_arc_token(auth_user())
#'
#' # get your own content items
#' self <- arc_user_self()
#' arc_user_content(self$username)
#'
#' # get a specific group's items
#' arc_group_content("2f0ec8cb03574128bd673cefab106f39")
#' }
#' @references
#' - [Group Content API Reference](https://developers.arcgis.com/rest/users-groups-and-items/group-content/)
#' - [User Content API Reference](https://developers.arcgis.com/rest/users-groups-and-items/user-content/)
#' @name content
#' @returns a `data.frame` of content item metadata
arc_group_content <- function(
  group,
  host = arc_host(),
  token = arc_token()
) {
  if (inherits(group, "PortalGroup")) {
    group <- group$id
  }

  if (!rlang::is_string(group)) {
    cli::cli_abort(
      "{.arg group} must be a string or {.cls PortalGroup} created with {.fn arc_group}"
    )
  }

  req <- arc_base_req(
    host,
    path = c("sharing", "rest", "content", "groups", group),
    query = c("f" = "json"),
    token = token
  )

  all_resps <- arc_paginate_req(req)

  results <- lapply(all_resps, function(.resp) {
    RcppSimdJson::fparse(httr2::resp_body_string(.resp))[["items"]]
  })

  res <- data_frame(rbind_results(results))

  res[["created"]] <- from_esri_date(res[["created"]])
  res[["modified"]] <- from_esri_date(res[["modified"]])
  res[["lastViewed"]] <- from_esri_date(res[["lastViewed"]])
  res
}


#' @name content
#' @export
arc_user_content <- function(
  user = arc_user_self(token = token),
  host = arc_host(),
  token = arc_token()
) {
  if (inherits(user, "PortalUser")) {
    user <- user$username
  }

  if (!rlang::is_string(user)) {
    cli::cli_abort(
      "{.arg user} must be a string or {.cls PortalUser} created with {.fn arc_user}"
    )
  }

  req <- arc_base_req(
    host,
    path = c("sharing", "rest", "content", "users", user),
    query = c("f" = "json"),
    token = token
  )

  all_resps <- arc_paginate_req(req)

  results <- lapply(all_resps, function(.resp) {
    RcppSimdJson::fparse(httr2::resp_body_string(.resp))[["items"]]
  })

  res <- data_frame(rbind_results(results))

  res[["created"]] <- from_esri_date(res[["created"]])
  res[["modified"]] <- from_esri_date(res[["modified"]])
  res[["lastViewed"]] <- from_esri_date(res[["lastViewed"]])
  res
}

#' List users in a group
#'
#' List all users in a given group.
#'
#' @param group a group ID as a scalar string or `PortalGroup` created via [`arc_group()`]
#' @param name a scalar string of a user to search for.
#' @param member_type default `NULL`. If provided must be one of `"admin"` or `"member"`.
#' @param joined_before default `NULL`. A scalar date to search for users who joined before this date.
#' @param joined_after default `NULL`. A scalar date to search for users who joined after this date.
#' @param sort_field default `"username"`. The field to sort by. Must be one of `"username"`, `"membertype"`, or `"joined"`.
#' @references [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/group-users-list/)
#' @inheritParams search_items
#' @export
#' @returns a data.frame
#' @examples
#' \dontrun{
#' set_arc_token(auth_user())
#' groups <- arc_user("r-bridge-docs")$groups
#' arc_group_users(groups$id[1])
#' }
arc_group_users <- function(
  group,
  name = NULL,
  member_type = NULL,
  joined_before = NULL,
  joined_after = NULL,
  sort_field = c("username", "membertype", "joined"),
  sort_order = c("asc", "desc"),
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
) {
  # validate user groups
  if (inherits(group, "PortalGroup")) {
    group <- group$id
    if (!rlang::is_string(group)) {
      cli::cli_abort(
        "{.cls PortalGroup} has an invalid group ID. Expected a string"
      )
    }
  } else if (rlang::is_string(group)) {
    group <- group
  } else {
    cli::cli_abort(
      "Expected a {.cls PortalGroup} or a bare string of the group ID. Found {obj_type_friendly(group)}"
    )
  }

  check_string(name, allow_null = TRUE)

  # if not-missing, validate
  if (!is.null(member_type)) {
    member_type <- rlang::arg_match(
      member_type,
      c("admin", "member")
    )
  }

  sort_field <- rlang::arg_match(
    sort_field
  )
  sort_order <- rlang::arg_match(sort_order)

  if (!is.null(joined_after)) {
    joined_after <- date_to_ms(joined_after)
  }

  if (!is.null(joined_before)) {
    joined_before <- date_to_ms(joined_before)
  }

  joined <- if (!is.null(joined_before) || !is.null(joined_after)) {
    sprintf("%s,%s", joined_after %||% "", joined_before %||% "")
  } else {
    NULL
  }

  params <- compact(list(
    sortField = sort_field,
    sortOrder = sort_order,
    joined = joined,
    name = name,
    joined = joined
  ))

  resps <- arc_base_req(
    host,
    path = c("sharing", "rest", "community", "groups", group, "userList"),
    query = c("f" = "json"),
    token = token
  ) |>
    httr2::req_url_query(!!!params) |>
    arc_paginate_req(
      max_pages = max_pages,
      page_size = page_size,
      .progress = .progress
    )

  results <- lapply(resps, function(.resp) {
    resp <- httr2::resp_body_string(.resp) |>
      yyjsonr::read_json_str() |>
      detect_errors()
    res <- resp[["users"]]
  })

  res <- data_frame(rbind_results(results))
  res[["joined"]] <- from_esri_date(as.numeric(res$joined))
  res
}

# Note: Any variables prefixed with `.` are used for text
# replacement in the Makevars.in and Makevars.win.in

# check the packages MSRV first
source("tools/msrv.R")

# check DEBUG and NOT_CRAN environment variables
env_debug <- Sys.getenv("DEBUG")
env_not_cran <- Sys.getenv("NOT_CRAN")
env_target <- Sys.getenv("CARGO_TARGET_DIR")
.target_dir <- if (nzchar(env_target)) env_target else "./rust/target"

# check if the vendored zip file exists
vendor_exists <- file.exists("src/rust/vendor.tar.xz")

is_not_cran <- env_not_cran != ""
is_debug <- env_debug != ""

if (is_debug) {
  # if we have DEBUG then we set not cran to true
  # CRAN is always release build
  is_not_cran <- TRUE
  message("Creating DEBUG build.")
}

if (!is_not_cran) {
  message("Building for CRAN.")
}

# we set cran flags only if NOT_CRAN is empty and if
# the vendored crates are present.
.cran_flags <- ifelse(
  !is_not_cran && vendor_exists,
  "-j 2 --offline",
  ""
)

# when DEBUG env var is present we use `--debug` build
.profile <- ifelse(is_debug, "", "--release")
.clean_targets <- ifelse(is_debug || is_not_cran, "", "$(TARGET_DIR)")

# We specify this target when building for webR
webr_target <- "wasm32-unknown-emscripten"

# here we check if the platform we are building for is webr
is_wasm <- identical(R.version$platform, webr_target)

# print to terminal to inform we are building for webr
if (is_wasm) {
  message("Building for WebR")
}

# we check if we are making a debug build or not
# if so, the LIBDIR environment variable becomes:
# LIBDIR = $(TARGET_DIR)/{wasm32-unknown-emscripten}/debug
# this will be used to fill out the LIBDIR env var for Makevars.in
target_libpath <- if (is_wasm) "wasm32-unknown-emscripten" else NULL
cfg <- if (is_debug) "debug" else "release"

# used to replace @LIBDIR@
.libdir <- paste(c(target_libpath, cfg), collapse = "/")

# use this to replace @TARGET@
# we specify the target _only_ on webR
# there may be use cases later where this can be adapted or expanded
.target <- ifelse(is_wasm, paste0("--target=", webr_target), "")

# add panic exports only for WASM builds
.panic_exports <- ifelse(
  is_wasm,
  "CARGO_PROFILE_DEV_PANIC=\"abort\" CARGO_PROFILE_RELEASE_PANIC=\"abort\" ",
  ""
)

# read in the Makevars.in file checking
is_windows <- .Platform[["OS.type"]] == "windows"

# if windows we replace in the Makevars.win.in
mv_fp <- ifelse(
  is_windows,
  "src/Makevars.win.in",
  "src/Makevars.in"
)

# set the output file
mv_ofp <- ifelse(
  is_windows,
  "src/Makevars.win",
  "src/Makevars"
)

# delete the existing Makevars{.win/.wasm}
if (file.exists(mv_ofp)) {
  message("Cleaning previous `", mv_ofp, "`.")
  invisible(file.remove(mv_ofp))
}

# read as a single string
mv_txt <- readLines(mv_fp)

# replace placeholder values
new_txt <- gsub("@CRAN_FLAGS@", .cran_flags, mv_txt) |>
  gsub("@PROFILE@", .profile, x = _) |>
  gsub("@CLEAN_TARGET@", .clean_targets, x = _) |>
  gsub("@LIBDIR@", .libdir, x = _) |>
  gsub("@TARGET@", .target, x = _) |>
  gsub("@PANIC_EXPORTS@", .panic_exports, x = _) |>
  gsub("@TARGET_DIR@", .target_dir, x = _)

message("Writing `", mv_ofp, "`.")
con <- file(mv_ofp, open = "wb")
writeLines(new_txt, con, sep = "\n")
close(con)

message("`tools/config.R` has finished.")

#' Combine multiple data.frames
#'
#' A general function that takes a list of `data.frame`s and returns a single
#' and combines them into a single object. It will use the fastest method
#' available. In order this is [`collapse::rowbind()`], [`data.table::rbindlist()`],
#' [`vctrs::list_unchop()`], then `do.call(rbind.data.frame, x)`.
#'
#' If all items in the list are `data.frame`s, then the result will be a `data.frame`.
#' If all elements are an `sf` object, then the result will be an `sf` object.
#' If the items are mixed, the result will be a `data.frame`.
#'
#' If any items are `NULL`, then an attribute `null_elements` will be attached
#' to the result. The attribute is an integer vector of the indices that
#' were `NULL`.
#'
#' @param x a list where each element is a `data.frame` or `NULL`.
#' @param .ptype currently unused. Reserved for a future release.
#' @export
#' @returns see details.
#' @inheritParams parse_esri_json
#' @examples
#'
#' x <- head(iris)
#' res <- rbind_results(list(x, NULL, x))
#' attr(res, "null_elements")
rbind_results <- function(
  x,
  call = rlang::current_env(),
  .ptype = data.frame()
) {
  # use for loop for side effects
  # check that each element is a data.frame
  for (item in x) {
    check_data_frame(
      item,
      allow_null = TRUE,
      call = call,
      arg = rlang::caller_arg(x)
    )
  }

  # check if all results are sf, if so, we must return sf
  return_sf <- all(vapply(x, inherits_or_null, logical(1), class = "sf"))

  # identify which results are missing
  missing_elements <- vapply(x, is.null, logical(1))
  are_missing <- which(missing_elements)

  if (all(missing_elements)) {
    # return empty data.frame() if all missing
    # FIXME should we take a ptype here? to
    return(structure(data.frame(), null_elements = are_missing))
  }

  if (rlang::is_installed("collapse", version = "2.0.0")) {
    # ensure that a data.frame is always returned via return = 2L
    x <- collapse::rowbind(x, return = 2L, fill = TRUE)
  } else if (rlang::is_installed("data.table")) {
    x <- data.table::rbindlist(x)
    data.table::setDF(x)
  } else if (rlang::is_installed("vctrs")) {
    # vctrs::vec_rbind() doesn't handle NULL
    x <- vctrs::list_unchop(x, error_call = call)
  } else {
    x <- do.call(rbind.data.frame, x)
  }

  # cast to sf if not already (the case with collapse)
  if (return_sf && !rlang::inherits_any(x, "sf")) {
    x <- sf::st_as_sf(x)
  }

  if (length(are_missing) > 0) {
    attr(x, "null_elements") <- are_missing
  }

  x
}

#' Check if an object is NULL or inherits a class
#'
#' Uses [`rlang::inherits_any()`] for the class check.
#' @keywords internal
#' @noRd
inherits_or_null <- function(x, class) {
  if (is.null(x)) {
    return(TRUE)
  } else {
    rlang::inherits_any(x, class)
  }
}

test_that("rbind sf objects", {
  nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"))
  x <- list(nc, nc, NULL, nc)

  res <- rbind_results(x)

  # should be an sf object
  expect_true(inherits(res, "sf"))
  # should have 1 missing element
  expect_identical(attr(res, "null_elements"), 3L)
})

test_that("rbind mixed results", {
  nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"))
  items <- list(as.data.frame(nc), nc)

  res <- rbind_results(items)
  expect_true(rlang::inherits_only(res, "data.frame"))
})

test_that("rbind data.frames", {
  res <- rbind_results(list(iris, NULL, iris))
  expect_true(rlang::inherits_only(res, "data.frame"))
})

test_that("rbind NULL & list(NULL)", {
  # should return empty df
  res <- rbind_results(NULL)
  expect_identical(res, structure(data.frame(), null_elements = integer()))

  # one null
  res <- rbind_results(list(NULL))
  expect_identical(res, structure(data.frame(), null_elements = 1L))

  # multiple
  res <- rbind_results(list(NULL, NULL))
  expect_identical(res, structure(data.frame(), null_elements = 1:2))
})

test_that("rbind errors on non-df objects", {
  expect_error(rbind_results(list(iris, NULL, "a")))
})

test_that("rbind works with list columns and collapse", {
  # this object caused pain
  x <- list(
    loc_name = "World",
    extents = list(
      c(
        xmin = -51.129958180141,
        ymin = 0.023178982724,
        xmax = -51.127958180141,
        ymax = 0.025178982724
      )
    ),
    geometry = list(
      c(-51.128958180141, 0.024178982724) |>
        structure(class = c("XY", "POINT", "sfg"))
    ) |>
      structure(
        class = c("sfc_POINT", "sfc"),
        precision = 0,
        bbox = c(
          xmin = -51.128958180141,
          ymin = 0.024178982724,
          xmax = -51.128958180141,
          ymax = 0.024178982724
        ) |>
          structure(class = "bbox"),
        crs = list(input = NA_character_, wkt = NA_character_) |>
          structure(class = "crs"),
        n_empty = 0L
      )
  ) |>
    structure(
      row.names = 1L,
      sf_column = "geometry",
      agr = factor(
        c(loc_name = NA_character_, extents = NA_character_),
        levels = c("constant", "aggregate", "identity")
      ),
      class = c("sf", "data.frame")
    )

  to_bind <- list(NULL, x, NULL, x)

  res <- rbind_results(to_bind)
  expect_equal(rbind(x, x), res, ignore_attr = TRUE)
})

#' @export
#' @name gp_params
#' @inheritParams arc_base_req
#' @examples
#' x <- list(
#'   xmin = -122.4195,
#'   ymin = 37.330219000000056,
#'   xmax = -122.030757,
#'   ymax = 37.77650360000007,
#'   spatialReference = list(wkid = 4326L, latestWkid = 4326L)
#' )
#'
#' from_envelope(x)
from_envelope <- function(x, error_call = rlang::caller_call()) {
  e_msg <- "Envelope must be a bare list with fields {.code xmin}, {.code ymin}, {.code xmax}, {.code ymax}, and {.code spatialReference}"

  if (!rlang::is_bare_list(x)) {
    cli::cli_abort(
      e_msg,
      call = error_call
    )
  }

  if (
    !all(names(x) %in% c("xmin", "ymin", "xmax", "ymax", "spatialReference"))
  ) {
    cli::cli_abort(
      e_msg,
      call = error_call
    )
  }

  crs <- from_spatial_reference(x$spatialReference, error_call)

  sf::st_bbox(
    unlist(x[c("xmin", "ymin", "xmax", "ymax")]),
    crs = crs
  )
}

#' Authenticate with Shiny
#'
#' @description
#'
#' `oath_provider_arcgis()` and `auth_shiny()` provide an integrated
#' authentication experience for `{shiny}` applications via the package
#' [shinyOAuth](https://github.com/lukakoning/shinyOAuth/). Applications that
#' use `auth_shiny()` will enable user-specific behavior and authentications.
#'
#' @details
#'
#' Authentication with `auth_shiny()` requires the package `shinyOAuth`.
#' When creating an OAuth app in ArcGIS Online / Enterprise, ensure that
#' the app has a valid redirect URI. This **must be** the same redirect URI
#' that is used by `shinyOAuth` and **must be** the same port as the
#' shiny application is served on.
#'
#' `oauth_provider_arcgis()` is a low-level provider function that mimics
#' other `oauth_provider_*()` functions from `shinyOAuth`. Use `auth_shiny()`
#' to create an OAuth client.
#'
#' **Important:** The client object returned by `auth_shiny()` must be created
#' outside of the server function to ensure a shared state across the OAuth flow.
#' Creating the client inside the server function will result in authentication failures.
#'
#' The user info returned from `auth$token@userinfo` has the same structure
#' as [arc_portal_self()]. See that function's documentation for available fields.
#'
#' The below example is derived from the shinyOAuth documentation.
#'
#' ```r
#' library(shiny)
#' library(shinyOAuth)
#' library(arcgisutils)
#'
#' # Simple UI
#' ui <- fluidPage(
#'   use_shinyOAuth(),
#'   uiOutput("login_information")
#' )
#'
#' client <- auth_shiny()
#'
#' server <- function(input, output, session) {
#'   # Set auto_redirect = FALSE for manual login
#'   auth <- shinyOAuth::oauth_module_server(
#'     "AGOL",
#'     client,
#'     auto_redirect = FALSE
#'   )
#'
#'   output$login_information <- renderUI({
#'     if (auth$authenticated) {
#'       user_info <- auth$token@userinfo
#'       tagList(
#'         tags$p("You are logged in! Your details:"),
#'         tags$pre(paste(capture.output(str(user_info)), collapse = "\n"))
#'       )
#'     } else {
#'       tagList(
#'         tags$p("You are not logged in."),
#'         actionButton("login_btn", "Login with ArcGIS")
#'       )
#'     }
#'   })
#'
#'   # Trigger login when button is clicked
#'   observeEvent(input$login_btn, {
#'     auth$request_login()
#'   })
#' }
#'
#' runApp(
#'   shinyApp(ui, server),
#'   port = 8100,
#'   launch.browser = FALSE
#' )
#'
#' # Open the app in your regular browser at http://localhost:8100
#' ```
#'
#' @export
#' @name auth_shiny
oauth_provider_arcgis <- function(
  host = arc_host()
) {
  ver <- utils::packageVersion("arcgisutils")
  arc_ver <- paste0("arcgisutils v", ver)

  rlang::check_installed("shinyOAuth")
  shinyOAuth::oauth_provider(
    name = arc_ver,
    auth_url = paste0(host, "/sharing/rest/oauth2/authorize"),
    token_url = paste0(host, "/sharing/rest/oauth2/token"),
    token_auth_style = "body",
    userinfo_url = paste0(host, "/sharing/rest/portals/self?f=json"),
    userinfo_id_selector = function(x) x$id
  )
}

#' @param redirect_uri default `Sys.getenv("ARCGIS_REDIRECT_URI")`. The redirect URL after completing authentication flow.
#' @param ... additional arguments passed to [shinyOAuth::oauth_client()]
#' @inheritParams auth_client
#' @export
#' @name auth_shiny
auth_shiny <- function(
  client = Sys.getenv("ARCGIS_CLIENT"),
  secret = Sys.getenv("ARCGIS_SECRET"),
  redirect_uri = Sys.getenv("ARCGIS_REDIRECT_URI"),
  host = arc_host(),
  ...
) {
  rlang::check_installed("shinyOAuth")
  dots <- rlang::list2(...)
  check_dots_named(dots)

  shinyOAuth::oauth_client(
    provider = oauth_provider_arcgis(host),
    client_id = client,
    client_secret = secret,
    redirect_uri = redirect_uri,
    ...
  )
}

#' Parse an ArcGIS service or content URL into its components
#'
#' [arc_url_parse()] uses [httr2::url_parse()] to parse URL components and
#' combine the components with a service or content URL `type` and a `layer`
#' number if applicable. A `layer` component is only included if the `type` is `"MapServer"` or
#' `"FeatureServer"` and the URL includes a trailing digit. A full `url` value is
#' also included in the returned list. The `url`, `type`, and `layer` components
#' are not part of the `httr2_url` class object returned by
#' [httr2::url_parse()].
#'
#' `r lifecycle::badge("experimental")`
#'
#' @inheritParams httr2::url_parse
#' @inheritParams arc_base_req
#' @examples
#' arc_url_parse(
#'   "https://services.arcgisonline.com/arcgis/rest/services/USA_Topo_Maps/MapServer/0"
#' )
#' arc_url_parse(
#'   "https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer"
#' )
#' arc_url_parse(
#'   "https://services.arcgisonline.com/arcgis/rest/services/WorldElevation3D/Terrain3D/ImageServer"
#' )
#'
#' @returns A named list with the following components: scheme, hostname, username, password, port, path, query, fragment, url, type, and layer.
#' @export
#' @name url
arc_url_parse <- function(
  url,
  base_url = NULL,
  error_call = rlang::caller_call()
) {
  check_string(url, call = error_call)
  check_string(base_url, allow_null = TRUE, call = error_call)

  # url encode the base url
  if (!is.null(base_url)) {
    base_url <- utils::URLencode(base_url)
  }

  # Parse URL into components
  httr2_url <- httr2::url_parse(
    url = utils::URLencode(url),
    base_url = base_url
  )

  # Extract the service or content type from URL
  type <- arc_url_type(url)

  # Extract trailing layer number from URL path
  layer <- NULL
  if (has_arc_url_layer(httr2_url[["path"]])) {
    layer <- str_extract(
      httr2_url[["path"]],
      "(?<=/)([0-9]+(?=/)|[0-9]+$)"
    )
  }

  c(
    unclass(httr2_url),
    list(
      url = httr2::url_build(httr2_url),
      type = type,
      layer = layer
    )
  )
}

#' Does the path have a layer at the end?
#' @noRd
has_arc_url_layer <- function(
  url,
  services = c("FeatureServer", "MapServer"),
  layer = NULL
) {
  layer <- layer %||% "[0-9]+"
  pattern <- paste0(
    "(",
    paste0(services, collapse = "|"),
    ")/",
    layer,
    "/?$"
  )

  grepl(pattern, url)
}


#' @export
#' @name url
arc_url_type <- function(url, error_call = rlang::caller_call()) {
  check_string(url, call = error_call)

  pattern <- c(
    rlang::set_names(
      arc_service_types,
      arc_service_types
    ),
    arc_content_types
  )

  # TODO: Add check for "rest/services" or other patterns to ensure the URL is a
  # valid ArcGIS service or content url and warn otherwise
  matches <- vapply(pattern, grepl, x = url, FUN.VALUE = logical(1))

  if (!any(matches)) {
    return(NULL)
  }

  names(pattern[matches])
}

#' @export
#' @name url
#' @inheritParams arc_base_req
is_url <- function(
  url,
  error_call = rlang::caller_call()
) {
  check_character(url, call)
  if (!rlang::is_vector(url) || rlang::is_empty(url)) {
    return(FALSE)
  }

  url_pattern <-
    "http[s]?://(?:[[:alnum:]]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"

  grepl(url_pattern, url)
}


#' Vector of valid ArcGIS service types
#'
#' Length 9 character vector
#' @noRd
arc_service_types <- c(
  "FeatureServer",
  "MapServer",
  "ImageServer",
  "GeoDataServer",
  "GeocodeServer",
  "GeometryServer",
  "GPServer",
  "WFSServer",
  "WFCServer",
  "SceneServer"
)

#' Vector of valid ArcGIS layer types
#'
#' Length 3 character vector
#' @noRd
arc_layer_types <- c(
  "Feature Layer",
  "Table",
  "Group Layer"
)

#' Named vector of ArcGIS Content URL patterns
#'
#' Length 10 character vector of item or content URL patterns.
#' @noRd
arc_content_types <- c(
  "content" = "/content\\.html\\?view=",
  "search" = "/search\\.html",
  "item" = "/item\\.html\\?id=",
  "group" = "/group\\.html\\?id=",
  "user" = "/user\\.html\\?user=",
  "webscene" = "/home/webscene/viewer\\.html\\?webscene=",
  "webmap" = "(webmap/viewer|mapviewer/index)\\.html\\?webmap=",
  "app" = "/index\\.html\\?appid=",
  "notebook" = "/notebook/notebook\\.html\\?id=",
  "experience" = "/experience/[a-zA-Z0-9]+",
  "storymap" = "/stories/[a-zA-Z0-9]+",
  "dashboard" = "/dashboards/[a-zA-Z0-9]+",
  "datapipeline" = "/datapipelines/editor\\?item\\=",
  "webapp" = "webappviewer/index.html\\?id=[a-zA-Z0-9]+",
  "experience_builder" = "builder/\\?id=[a-zA-Z0-9]+"
)

#' @noRd
str_extract_layer <- function(x) {
  trailing_txt <- str_extract(
    x,
    "(?:\\/)([[:digit:]]+)(?:\\/?)$",
    perl = FALSE
  )

  str_extract(trailing_txt, "[[:digit:]]+", perl = FALSE)
}

test_that("arc_url_parse works", {
  furl <- "https://services.arcgisonline.com/arcgis/rest/services/USA_Topo_Maps/MapServer"

  furl_parsed <- arc_url_parse(furl)

  expect_identical(furl_parsed[["type"]], "MapServer")
  expect_identical(furl_parsed[["url"]], furl)

  furl_alt <- "https://services1.arcgis.com/hLJbHVT9ZrDIzK0I/arcgis/rest/services/North%20Carolina%20SIDS%20sample/FeatureServer"

  expect_identical(arc_url_parse(furl_alt)[["type"]], "FeatureServer")
})


test_that("arc_url_type() is inferred correctly", {
  skip_if_offline()

  test_cases <- c(
    map_server = "https://image.discomap.eea.europa.eu/arcgis/rest/services/Corine/CLC2000_WM/MapServer",
    feature_layer = "https://image.discomap.eea.europa.eu/arcgis/rest/services/Corine/CLC2000_WM/MapServer/0",
    feature_server = "https://services.arcgis.com/P3ePLMYs2RVChkJx/arcgis/rest/services/USA_Major_Cities_/FeatureServer",
    scene = "https://tiles.arcgis.com/tiles/oPre3pOfRfefL8y0/arcgis/rest/services/3D_Buildings_Switzerland_wgs84/SceneServer",
    tile_imagery = "https://image.arcgisonline.nl/arcgis/rest/services/KEA/Maximale_overstromingsdiepte/ImageServer",
    elevation = "https://tiles.arcgis.com/tiles/qHLhLQrcvEnxjtPr/arcgis/rest/services/British_National_Grid_Terrain_3D/ImageServer",
    webmap_app = "https://esri2.maps.arcgis.com/apps/instant/media/index.html?appid=80eb92ffc89b4086abe8cedd58ab160c",
    storymap = "https://storymaps.arcgis.com/stories/ad791fda858c46fdbe79636aa5f35dd8",
    instant_app = "https://actgov.maps.arcgis.com/apps/instant/interactivelegend/index.html?appid=f2dfd67d29ed4cabbb91e742e0297955",
    dashboard = "https://www.arcgis.com/apps/dashboards/84ba9c03786e462d960e3172bc1b2204",
    datapipeline = "https://analysis-1.maps.arcgis.com/apps/datapipelines/editor?item=10db3e15b22948faabf8ecdf9af30065",
    experience = "https://experience.arcgis.com/experience/6e360741bfd84db79d5db774a1147815",
    webapp = "https://governmentofbc.maps.arcgis.com/apps/webappviewer/index.html?id=950b4eec577a4dc5b298a61adab41c06",
    notebook_item = "https://geosaurus.maps.arcgis.com/home/item.html?id=9a9fca3f09bb41dd856c9cd4239b8519",
    notebook = "https://geosaurus.maps.arcgis.com/home/notebook/notebook.html?id=9a9fca3f09bb41dd856c9cd4239b8519",
    webscene = "https://analysis-1.maps.arcgis.com/home/webscene/viewer.html?webscene=7b506043536246faa4194d4c3d4c921b",
    group = "https://analysis-1.maps.arcgis.com/home/group.html?id=2f0ec8cb03574128bd673cefab106f39#overview",
    user = "https://analysis-1.maps.arcgis.com/home/user.html?user=jparry_ANGP",
    item_db = "https://analysis-1.maps.arcgis.com/home/item.html?id=84ba9c03786e462d960e3172bc1b2204",
    item_mapserver = "https://analysis-1.maps.arcgis.com/home/item.html?id=1d150c40d9f642cb8bd691017bf22cee",
    feature_collection = "https://analysis-1.maps.arcgis.com/home/item.html?id=24aa36ce1d7747c2b5a6aa57711d03fb",
    gp_server = "https://gis.pikepa.org/arcgis/rest/services/Utilities/GeocodingTools/GPServer",
    geometry_server = "https://sampleserver6.arcgisonline.com/arcgis/rest/services/Utilities/Geometry/GeometryServer"
  )

  expected_types <- c(
    "MapServer",
    "MapServer",
    "FeatureServer",
    "SceneServer",
    "ImageServer",
    "ImageServer",
    "app",
    "storymap",
    "app",
    "dashboard",
    "datapipeline",
    "experience",
    "webapp",
    "item",
    "notebook",
    "webscene",
    "group",
    "user",
    "item",
    "item",
    "item",
    "GPServer",
    "GeometryServer"
  )

  for (i in seq_along(test_cases)) {
    expect_identical(arc_url_type(test_cases[[i]]), expected_types[i])
  }
})

#' Validate CRS object
#'
#'
#' Takes a representation of a CRS and ensures that it is a valid one. The CRS
#' is validated using `sf::st_crs()` if it cannot be validated, a null CRS is returned.
#'
#' See [`sf::st_crs()`] for more details on valid representations.
#'
#' @param crs a representation of a coordinate reference system.
#' @inheritParams cli::cli_abort
#' @inheritParams rlang::caller_arg
#'
#' @export
#' @examples
#'
#' # using epsg code integer or string representation
#' validate_crs(3857)
#' validate_crs("EPSG:4326")
#'
#' # using a custom proj4 string
#' proj4string <- "+proj=longlat +datum=WGS84 +no_defs"
#'
#' crs <- validate_crs(proj4string)
#'
#' # using wkt2 (from above result)
#' crs <- validate_crs(crs$spatialReference$wkt)
#'
#' @returns
#'
#' Returns a list of length 1 with an element named `spatialReference` which is itself
#' a named list.
#'
#' If the provided CRS returns a valid well-known ID (WKID) `spatialReference` contains
#' a named element called `wkid` which is the integer value of the WKID. If the WKID
#' is not known but the CRS returned is a valid well-known text representation the `wkid` field
#' is `NA` and another field `wkt` contains the valid wkt.
#'
#'
validate_crs <- function(
  crs,
  arg = rlang::caller_arg(crs),
  call = rlang::caller_env()
) {
  if (rlang::is_null(crs)) {
    return(list(spatialReference = c()))
  }
  if (
    !((inherits(crs, "character") ||
      inherits(crs, "crs") ||
      inherits(crs, "numeric")))
  ) {
    cli::cli_abort(c(
      "Invalid CRS supplied: {.arg {arg}}",
      "i" = "must be compatible with {.fn sf::st_crs}",
      arg = arg,
      call = call
    ))
  }

  crs <- sf::st_crs(crs)

  # extract SRID
  srid <- crs$srid

  if (!is.null(srid) && !is.na(srid)) {
    wkid <- as.integer(strsplit(srid, ":")[[1]][2])
    wkt <- NULL
  } else {
    wkid <- NULL
    wkt <- crs$wkt
  }

  sr_components <- compact(list(wkid = wkid, wkt = wkt))
  list(spatialReference = sr_components)
}

#' Determine Esri Geometry type
#'
#' Takes an `sf` or `sfc` object and returns the appropriate Esri geometry type.
#'
#' ### Geometry type mapping
#'
#' - `POINT`:  `esriGeometryPoint`
#' - `MULTIPOINT`:  `esriGeometryMultipoint`
#' - `LINESTRING`:  `esriGeometryPolyline`
#' - `MULTILINESTRING`:  `esriGeometryPolyline`
#' - `POLYGON`:  `esriGeometryPolygon`
#' - `MULTIPOLYGON`:  `esriGeometryPolygon`
#'
#' @param x an object of class `data.frame`, `sf`, `sfc`, or `sfg`.
#' @inheritParams cli::cli_abort
#' @returns returns a character scalar of the corresponding Esri geometry type
#' @export
#' @examples
#' determine_esri_geo_type(sf::st_point(c(0, 0)))
determine_esri_geo_type <- function(x, call = rlang::caller_env()) {
  # if `geom` is a data.frame return NULL
  if (inherits(x, "data.frame") && !inherits(x, "sf")) return(NULL)

  geom_type <- as.character(sf::st_geometry_type(x, by_geometry = FALSE))

  switch(
    geom_type,
    "POINT" = "esriGeometryPoint",
    "MULTIPOINT" = "esriGeometryMultipoint",
    "LINESTRING" = "esriGeometryPolyline",
    "MULTILINESTRING" = "esriGeometryPolyline",
    "POLYGON" = "esriGeometryPolygon",
    "MULTIPOLYGON" = "esriGeometryPolygon",
    cli::cli_abort(
      "{.val {geom_type}} is not a supported Esri geometry type",
      call = call
    )
  )
}

#' Convert an object to an extent
#'
#' Given an sf or sfc object create a list that represents the extent of the
#' object. The result of this function can be parsed directly into json using
#' `jsonify::to_json(x, unbox = TRUE)` or included into a list as the extent
#' component that will be eventually converted into json using the above function.
#'
#' @param x an sf or sfc object
#' @param crs the CRS of the object. Must be parsable by `sf::st_crs()`
#' @inheritParams cli::cli_abort
#' @export
#' @examples
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' as_extent(nc)
#' @returns
#' An extent json object. Use `jsonify::to_json(x, unbox = TRUE)` to convert
#' to json.
as_extent <- function(x, crs = sf::st_crs(x), call = rlang::caller_env()) {
  # if a Table (no spatial dimensions) return NULL
  if (inherits(x, "data.frame") && !inherits(x, "sf")) {
    return(NULL)
  }

  if (is.na(crs)) {
    crs <- NULL
  } else {
    crs <- validate_crs(crs, call = call)
  }

  bbox <- as.list(sf::st_bbox(x))
  compact(c(bbox, crs))
}

#' Geoprocessing Parameter Types
#'
#' Functions for converting R objects to and from ArcGIS geoprocessing parameter
#' types. These functions handle the serialization and parsing of various data
#' types used in ArcGIS geoprocessing services.
#'
#' @details
#' `r lifecycle::badge("experimental")`
#'
#' This package provides support for the following geoprocessing parameter types:
#'
#' ## Implemented Types
#'
#' - **GPFeatureRecordSetLayer**: Feature collections with geometry and attributes
#' - **GPRecordSet**: Tabular data without geometry
#' - **GPRasterDataLayer**: Raster datasets from Portal items, Image Servers, or URLs
#' - **GPLinearUnit**: Linear distance measurements with units
#' - **GPArealUnit**: Area measurements with units
#' - **GPDate**: Date/time values in milliseconds since epoch
#' - **GPSpatialReference**: Coordinate reference systems
#'
#' ## Not Yet Implemented
#'
#' The following types are planned for future implementation:
#'
#' - **GPField**: Field definitions with name, type, and properties
#' - **GPMultiValue**: Arrays of values for a single data type
#' - **GPValueTable**: Flexible table-like objects with rows and columns
#' - **GPComposite**: Parameters that accept multiple data types
#' - **GPEnvelope**: Bounding box extents (use `as_extent()` for GPExtent)
#'
#' @section Usage Patterns:
#'
#' Most functions follow a consistent pattern:
#' - `as_gp_*()`: Convert R objects to geoprocessing parameter JSON
#' - `parse_gp_*()`: Parse geoprocessing response JSON to R objects
#' - Constructor functions (e.g., `gp_linear_unit()`, `gp_areal_unit()`) create typed S7 objects
#'
#' @section Examples:
#'
#' ```r
#' # Create a linear unit
#' distance <- gp_linear_unit(distance = 100, units = "esriMeters")
#'
#' # Convert spatial data to feature record set
#' as_gp_feature_record_set(my_sf_data)
#'
#' # Parse a geoprocessing response
#' parse_gp_feature_record_set(response_json)
#' ```
#'
#' @references [API Documentation](https://developers.arcgis.com/rest/services-reference/enterprise/gp-data-types/)
#'
#' @name gp_params
#' @export
#' @family geoprocessing
#' @param json raw json to parse
#' @examples
#' # create a feature record set
#' fset <- as_gp_feature_record_set(iris[1,])
#' fset
#'
#' # create fake gp feature record set to parse
#' fset_list <- list(
#'   list(
#'     dataType = "GPFeatureRecordSetLayer",
#'     paramName = "example",
#'     value = as_featureset(iris[1,])
#'   )
#' )
#'
#' # create the json
#' json <- yyjsonr::write_json_str(fset_list, auto_unbox = TRUE)
#'
#' # parse the record set json
#' parse_gp_feature_record_set(json)
#'
#' # linear units
#' lu <- gp_linear_unit(10, "esriMeters")
#' lu
#' as_gp_linear_unit(lu)
#'
#' # areal units
#' au <- gp_areal_unit(10, "esriSquareMeters")
#' au
#'
#' as_gp_areal_unit(au)
#'
#' # dates
#' json <- r"({
#'   "paramName": "Output_Date",
#'   "dataType": "GPDate",
#'   "value": 1199145600000
#' })"
#'
#' parse_gp_date(json)
parse_gp_feature_record_set <- function(json) {
  check_string(json, allow_empty = FALSE)

  geo <- parse_esri_json(
    json,
    query = "/0/value"
  )

  res <- RcppSimdJson::fparse(
    json,
    query = c(param_name = "/0/paramName", data_type = "/0/dataType")
  )

  # inject the geometry
  res[["geometry"]] <- geo
  res
}

#' @name gp_params
#' @export
#' @param x the object to convert into json
as_gp_feature_record_set <- function(x) {
  # this handles sf objects, data.frames, tbl, and tibble
  if (inherits(x, "data.frame")) {
    return(as_esri_featureset(x))
  }

  # this handles the other case of layers or item ids
  switch(
    class(x)[1],
    "PortalItem" = yyjsonr::write_json_str(
      list(
        itemID = x[["id"]]
      ),
      list(auto_unbox = TRUE)
    ),
    "FeatureLayer" = yyjsonr::write_json_str(
      list(url = x[["url"]]),
      list(auto_unbox = TRUE)
    ),
    "Table" = yyjsonr::write_json_str(
      list(url = x[["url"]]),
      list(auto_unbox = TRUE)
    ),
    NULL = cli::cli_abort(
      "Cannot convert {obj_type_friendly(x)} to a GPFeatureRecordSet"
    )
  )
}


# as_gp_feature_record_set(x)
# as_gp_feature_record_set(sfdep::guerry)
# x <- arc_item("35bfb71133704d89b62d33dc9282fdb4")
# x$id

# GPRecordSet ------------------------------------------------------------

#' @name gp_params
#' @export
parse_gp_record_set <- function(json) {
  res <- yyjsonr::read_json_str(json)
  res$value$fields <- data_frame(res$value$fields)
  res$value$features <- data_frame(res$value$fields)
  res
}


#' @name gp_params
#' @export
as_record_set <- function(x) {
  # this handles sf objects, data.frames, tbl, and tibble
  if (inherits(x, "data.frame")) {
    feats <- as_features(x)
    fields <- infer_esri_type(x)

    record_set <- yyjsonr::write_json_str(
      list(fields = fields, features = feats),
      auto_unbox = TRUE
    )

    return(record_set)
  }
  switch(
    class(x)[1],
    "PortalItem" = yyjsonr::write_json_str(
      list(
        itemID = x[["id"]]
      ),
      list(auto_unbox = TRUE)
    ),
    "FeatureLayer" = yyjsonr::write_json_str(
      list(url = x[["url"]]),
      list(auto_unbox = TRUE)
    ),
    "Table" = yyjsonr::write_json_str(
      list(url = x[["url"]]),
      list(auto_unbox = TRUE)
    ),
    "sf" = as_esri_featureset(x),
    NULL = cli::cli_abort(
      "Cannot convert {obj_type_friendly(x)} to a GPFeatureRecordSet"
    )
  )
}


# GPRasterDataLayer ------------------------------------------------------

#' @name gp_params
#' @export
as_gp_raster_layer <- function(x) {
  # must be either PortalItem, ImageServer, or URL

  # use item ID
  if (inherits(x, "PortalItem")) {
    res <- yyjsonr::write_json_str(
      list(
        itemID = x[["id"]]
      ),
      auto_unbox = TRUE
    )
    return(res)
  }

  # use image server
  if (inherits(x, "ImageServer")) {
    return(
      yyjsonr::write_json_str(
        list(url = x[["url"]]),
        auto_unbox = TRUE
      )
    )
  }

  # return URL
  if (rlang::is_string(x)) {
    if (is_url(x)) {
      return(
        yyjsonr::write_json_str(
          list(url = x, auto_unbox = TRUE)
        )
      )
    }
  }
}


# GPLinearUnit -----------------------------------------------------------

#' @name gp_params
#' @export
#' @param units a valid unit name. Must be one of "esriUnknownUnits", "esriInches", "esriPoints", "esriFeet", "esriYards", "esriMiles", "esriNauticalMiles", "esriMillimeters", "esriCentimeters", "esriMeters", "esriKilometers", "esriDecimalDegrees", "esriDecimeters", "esriIntInches", "esriIntFeet", "esriIntYards", "esriIntMiles", "esriIntNauticalMiles".
#' @param distance a scalar number of the distance.
gp_linear_unit <- S7::new_class(
  "GPLinearUnit",
  package = "arcgisutils",
  properties = list(
    distance = S7::class_numeric,
    units = S7::class_character
  ),
  validator = function(self) {
    check_number_decimal(self@distance)
    check_string(self@units, allow_empty = FALSE)
    valid_units <- c(
      "esriUnknownUnits",
      "esriInches",
      "esriPoints",
      "esriFeet",
      "esriYards",
      "esriMiles",
      "esriNauticalMiles",
      "esriMillimeters",
      "esriCentimeters",
      "esriMeters",
      "esriKilometers",
      "esriDecimalDegrees",
      "esriDecimeters",
      "esriIntInches",
      "esriIntFeet",
      "esriIntYards",
      "esriIntMiles",
      "esriIntNauticalMiles"
    )

    if (!self@units %in% valid_units) {
      cli::cli_abort("{.arg units} must be one of {.val {valid_units}}")
    }
  }
)

#' @name gp_params
#' @export
as_gp_linear_unit <- function(x) {
  if (!rlang::inherits_any(x, "arcgisutils::GPLinearUnit")) {
    cli::cli_abort(
      c(
        "Expected {.cls arcgisutils::GPLinearUnit} found {obj_type_friendly(x)}",
        "i" = "Create one with {.fn gp_linear_unit}"
      )
    )
  }

  yyjsonr::write_json_str(
    list(distance = x@distance, units = x@units),
    auto_unbox = TRUE
  )
}

#' @name gp_params
#' @export
parse_gp_linear_unit <- function(json) {
  rlang::inject(
    gp_linear_unit(!!!yyjsonr::read_json_str(json))
  )
}


#' @name gp_params
#' @export
#' @param area a scalar number of the measurement.
#' @param units the unit of the measurement. Must be one of "esriUnknownAreaUnits", "esriSquareInches", "esriSquareFeet", "esriSquareYards", "esriAcres", "esriSquareMiles", "esriSquareMillimeters", "esriSquareCentimeters", "esriSquareDecimeters", "esriSquareMeters", "esriAres", "esriHectares", "esriSquareKilometers", "esriSquareInchesUS", "esriSquareFeetUS", "esriSquareYardsUS", "esriAcresUS", "esriSquareMilesUS".
gp_areal_unit <- S7::new_class(
  "GPArealUnit",
  package = "arcgisutils",
  properties = list(
    area = S7::class_numeric,
    units = S7::class_character
  ),
  validator = function(self) {
    check_string(self@units, allow_empty = FALSE)
    valid_units <- c(
      "esriUnknownAreaUnits",
      "esriSquareInches",
      "esriSquareFeet",
      "esriSquareYards",
      "esriAcres",
      "esriSquareMiles",
      "esriSquareMillimeters",
      "esriSquareCentimeters",
      "esriSquareDecimeters",
      "esriSquareMeters",
      "esriAres",
      "esriHectares",
      "esriSquareKilometers",
      "esriSquareInchesUS",
      "esriSquareFeetUS",
      "esriSquareYardsUS",
      "esriAcresUS",
      "esriSquareMilesUS"
    )

    if (!self@units %in% valid_units) {
      cli::cli_abort("{.arg units} must be one of {.val {valid_units}}")
    }
  }
)

# GPArealUnit ------------------------------------------------------------

#' @name gp_params
#' @export
as_gp_areal_unit <- function(x) {
  if (!rlang::inherits_any(x, "arcgisutils::GPArealUnit")) {
    cli::cli_abort(
      c(
        "Expected {.cls arcgisutils::GPArealUnit} found {obj_type_friendly(x)}",
        "i" = "Create one with {.fn gp_areal_unit}"
      )
    )
  }

  yyjsonr::write_json_str(
    list(area = x@area, units = x@units),
    auto_unbox = TRUE
  )
}

#' @name gp_params
#' @export
parse_gp_areal_unit <- function(json) {
  rlang::inject(
    gp_areal_unit(!!!yyjsonr::read_json_str(json))
  )
}


# GPDate -----------------------------------------------------------------

#' @name gp_params
#' @export
as_gp_date <- function(x) {
  date_to_ms(x)
}

#' @name gp_params
#' @export
parse_gp_date <- function(json) {
  res <- yyjsonr::read_json_str(json)
  res$value <- from_esri_date(as.numeric(res$value))
  res
}

# json <- r"({
#   "paramName": "Output_Date",
#   "dataType": "GPDate",
#   "value": 1199145600000
# })"

# parse_gp_date(json)

# GPSpatialReference and GPCoordinateSystem ------------------------------

# TODO: Handle OUTPUT

#' @name gp_params
#' @export
as_spatial_reference <- function(x) {
  crs <- validate_crs(x)
  yyjsonr::write_json_str(crs, auto_unbox = TRUE)
}


#' @name gp_params
#' @export
#' @inheritParams arc_base_req
#' @param sr a list with fields `latestWkid`, `wkid`, or `wkt` representing a spatial reference
#' @examples
#' sr <- list(wkid = 4326L)
#' from_spatial_reference(sr)
from_spatial_reference <- function(sr, error_call = rlang::caller_call()) {
  if (!rlang::is_list(sr)) {
    cli::cli_abort(
      "spatial reference must be a bare list with fields {.code latestWkid}, {.code wkid}, or {.code wkt}",
      call = error_call
    )
  }

  crs <- sr[["latestWkid"]] %||% sr[["wkid"]] %||% sr[["wkt"]]

  # WKIDs 32767 and higher fall within the ESRI authority
  if (is.numeric(crs) && !is.na(crs) && crs >= 32767) {
    crs <- paste0("ESRI:", crs)
  }

  sf::st_crs(crs)
}

#' @name gp_params
#' @export
parse_spatial_reference <- function(json) {
  sr <- yyjsonr::read_json_str(json)
  from_spatial_reference(sr)
}

test_that("from_spatial_reference works for EPSG WKIDs", {
  sr <- list(wkid = 4326L)

  expect_identical(
    from_spatial_reference(sr),
    sf::st_crs(sr$wkid)
  )
})

test_that("from_spatial_reference works for ESRI WKIDs", {
  sr <- list(wkid = 102003L)

  expect_identical(
    from_spatial_reference(sr),
    sf::st_crs(paste0("ESRI:", sr$wkid))
  )
})

test_that("from_spatial_reference works for NA WKIDs", {
  sr <- list(wkid = NA_integer_)

  expect_identical(
    from_spatial_reference(sr),
    sf::NA_crs_
  )
})

#' Create Esri FeatureSet Objects
#'
#' These functions create an Esri FeatureSet object. A FeatureSet contains an inner array
#' of features as well as additional metadata about the the collection such as the
#' geometry type, spatial reference, and object ID field.
#'
#' @inheritParams as_esri_features
#' @export
#' @rdname featureset
#' @references [API Reference](https://developers.arcgis.com/documentation/common-data-types/featureset-object.htm)
#' @examples
#'
#' library(sf)
#' # POINT
#' # create sfg points
#' xy <- st_sfc(st_point(c(1, 2)))
#' xyz <- st_sfc(st_point(c(1, 2, 3)))
#' xym <- st_sfc(st_point(c(1, 2, 3), dim = "XYM"))
#'
#' as_esri_featureset(xy)
#' as_esri_featureset(xyz)
#' as_esri_featureset(xym)
#'
#' # MULTIPOINT
#' # vector to create matrix points
#' set.seed(0)
#' x <- rnorm(12)
#'
#' xy <- st_sfc(st_multipoint(matrix(x, ncol = 2)))
#' xyz <- st_sfc(st_multipoint(matrix(x, ncol = 3)))
#' xym <- st_sfc(st_multipoint(matrix(x, ncol = 3), dim = "XYM"))
#'
#' as_esri_featureset(xy)
#' as_esri_featureset(xyz)
#' as_esri_featureset(xym)
#'
#' # LINESTRING
#' xy <- st_sfc(st_linestring(matrix(x, ncol = 2)))
#' xyz <- st_sfc(st_linestring(matrix(x, ncol = 3)))
#' xym <- st_sfc(st_linestring(matrix(x, ncol = 3), dim = "XYM"))
#'
#' as_esri_featureset(xy)
#' as_esri_featureset(xyz)
#' as_esri_featureset(xym)
#'
#' # MULTILINESTRING
#' as_esri_featureset(st_sfc(st_multilinestring(list(xy[[1]], xy[[1]]))))
#' as_esri_featureset(st_sfc(st_multilinestring(list(xyz[[1]], xyz[[1]]))))
#' as_esri_featureset(st_sfc(st_multilinestring(list(xym[[1]], xym[[1]]))))
#'
#' # POLYGON
#' coords <- rbind(
#'   c(0, 0, 0, 1),
#'   c(0, 1, 0, 1),
#'   c(1, 1, 1, 1),
#'   c(1, 0, 1, 1),
#'   c(0, 0, 0, 1)
#' )
#'
#' xy <- st_sfc(st_polygon(list(coords[, 1:2])))
#' xyz <- st_sfc(st_polygon(list(coords[, 1:3])))
#' xym <- st_sfc(st_polygon(list(coords[, 1:3]), dim = "XYM"))
#'
#' as_esri_featureset(xy)
#' as_esri_featureset(xyz)
#' as_esri_featureset(xym)
#'
#' # MULTIPOLYGON
#' as_esri_featureset(st_sfc(st_multipolygon(list(xy[[1]], xy[[1]]))))
#' as_esri_featureset(st_sfc(st_multipolygon(list(xyz[[1]], xyz[[1]]))))
#' as_esri_featureset(st_sfc(st_multipolygon(list(xym[[1]], xym[[1]]))))
#' @returns a list or a json string
as_featureset <- function(x, crs = sf::st_crs(x), call = rlang::caller_env()) {
  # class check
  valid_sfg_classes <- c(
    "sf",
    "data.frame",
    "sfc"
  )

  # exit if an invalid geometry type is provided
  if (!rlang::inherits_any(x, valid_sfg_classes)) {
    cli::cli_abort(
      "{.arg x} must inherit one of the following classes {.cls {valid_sfg_classes}}",
      call = call
    )
  }

  sr <- validate_crs(crs)[[1]] %||% list()

  # store the class name to be used in a switch statement
  x_class <- inherits_which(x, valid_sfg_classes)[1]

  switch(
    x_class,
    "sf" = as_featureset_sf(x, sr, call = call),
    "data.frame" = as_featureset_sf(x, sr, call = call),
    "sfc" = as_featureset_sfc(x, sr, call = call)
  )
}

as_featureset_sfc <- function(x, crs = NULL, call = rlang::caller_env()) {
  # TODO handle CRS
  sr <- crs %||% list()

  # class check
  valid_sfg_classes <- c(
    "sfc_POINT",
    "sfc_MULTIPOINT",
    "sfc_LINESTRING",
    "sfc_MULTILINESTRING",
    "sfc_POLYGON",
    "sfc_MULTIPOLYGON"
  )

  # exit if an invalid geometry type is provided
  if (!rlang::inherits_any(x, valid_sfg_classes)) {
    cli::cli_abort(
      "{.arg x} must inherit one of the following classes {.cls {valid_sfg_classes}}",
      call = call
    )
  }

  # store the class name to be used in a switch statement
  sfc_class <- inherits_which(x, valid_sfg_classes)

  # dimension check
  z <- has_z(x)
  m <- has_m(x)

  # abort if 4D
  if (z && m) {
    cli::cli_abort(
      c(
        "{.cls XYZM} geometries detected. Only {.cls XY}, {.cls XYZ}, and {.cls XYM} geometries supported. ",
        ">" = "{.href [please make an issue on GitHub](https://github.com/r-arcgis/arcgisutils)}"
      ),
      call = call
    )
  }

  # determine if 3D
  three_dim <- z || m

  # switch based on dimensions
  if (three_dim) {
    switch(
      sfc_class,
      "sfc_POINT" = sfc_point_featureset_3d_list(x, sr),
      "sfc_MULTIPOINT" = sfc_multipoint_featureset_3d_list(x, sr),
      "sfc_LINESTRING" = sfc_linestring_featureset_3d_list(x, sr),
      "sfc_MULTILINESTRING" = sfc_multilinestring_featureset_3d_list(x, sr),
      "sfc_POLYGON" = sfc_polygon_featureset_3d_list(x, sr),
      "sfc_MULTIPOLYGON" = sfc_multipolygon_featureset_3d_list(x, sr),
    )
  } else {
    switch(
      sfc_class,
      "sfc_POINT" = sfc_point_featureset_2d_list(x, sr),
      "sfc_MULTIPOINT" = sfc_multipoint_featureset_2d_list(x, sr),
      "sfc_LINESTRING" = sfc_linestring_featureset_2d_list(x, sr),
      "sfc_MULTILINESTRING" = sfc_multilinestring_featureset_2d_list(x, sr),
      "sfc_POLYGON" = sfc_polygon_featureset_2d_list(x, sr),
      "sfc_MULTIPOLYGON" = sfc_multipolygon_featureset_2d_list(x, sr),
    )
  }
}


as_featureset_sf <- function(x, crs = NULL, call = rlang::caller_env()) {
  # must be a data.frame
  check_data_frame(x)

  # TODO handle CRS
  sr <- crs %||% list()

  # get the geometry column
  geom_col <- attr(x, "sf_column") %||% ""
  # if it is null then make it into a list
  .geoms <- x[[geom_col]] %||% list()

  # determine dims
  if (rlang::is_empty(.geoms)) {
    z <- FALSE
    m <- FALSE
  } else {
    z <- has_z(.geoms)
    m <- has_m(.geoms)
  }

  # abort if 4D
  if (z && m) {
    cli::cli_abort(
      c(
        "{.cls XYZM} geometries detected. Only {.cls XY}, {.cls XYZ}, and {.cls XYM} geometries supported. ",
        ">" = "{.href [please make an issue on GitHub](https://github.com/r-arcgis/arcgisutils)}"
      ),
      call = call
    )
  }

  # check if three dim
  three_dim <- z || m

  # extract the data.frame without using sf
  # we would otherwise use st_drop_geometry
  .data <- x
  .data[[geom_col]] <- NULL
  # we've verified that x is a data.frame so here we remove any other subclasses
  attr(.data, "class") <- "data.frame"

  # find any factors and convert them to character
  factor_check <- which(vapply(.data, is.factor, logical(1)))

  # convert to factor
  # this wont execute if there aren't any factors
  for (idx in factor_check) {
    .data[[idx]] <- as.character(.data[[idx]])
  }

  are_dates <- which(vapply(.data, is_date, logical(1)))
  for (col in are_dates) {
    .data[[col]] <- date_to_ms(x[[col]])
  }

  # find columns that cannot be supported
  esri_types <- infer_esri_type(.data)
  invalid_types <- is.na(esri_types$type)
  bad_names <- esri_types$name[invalid_types]

  # here we extract the friendly label for error printing
  bad_types <- lapply(.data[, invalid_types], obj_type_friendly)

  # report an error in the case that theres a list column or something else weird
  if (!rlang::is_empty(bad_types)) {
    cli::cli_abort(
      c(
        "{.var {bad_names}} cannot be converted into EsriJSON. Found {bad_types}"
      ),
      call = call
    )
  }

  # use the appropriate function based on if there is a Z or M value
  # note that st_as_features_2d is used in the case of data.frame
  if (three_dim) {
    as_featureset_3d_list(.data, .geoms, nrow(.data), sr)
  } else {
    as_featureset_2d_list(.data, .geoms, nrow(.data), sr)
  }
}

test_that("date handling is correct", {
  skip_if_not_installed("jsonify")
  # create sf object
  x <- sf::st_sf(
    data.frame(today = as.POSIXct("2001-01-01", tz = "UTC")),
    geometry = sf::st_sfc(sf::st_point(c(0, 1)), crs = 4326)
  )

  # create json object
  l <- RcppSimdJson::fparse(as_esri_featureset(x))
  # l <- as_featureset(x)
  # add fields since they are not included by default
  l[["fields"]] <- infer_esri_type(x)

  # create featurset geometry
  json <- jsonify::to_json(l)

  parsed <- parse_esri_json(json)

  expect_identical(parsed$today, x$today)
})

# create a package level environment that will be used to store the tokens
# This is where all tokens are going to go
token_env <- rlang::env()

#' Manage authorization tokens
#'
#' These functions are used to set, fetch, and check authorization tokens.
#'
#' @param token for `arc_token()`, the name of a token to fetch. For `set_arc_token()`,
#'   it is an `httr2_token` that will be set. For `unset_arc_token()`, a character
#'   vector of token names to be unset.
#' @param ... named arguments to set `httr2_token`. Must be valid names and must be an `httr2_token`.
#' @inheritParams cli::cli_abort
#' @details
#'
#' It is possible to have multiple authorization tokens in one session. These
#' functions assist you in managing them.
#'
#' `arc_token()` is used to fetch tokens by name. The default token is `ARCGIS_TOKEN`.
#' However, they can be any valid character scalar. `set_arc_token()` will create
#' store a token with the name `ARCGIS_TOKEN`. However, you can alternatively
#' set the tokens by name using a key-value pair. The key is what you would pass
#' to `arc_token()` to fetch the `httr2_token` object. To remove a token that has
#' been set, use `unset_arc_token()`.
#'
#' `obj_check_token()` is a developer oriented function that can be used to check
#' if an object is indeed an `httr2_token`. To check if a token has expired,
#' [`validate_or_refresh_token()`] will do so.
#'
#' `check_token_has_user()` is a developer oriented function that checks to see
#' if a token has a `username` field associated with it.
#'
#' For developers:
#'
#' `set_arc_token()` uses a package level environment to store the tokens. The
#' tokens are fetched from the environment using `arc_token()`.
#'
#' @export
#' @rdname token
#' @examples
#' # create fake tokens
#' token_a <- httr2::oauth_token("1234", arcgis_host = arc_host())
#' token_b <- httr2::oauth_token("abcd", arcgis_host = arc_host())
#'
#' # set token to the default location
#' set_arc_token(token_a)
#'
#' # fetch token from the default location
#' arc_token()
#'
#' # set token by name
#' set_arc_token(org_a = token_a, org_b = token_b)
#'
#' # fetch token by name
#' arc_token("org_a")
#' arc_token("org_b")
#'
#' # unset tokens
#' unset_arc_token()
#' unset_arc_token(c("org_a", "org_b"))
arc_token <- function(token = "ARCGIS_TOKEN") {
  # returns NULL if not found
  token_env[[token]]
}


#' @export
#' @rdname token
set_arc_token <- function(token, ...) {
  # one of the two must be provided
  if (rlang::is_missing(token) && rlang::dots_n(...) == 0) {
    cli::cli_abort("Must provide {.arg token} or {.arg ...}")
  }

  # if token arg is not missing set it
  # note that this must come before the for loop because it
  # adjusts the scoped variable `token`
  if (!rlang::is_missing(token)) {
    # check if token is the right class
    obj_check_token(token)
    # set the environment ARCGIS_TOKEN
    rlang::env_bind(token_env, "ARCGIS_TOKEN" = token)
  }

  # handle dots if present
  if (rlang::dots_n(...) > 0) {
    # capture dots
    dots <- rlang::list2(...)

    # check em!
    check_dots_named(dots)

    # check all tokens
    for (token in dots) {
      obj_check_token(token)
    }

    # bind tokens to token environment
    rlang::env_bind(token_env, ...)

    # NOTE Worry is being overly verbose here:
    # provide informative message on how to access
    cli::cli_alert_success("Named tokens set: {.var {rlang::names2(dots)}}\n")
    cli::cli_alert_info("Access named tokens with {.code arc_token(\"name\")}")
  }
}

#' @rdname token
#' @export
unset_arc_token <- function(token = NULL) {
  check_character(token, allow_null = TRUE)
  if (is.null(token)) {
    token <- "ARCGIS_TOKEN"
    rlang::env_unbind(token_env, token)
  } else {
    rlang::env_unbind(token_env, token)
  }
  cli::cli_alert_success("Token{?s} {.var {token}} {?has/have} been unset.")
}

#' @export
#' @rdname token
obj_check_token <- function(token, call = rlang::caller_env()) {
  # if object is not a httr2 token
  if (!rlang::inherits_only(token, "httr2_token")) {
    cli::cli_abort(
      "{.arg token} must be an {.cls httr2_token} not {.cls {class(token)}}",
      call = call
    )
  }

  # if no host is set
  if (is.null(token[["arcgis_host"]])) {
    cli::cli_abort(
      c(
        "{.arg token} does not have {.val arcgis_host}.",
        "i" = "was your token created using {.pkg arcgisutils}?"
      ),
      call = call
    )
  }

  # if more than one host is set, error
  if (length(token[["arcgis_host"]]) > 1) {
    cli::cli_abort(
      c(
        "{.arg token} has more than one {.val arcgis_host}.",
        "i" = "was your token created using {.pkg arcgisutils}?"
      ),
      call = call
    )
  }

  invisible(token)
}

#' @export
#' @rdname token
check_token_has_user <- function(token, call = rlang::caller_env()) {
  obj_check_token(token, call = call)
  if (is.null(token[["username"]])) {
    cli::cli_abort(
      c(
        "{.arg token} does not have an associated {.val username}."
      )
    )
  }
  invisible(token)
}

test_that("tokens set and fetched by name are identical", {
  # create fake tokens
  token_a <- httr2::oauth_token("1234", arcgis_host = arc_host())
  token_b <- httr2::oauth_token("abcd", arcgis_host = arc_host())

  # set token by name
  set_arc_token(org_a = token_a, org_b = token_b)

  # fetch token by name
  expect_identical(arc_token("org_a"), token_a)
  expect_identical(arc_token("org_b"), token_b)
})

test_that("Default token set and fetched", {

  # create fake token
  token <- httr2::oauth_token("1234", arcgis_host = arc_host())

  # set token to the default location
  set_arc_token(token)

  # fetch token from the default location
  expect_identical(arc_token(), token)
})


test_that("Default token and named tokens", {

  # create fake tokens
  token <- httr2::oauth_token("....", arcgis_host = arc_host())
  token_a <- httr2::oauth_token("1234", arcgis_host = arc_host())
  token_b <- httr2::oauth_token("abcd", arcgis_host = arc_host())


  # set token by name and default
  set_arc_token(token, org_a = token_a, org_b = token_b)

  # fetch token by name
  expect_identical(arc_token(), token)
  expect_identical(arc_token("org_a"), token_a)
  expect_identical(arc_token("org_b"), token_b)

})


test_that("Host must be set:", {
  token <- httr2::oauth_token("....")
  expect_error(
    obj_check_token(token),
    "`token` does not have \"arcgis_host\""
  )
})

test_that("Only one host can be set:", {
  token <- httr2::oauth_token("....", arcgis_host = c("a", "b"))
  expect_error(
    obj_check_token(token),
    "`token` has more than one \"arcgis_host\""
  )
})

# Date handling -----------------------------------------------------------

#' Date handling
#'
#' Esri date fields are represented as milliseconds from the Unix Epoch.
#'
#' - `is_date()`: checks if an object is a `Date` or `POSIXt` class object.
#' - `date_to_ms()` converts a date object to milliseconds from the Unix Epoch in the specified time zone.
#'
#' @param x an object of class `Date` or `POSIXt`. In the case of `is_date()`, any R object.
#' @inheritParams base::as.POSIXlt
#'
#' @rdname dates
#' @export
#' @returns
#' - `is_date()` returns a logical scalar
#' - `date_to_ms()` returns a numeric vector of times in milliseconds from the Unix Epoch in the specified time zone.
#' @examples
#'
#' today <- Sys.Date()
#'
#' is_date(today)
#'
#' date_to_ms(today)
#'
is_date <- function(x, tz) inherits(x, c("Date", "POSIXt"))

# a function to convert dates to ms
#' @rdname dates
#' @export
date_to_ms <- function(x, tz = "UTC") {
  if (!is_date(x)) {
    stop("`x` must inherit the `Date` or `POSIXt` class")
  }
  as.numeric(as.POSIXlt(x, tz = tz)) * 1000
}

#' @export
#' @rdname dates
from_esri_date <- function(x) {
  is_na <- x == -1
  x[is_na] <- NA_real_
  as.POSIXct(
    x / 1000,
    tz = "UTC",
    origin = "1970-01-01"
  )
}

# See https://rconsortium.github.io/S7/articles/packages.html#backward-compatibility
# enable usage of <S7_object>@name in package code
#' @rawNamespace if (getRversion() < "4.3.0") importFrom("S7", "@")
NULL

#' Geoprocessing Job Status
#'
#' Represents the status of a geoprocessing job.
#'
#' @param status a scalar character. Must be one of `"esriJobSubmitted"`,
#' `"esriJobWaiting"`, `"esriJobExecuting"`, `"esriJobSucceeded"`, `"esriJobFailed"`,
#' `"esriJobTimedOut"`, `"esriJobCancelling"`, or `"esriJobCancelled`".
#'
#' @export
#' @family geoprocessing
#' @return an object of class `arc_job_status`
#' @examples
#' arc_job_status("esriJobSubmitted")
arc_job_status <- S7::new_class(
  "arc_job_status",
  package = "arcgisutils",
  properties = list(status = S7::class_character),
  validator = function(self) {
    job_statuses <- c(
      "esriJobSubmitted",
      "esriJobWaiting",
      "esriJobExecuting",
      "esriJobSucceeded",
      "esriJobFailed",
      "esriJobTimedOut",
      "esriJobCancelling",
      "esriJobCancelled"
    )

    # permit length 0
    if (length(self@status) == 0L) {
      return(NULL)
    }
    # if it is longer than 0, it should have been set to something
    if (length(self@status) > 1) {
      cli::cli_abort("Job status must be a scalar value")
    } else if (!self@status %in% job_statuses) {
      cli::cli_abort("Job status must be one of {.val {job_statuses}}")
    }
  }
)


#' Form request parameters
#'
#' ArcGIS endpoints make extensive use of form encoded data for the body
#' of http requests. Form requests require that each element has a name
#' and is encoded as a single string—often as json.
#'
#' The `arc_form_params` class provides validation of form body parameters
#' ensuring that each element is a scalar string. It uses a named list
#' internally to store the parameters.
#'
#' The helper function `as_form_params()` converts a named list to form
#' parameters by automatically JSON-encoding each element using
#' `yyjsonr::write_json_str()` with `auto_unbox = TRUE`.
#'
#' @export
#' @family geoprocessing
#' @return an object of class `arc_form_params`
#' @param params a named list with scalar character elements
#' @param x for `as_form_params()`, a named list to convert to form parameters
#' @examples
#' arc_form_params(
#'   list(f = "json", outFields = "*", where = "1 = 1")
#' )
arc_form_params <- S7::new_class(
  "arc_form_params",
  package = "arcgisutils",
  properties = list(params = S7::class_list),
  validator = function(self) {
    lapply(
      self@params,
      check_string,
      allow_null = TRUE,
      allow_na = FALSE,
      call = rlang::caller_call()
    )
    # must return null
    return(NULL)
  }
)

#' @export
#' @family geoprocessing
#' @rdname arc_form_params
as_form_params <- function(x) {
  if (!rlang::is_list(x)) {
    cli::cli_abort("{.arg x} must be a list.")
  }

  if (!rlang::is_named(x)) {
    cli::cli_abort("{.arg x} must be a named list.")
  }

  params <- lapply(x, \(.x) {
    if (rlang::is_string(.x)) {
      .x
    } else {
      yyjsonr::write_json_str(.x, auto_unbox = TRUE)
    }
  })
  arc_form_params(params)
}

#' @title Create a Geoprocessing Service Job
#'
#' @description
#' The `arc_gp_job` class is used to interact with Geoprocessing Services in
#' ArcGIS Online and Enterprise.
#'
#' @details
#'
#' The `arc_gp_job` uses S7 classes for the job request parameters and job status
#' via [arc_form_params()] and [arc_job_status()] respectively. Importantly,
#' [arc_form_params()] ensures that parameters provided to a geoprocessing
#' service are all character scalars as required by the form body.
#'
#' @section Associated Functions:
#' \describe{
#'   \item{\code{from_url(url, token = arc_token())}}{Create a GP Job object from an existing job URL}
#' }
#'
#' @importFrom R6 R6Class
#' @export
#' @returns
#' An object of class `arc_gp_job`.
#' @family geoprocessing
#' @rdname gp_job
#' @examples
#' url <- paste0(
#'   "https://logistics.arcgis.com/arcgis/",
#'   "rest/services/World/ServiceAreas/",
#'   "GPServer/GenerateServiceAreas"
#' )
#' job <- new_gp_job(url, list(f = "json"))
#' job
#'
#' # extract params S7 class
#' params <- job$params
#' params
#'
#' # view underlying list
#' params@params
arc_gp_job <- R6::R6Class(
  "arc_gp_job",
  #' @field base_url the URL of the job service (without `/submitJob`)
  #' @field id the ID of the started job. `NULL` `self$start()` has not been called.
  #' @field params returns an S7 object of class `arc_form_params` (see [`arc_form_params()`]) the list can be accessed via `self$params@params`.
  #' @field status returns the status of the geoprocessing job as an S7 object of class `gp_job_status` (see [arc_job_status()]) by querying the `/jobs/{job-id}` endpoint.
  #' @field results returns the current results of the job by querying the `/jobs/{job-id}/results` endpoint.
  public = list(
    base_url = NULL,
    id = NULL,
    #' @param base_url the URL of the job service (without `/submitJob`)
    #' @param params a named list where each element is a scalar character
    #' @param result_fn Default `NULL`. An optional function to apply to the results JSON. By default parses results using `RcppSimdJson::fparse()`.
    #' @param token default [arc_token()]. The token to be used with the job.
    #' @param error_call default `rlang::caller_call()` the calling environment.
    initialize = function(
      base_url,
      params = list(),
      result_fn = NULL,
      token = arc_token(),
      error_call = rlang::caller_call()
    ) {
      # use S7 to validate the form parameters
      if (!is_url(base_url)) {
        cli::cli_abort("{.arg base_url} is not a valid URL.", call = error_call)
      }

      if (!rlang::is_null(result_fn)) {
        check_function(result_fn)
        private$.result_fn <- result_fn
      }

      self$base_url <- base_url
      if (inherits(params, "arcgisutils::arc_form_params")) {
        private$.params <- params
      } else {
        private$.params <- arc_form_params(params)
      }

      private$token <- token
      self
    },
    #' @description  Starts the job by calling the `/submitJob` endpoint. This also sets the public field `id`.
    start = function() {
      # TODO make it possible to only do this once
      resp <- arc_base_req(
        self$base_url,
        token = private$token,
        path = "submitJob"
      ) |>
        httr2::req_body_form(!!!private$.params@params) |>
        httr2::req_error(is_error = function(e) FALSE) |>
        httr2::req_perform()

      res <- RcppSimdJson::fparse(httr2::resp_body_string(resp))
      detect_errors(res)
      self$id <- res$jobId
      self
    },

    #' @description Cancels a job by calling the `/cancel` endpoint.
    cancel = function() {
      resp <- arc_base_req(
        self$base_url,
        token = private$token,
        path = c("jobs", self$id, "cancel"),
        query = c(f = "json")
      ) |>
        httr2::req_body_form(!!!private$.params@params) |>
        httr2::req_error(is_error = function(e) FALSE) |>
        httr2::req_perform()

      res <- RcppSimdJson::fparse(httr2::resp_body_string(resp))

      # check for errors
      detect_errors(res)

      self$status <- arc_job_status(status = res$jobStatus)
      self$id <- res$jobId
      self
    },
    #' @description Waits for job completion and returns results.
    #' @param interval polling interval in seconds (default 0.1)
    #' @param verbose whether to print status messages (default FALSE)
    await = function(interval = 0.1, verbose = FALSE) {
      if (is.null(self$id)) {
        cli::cli_abort("Job has not been started.")
      }

      is_complete <- FALSE
      Sys.sleep(0.5) # initial sleep 500ms

      while (!is_complete) {
        cur_status <- self$status@status
        if (verbose) {
          cli::cli_alert_info(
            "Job status is {.val {cur_status}}. Waiting before polling again..."
          )
        }

        is_complete <- cur_status %in%
          c("esriJobFailed", "esriJobCancelled", "esriJobSucceeded")

        if (cur_status %in% c("esriJobCancelled", "esriJobFailed")) {
          cli::cli_alert_danger(
            "Job ID {.val {self$id}} ended with {.val {cur_status}}",
            ">" = "Returning GP results"
          )
          return(self$results)
        } else if (is_complete) {
          return(self$results)
        }
        Sys.sleep(interval)
      }
    },

    #' @description Fetches Geoprocessing Messages
    messages = function() {
      resp <- arc_base_req(
        self$base_url,
        token = private$token,
        path = c("jobs", self$id),
        query = c(f = "json")
      ) |>
        httr2::req_perform()

      res <- RcppSimdJson::fparse(httr2::resp_body_string(resp))

      # check for errors
      detect_errors(res)
      res
    }
  ),
  private = list(
    .params = NULL,
    .status = function() {
      if (is.null(self$id)) {
        return(NULL)
      }
      # check the status
      resp <- arc_base_req(
        self$base_url,
        token = private$token,
        path = c("jobs", self$id),
        query = c(f = "json")
      ) |>
        httr2::req_error(is_error = function(e) FALSE) |>
        httr2::req_perform() |>
        httr2::resp_body_string()

      # read the string
      res <- RcppSimdJson::fparse(resp)

      detect_errors(res)

      arc_job_status(res[["jobStatus"]])
    },
    .result_fn = NULL,
    token = NULL
  ),
  active = list(
    params = function() {
      private$.params
    },
    status = function() private$.status(),
    results = function() {
      # if there is a NULL job ID we abort
      if (is.null(self$id)) {
        cli::cli_abort(
          c(
            "There is no job ID present.",
            ">" = " Have you started the job with `x$start()`?"
          )
        )
      }

      # check the status
      resp <- arc_base_req(
        self$base_url,
        token = private$token,
        path = c("jobs", self$id, "results"),
        query = c(f = "json")
      ) |>
        httr2::req_error(is_error = function(e) FALSE) |>
        httr2::req_perform() |>
        httr2::resp_body_string()

      # optional .fn arg which is used to parse the results
      # of the geoprocessing body
      if (!rlang::is_null(private$.result_fn)) {
        return(private$.result_fn(resp))
      }

      # read the string
      res <- RcppSimdJson::fparse(resp)
      detect_errors(res)

      if (!is.null(res) && is.data.frame(res)) {
        return(data_frame(res))
      } else {
        res
      }
    }
  ),
  lock_class = TRUE
)

#' Create GP Job from existing URL
#'
#' @export
#' @family geoprocessing
#' @param url the url of an existing geoprocessing job
#' @inheritParams arc_gp_job
#'
#' @examples
#' if (interactive()) {
#' job_url <- paste0(
#'   "https://hydro.arcgis.com/arcgis/rest/services/Tools/Hydrology/",
#'   "GPServer/TraceDownstream/jobs/jfde67910074649e4a567f0adbb8af870"
#' )
#'
#' gp_job_from_url(
#'   job_url,
#'   token = auth_user()
#' )
#' }
gp_job_from_url = function(url, token = arc_token()) {
  obj_check_token(token)
  check_string(url, allow_empty = FALSE)

  job_info <- arc_base_req(
    url,
    token = token,
    query = c(f = "json")
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    yyjsonr::read_json_str()

  # parse the input url
  url_parts <- httr2::url_parse(url)
  path_parts <- strsplit(url_parts$path, "/")[[1]]
  end_idx <- which(path_parts == "jobs") - 1

  # combine back together
  url_parts[["path"]] <- paste0(path_parts[1:end_idx], collapse = "/")

  # build the url without the job ID
  base_url <- httr2::url_build(url_parts)

  # Create and return new instance
  instance <- new_gp_job(base_url, token = token)
  instance$id <- job_info$jobId
  instance
}

#' @export
#' @rdname gp_job
#' @param base_url the URL of the job service (without `/submitJob`)
#' @param params a named list where each element is a scalar character
#' @param token default [arc_token()]. The token to be used with the job.
new_gp_job <- function(
  base_url,
  params = list(),
  token = arc_token()
) {
  check_string(base_url)
  if (!rlang::is_null(token)) {
    obj_check_token(token)
  }
  arc_gp_job$new(base_url, params = params, token = token)
}

#' @export
print.arc_gp_job <- function(x, ...) {
  status <- if (!is.null(x$status)) {
    x$status@status
  } else {
    "not started"
  }
  contents <- c(
    sprintf("<%s>", class(x)[1]),
    sprintf("Job ID: %s", x$id %||% "not initiated"),
    sprintf("Status: %s", status),
    sprintf("Resource: /%s", utils::tail(strsplit(x$base_url, "/")[[1]], 1)),
    "Params:",
    cli::cli_fmt(cli::cli_ul(names(compact(x$params@params))))
  )
  cat(contents, sep = "\n")
  invisible(x)
}

test_that("as_form_params() does not double-encode string scalars", {
  features <- as_esri_features(penguins[1, ])
  params <- as_form_params(list(f = "json", features = features))

  expect_equal(params@params[["f"]], "json")
  expect_equal(params@params[["features"]], features)
})

#' List ArcGIS Enterprise Servers
#'
#' The servers resource lists the ArcGIS Server sites that have been federated with the portal.
#'
#' @inheritParams arc_portal_users
#' @export
#' @examples
#' \dontrun{
#' set_arc_token(auth_user())
#' arc_portal_servers()
#' }
#' @returns a data.frame of servers
arc_portal_servers <- function(
  id = arc_portal_self(token)[["id"]],
  host = arc_host(),
  token = arc_token()
) {
  check_string(id)
  obj_check_token(token)
  resp <- arc_base_req(
    host,
    path = c("sharing", "rest", "portals", id, "servers"),
    query = c(f = "json"),
    token = token
  ) |>
    httr2::req_perform() |>
    httr2::resp_body_string() |>
    yyjsonr::read_json_str() |>
    detect_errors()

  if (!rlang::is_empty(resp[["servers"]])) {
    data_frame(resp[["servers"]])
  } else {
    resp
  }
}

#' Determine the dimensions of a geometry object
#'
#' Given an sfc or sfg object determine what dimensions are represented.
#'
#' @returns
#'
#' `determine_dims()` returns a scalar character of the value `"xy"`, `"xyz"`, or `"xyzm"` depending
#' on what dimensions are represented.
#'
#' `has_m()` and `has_z()` returns a logical scalar of `TRUE` or `FALSE` if the
#' geometry has a Z or M dimension.
#'
#'
#' @param x an object of class `sfc` or `sfg`
#' @export
#'
#' @examples
#'
#' geo <- sf::st_read(system.file("shape/nc.shp", package="sf"), quiet = TRUE)[["geometry"]]
#'
#' determine_dims(geo)
#' has_z(geo)
#' has_m(geo)
determine_dims <- function(x) UseMethod("determine_dims")

#' @export
determine_dims.sfc <- function(x) {
  c("xyzm", "xyz", "xy")[c(has_m(x), has_z(x), TRUE)][1]
}

#' @export
determine_dims.sfg <- function(x) {
  tolower(class(x)[1])
}


# helpers to check for z or m dimension

#' @export
#'@rdname determine_dims
has_m <- function(x) UseMethod("has_m")
#' @export
has_m.sfc <- function(x) !is.null(attr(x, "m_range"))
#' @export
has_m.sfg <- function(x) grepl("M", class(x)[1])

#' @export
#' @rdname determine_dims
has_z <- function(x) UseMethod("has_z")
#' @export
has_z.sfc <- function(x) !is.null(attr(x, "z_range"))
#' @export
has_z.sfg <- function(x) grepl("Z", class(x)[1])

#' General utility functions
#'
#' @details
#'
#' - `compact()` removes any `NULL` list elements
#' - `%||%` is a special pipe operator that returns `b` if `a` is `NULL`
#'
#' @examples
#'
#' # remove null elements
#' compact(list(a = NULL, b = 1))
#'
#' # if NULL return rhs
#' NULL %||% 123
#'
#' # if not NULL return lhs
#' 123 %||% NULL
#'
#' @param .x a list
#' @export
#' @rdname utilities
#' @returns
#' - `compact()` a list
#' - `%||%` the first non-null item or `NULL` if both are `NULL`
compact <- function(.x) Filter(length, .x)

#' @export
#' @rdname utilities
#' @param a an R object
#' @param b an R object
`%||%` <- function(a, b) if (is.null(a)) b else a


#' @param dots a list collected from dots via `rlang::list2(...)`
#' @param call default `rlang::caller_env()`. The caller environment passed to `cli::cli_abort()`
#' @export
#' @rdname utilities
check_dots_named <- function(dots, call = rlang::caller_env()) {
  if (!rlang::is_named2(dots)) {
    cli::cli_abort(
      "All arguments provided to {.arg ...} must be named",
      call = call
    )
  }
  invisible(dots)
}

#' @param x a data.frame
#' @param call default [rlang::caller_call()].
#' @export
#' @rdname utilities
data_frame <- function(x, call = rlang::caller_call()) {
  check_data_frame(x, call = call)
  structure(x, class = c("tbl", "data.frame"))
}

#' Extract matching patterns from a string
#' @noRd
str_extract <- function(string, pattern, perl = TRUE) {
  matches <- regexpr(pattern, string, perl = perl)
  regmatches(string, matches)
}

#' Generate base request
#'
#' This function takes a url and creates a basic httr2 request that
#' adds the user-agent and adds an authorization token to the
#' `X-Esri-Authorization` header.
#'
#' @param url a valid url that is passed to [`httr2::request()`]
#' @param token an object of class `httr2_token` as generated by [`auth_code()`]
#'  or related function
#' @param path a character vector of paths to be appended to url using [`httr2::req_url_path_append()`]
#' @param query a named vector or named list of query parameters to be appended to the url using [`httr2::req_url_query()`]
#' @param error_call the caller environment to be used when propagating errors.
#' @export
#' @examples
#' arc_base_req("https://arcgis.com")
#' @returns an `httr2_request` with the `X-Esri-Authorization` header and `User-Agent` set.
arc_base_req <- function(
  url,
  token = NULL,
  path = NULL,
  query = NULL,
  error_call = rlang::caller_env()
) {
  # set the user agent
  req <- arc_agent(httr2::request(url))

  # handle path if provided
  if (!is.null(path)) {
    # check path is a character vector
    if (!rlang::is_bare_character(path)) {
      cli::cli_abort(
        "{.arg path} must be a character vector",
        call = error_call
      )
    }

    # add path to the url
    req <- httr2::req_url_path_append(
      req,
      rlang::inject(!!path)
    )
  }

  # append a query as well
  if (!is.null(query)) {
    if (!rlang::is_named2(query)) {
      cli::cli_abort(
        "{.arg query} must be a named list",
        call = error_call
      )
    }

    req <- httr2::req_url_query(req, !!!query)
  }

  # if token is not missing, check it
  if (!is.null(token)) {
    # ensure that the token is an httr2_token
    # token <- validate_or_refresh_token(token, error_call)

    # set the auth header
    req <- httr2::req_headers(
      req,
      "X-Esri-Authorization" = paste("Bearer", token[["access_token"]])
    )
  }

  req
}

test_that("agent is arcgisutils", {
  req <- arc_base_req("http://google.com")

  expect_true(
    grepl("^arcgisutils", req$options$useragent)
  )
})

test_that("token is added as header", {

  token_value <- "my-secret-token"

  # create fake token
  token <- httr2::oauth_token(
    token_value,
    arcgis_host = arc_host()
  )

  req <- arc_base_req("http://google.com", token)

  expect_identical(
    paste0("Bearer ", token_value),
    req$headers[["X-Esri-Authorization"]]
  )

})

test_that("path is appended", {
  path_vals <- c("a", "b", "c")

  # create a request
  req <- arc_base_req(
    "http://google.com",
    path = path_vals
  )

  # parse the url of the request
  req_url <- httr2::url_parse(req[["url"]])

  # split
  expect_identical(
    path_vals,
    strsplit(req_url$path, "/")[[1]][-1]
  )
})


test_that("query is appended", {
  query <- list(a = "1", b = "123")

  # create a request
  req <- arc_base_req(
    "http://google.com",
    query = query
  )

  # parse the url of the request
  req_url <- httr2::url_parse(req[["url"]])

  expect_identical(
    query,
    req_url$query
  )
})

#' Portal File Resources
#'
#' The resources endpoint lists all file resources for the organization.
#'
#' @references [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/resources-portal/)
#' @returns a data.frame of resources available to your portal.
#' @family portal
#' @export
#' @inheritParams arc_paginate_req
#' @inheritParams arc_base_req
#' @inheritParams arc_portal_users
#' @examples
#' \dontrun{
#'   set_arc_token(auth_user())
#'   arc_portal_resources()
#' }
#'
arc_portal_resources <- function(
  id = arc_portal_self(token)[["id"]],
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
) {
  check_string(id)
  obj_check_token(token)
  resps <- arc_base_req(
    host,
    path = c("sharing", "rest", "portals", id, "resources"),
    query = c(f = "json"),
    token = token
  ) |>
    arc_paginate_req(
      max_pages = max_pages,
      page_size = page_size,
      .progress = .progress
    )

  results <- lapply(resps, function(.resp) {
    resp <- httr2::resp_body_string(.resp) |>
      yyjsonr::read_json_str() |>
      detect_errors()
    resp[["resources"]]
  })

  res <- data_frame(rbind_results(results))
  res[["created"]] <- from_esri_date(as.numeric(res[["created"]]))
  res
}

#' Portal Users
#'
#' This function lists all users in a portal.
#'
#' @param id the portal ID. By default it fetches the `id` from [`arc_portal_self()`].
#' @param sort_field optional field to sort by. It must be one of "username", "fullname", "created", "lastlogin", "mfaenabled", "level", "role".
#' @param provider optional filter users based on their identity provider. Must be one of "arcgis", "enterprise", "facebook", "google", "apple", or "github".
#' @param sort_order optional order to sort by. It must be one of `"asc"` or `"desc"`.
#' @param role optional role to filter down to. It must be one of "org_admin", "org_publisher", "org_user".
#' @param fullname optional string of the user's fullanme to search for.
#' @param username optional string of the user's user name to search for.
#' @param firstname optional string of the user's first name to search for.
#' @param lastname optional string of the user's last name to search for.
#' @param filter_intersection optional boolean value. If `TRUE` mutliple filters are treated as an `"and"` condition. If `FALSE`, treated as an `"or"`.
#' @inheritParams arc_base_req
#' @inheritParams arc_paginate_req
#' @inheritParams arc_user
#' @references [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/users/)
#' @examples
#' \dontrun{
#' set_arc_token(auth_user())
#' arc_portal_users()
#' }
#' @family portal
#' @returns a data.frame of users.
#' @export
arc_portal_users <- function(
  id = arc_portal_self(token)[["id"]],
  sort_field = NULL,
  provider = NULL,
  sort_order = NULL,
  role = NULL,
  fullname = NULL,
  username = NULL,
  firstname = NULL,
  lastname = NULL,
  filter_intersection = NULL,
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
) {
  obj_check_token(token)
  check_string(id)
  check_bool(filter_intersection, allow_null = TRUE)
  check_string(fullname, allow_null = TRUE, allow_empty = FALSE)
  check_string(firstname, allow_null = TRUE, allow_empty = FALSE)
  check_string(lastname, allow_null = TRUE, allow_empty = FALSE)
  check_string(username, allow_null = TRUE, allow_empty = FALSE)

  if (!is.null(sort_field)) {
    rlang::arg_match(
      sort_field,
      c(
        "username",
        "fullname",
        "created",
        "lastlogin",
        "mfaenabled",
        "level",
        "role"
      )
    )
  }

  if (!is.null(provider)) {
    rlang::arg_match(
      provider,
      c("arcgis", "enterprise", "facebook", "google", "apple", "github")
    )
  }

  if (!is.null(provider)) {
    rlang::arg_match(provider, c("org_admin", "org_publisher", "org_user"))
  }

  if (!is.null(sort_order)) {
    rlang::arg_match(sort_order, c("asc", "desc"))
  }

  params <- compact(
    list(
      sortField = sort_field,
      sortOrder = sort_order,
      provider = provider,
      role = role,
      fullname = fullname,
      username = username,
      lastname - lastname,
      firstname = firstname,
      applyFilterIntersection = filter_intersection
    )
  )
  resps <- arc_base_req(
    host,
    path = c("sharing", "rest", "portals", id, "users"),
    token = token
  ) |>
    httr2::req_body_form(!!!params, f = "json") |>
    arc_paginate_req(
      max_pages = max_pages,
      page_size = page_size,
      .progress = .progress
    )

  results <- lapply(resps, function(.resp) {
    resp <- httr2::resp_body_string(.resp) |>
      yyjsonr::read_json_str() |>
      detect_errors()

    # extract the users from this
    users <- resp[["users"]]

    # unlist these fields in the event that there is a `-1`
    # which causes the dates to become a list
    users[["lastLogin"]] <- unlist(users$lastLogin, recursive = FALSE)
    users[["modified"]] <- unlist(users$modified, recursive = FALSE)
    users
  })

  res <- data_frame(rbind_results(results))
  res[["created"]] <- from_esri_date(as.numeric(res[["created"]]))
  res[["modified"]] <- from_esri_date(as.numeric(res[["modified"]]))
  res[["emailStatusDate"]] <- from_esri_date(as.numeric(res[[
    "emailStatusDate"
  ]]))
  res
}

# Code flow ---------------------------------------------------------------

#' Authorization
#'
#' Authorize your R session to connect to an ArcGIS Portal. See details.
#'
#' @details
#'
#' ArcGIS Online and Enterprise Portals utilize OAuth2 authorization via their REST APIs.
#'
#' - `auth_code()` is the recommend OAuth2 workflow for interactive sessions
#' - `auth_client()` is the recommended OAuth2 workflow for non-interactive sessions
#' - `auth_user()` uses legacy username and password authorization using the `generateToken` endpoint. It is only recommended for legacy systems that do not implement OAuth2.
#' - `auth_binding()` fetches a token from the active portal set by `arcgisbinding`. Uses `arcgisbinding::arc.check_portal()` to extract the authorization token. Recommended if using arcgisbinding.
#'
#' @param client an OAuth 2.0 developer application client ID. By default uses the
#'  environment variable `ARCGIS_CLIENT`.
#' @param secret an OAuth 2.0 developer application secret. By default uses the environment
#'   variable `ARCGIS_SECRET`.
#' @param host default `"https://www.arcgis.com"`. The host of your ArcGIS Portal.
#' @param expiration the duration of the token in minutes.
#' @inheritParams cli::cli_abort
#'
#' @rdname auth
#' @export
#' @examples
#' \dontrun{
#' auth_code()
#' auth_client()
#' auth_user()
#' auth_key()
#' auth_binding()
#' }
#' @returns an `httr2_token`
auth_code <- function(
  client = Sys.getenv("ARCGIS_CLIENT"),
  host = arc_host()
) {
  check_string(client, allow_empty = FALSE)
  check_string(host, allow_empty = FALSE)

  token_url <- paste(host, "sharing", "rest", "oauth2", "token", sep = "/")

  auth_url <- paste(host, "sharing", "rest", "oauth2", "authorize", sep = "/")

  client <- httr2::oauth_client(
    id = client,
    token_url = token_url,
    name = "arcgisutils"
  )

  auth_url <- httr2::oauth_flow_auth_code_url(
    client,
    auth_url = auth_url,
    redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
  )

  # open browser
  utils::browseURL(auth_url)

  {
    code <- readline(prompt = "Enter code: ")
  }

  # use the internal function from httr2 to complete the token
  # transfer
  fx <- utils::getFromNamespace("oauth_client_get_token", "httr2")

  res <- fx(
    client,
    grant_type = "authorization_code",
    code = code,
    redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
  )

  # add host to the token
  res[["arcgis_host"]] <- host
  res
}
# Client auth -------------------------------------------------------------
#' @export
#' @rdname auth
auth_client <- function(
  client = Sys.getenv("ARCGIS_CLIENT"),
  secret = Sys.getenv("ARCGIS_SECRET"),
  host = arc_host(),
  expiration = 120
) {
  check_string(client, allow_empty = FALSE)
  check_string(secret, allow_empty = FALSE)
  check_string(host, allow_empty = FALSE)
  check_number_whole(expiration, min = 5, max = 20160)

  # https://developers.arcgis.com/documentation/mapping-apis-and-services/security/application-credentials/
  token_url <- paste(
    host,
    "sharing",
    "rest",
    "oauth2",
    "token",
    sep = "/"
  )

  cln <- httr2::oauth_client(
    client,
    token_url,
    secret,
    name = "arcgisutils"
  )

  res <- httr2::oauth_flow_client_credentials(
    cln,
    token_params = list(expiration = expiration)
  )

  # add host to the token
  res[["arcgis_host"]] <- host
  res
}


# arcgisbinding -----------------------------------------------------------

#' @export
#' @rdname auth
auth_binding <- function() {
  rlang::check_installed("arcgisbinding")
  tkn <- utils::getFromNamespace("arc.check_portal", "arcgisbinding")()
  res <- httr2::oauth_token(tkn[["token"]])
  # add host to the token
  res[["arcgis_host"]] <- tkn[["url"]]
  # add the username to the token as well
  res[["username"]] <- tkn[["user"]]
  res
}


# legacy auth -------------------------------------------------------------

# uses request ip does not support other forms
# use oauth as the recommended approach
#' @export
#' @rdname auth
#' @param username default `Sys.getenv("ARCGIS_USER")`.
#'  Your username to login. **Do not** hard code this value.
#' @param password default `Sys.getenv("ARCGIS_PASSWORD")`.
#'   Your password to login. **Do not** hard code this value.
auth_user <- function(
  username = Sys.getenv("ARCGIS_USER"),
  password = Sys.getenv("ARCGIS_PASSWORD"),
  host = arc_host(),
  expiration = 60
) {
  check_string(username, allow_empty = FALSE)
  check_string(password, allow_empty = FALSE)
  check_string(host, allow_empty = FALSE)
  check_number_whole(expiration, min = 5, max = 20160)

  if (expiration > 21600) {
    cli::cli_abort("{.arg expiration} cannot be more than 15 days (21600)")
  }

  burl <- paste0(host, "/sharing/rest/generateToken")
  b_req <- httr2::request(burl)

  req <- httr2::req_body_form(
    b_req,
    username = username,
    password = password,
    client = "referer",
    referer = host,
    expiration = expiration,
    f = "json"
  )

  # set the user agent
  req <- arc_agent(req)

  # perform request
  resp <- httr2::req_perform(req)

  # fetch token
  token_raw <- httr2::resp_body_string(resp)

  # parse the response
  token <- RcppSimdJson::fparse(token_raw)

  # detect the errors
  detect_errors(token)

  # return the token
  httr2::oauth_token(
    token[["token"]],
    expires_in = expiration,
    username = username,
    arcgis_host = host,
  )
}


# API Key ----------------------------------------------------------------

#' @param api_key default `Sys.getenv("ARCGIS_API_KEY")`. A character scalar of an ArcGIS Developer API key.
#' @export
#' @rdname auth
auth_key <- function(
  api_key = Sys.getenv("ARCGIS_API_KEY"),
  host = arc_host()
) {
  check_string(
    api_key,
    allow_empty = FALSE,
    allow_na = FALSE,
    allow_null = FALSE
  )

  httr2::oauth_token(
    api_key,
    arcgis_host = host
  )
}

# refreshment mmm tasty --------------------------------------------------------

#' @param token an `httr2_token` as created by `auth_code()` or similar
#' @rdname auth
#' @export
refresh_token <- function(
  token,
  client = Sys.getenv("ARCGIS_CLIENT"),
  host = arc_host()
) {
  # validate the object is a token
  obj_check_token(token)
  check_string(client, allow_empty = FALSE)

  # extract host from token
  host <- token[["arcgis_host"]]

  token_url <- paste(
    host,
    "sharing",
    "rest",
    "oauth2",
    "token",
    sep = "/"
  )

  cln <- httr2::oauth_client(client, token_url, name = "arcgisutils")

  # get the current time
  cur_time <- as.numeric(Sys.time())

  if (is.null(token[["refresh_token"]])) {
    cli::cli_abort(
      "{.arg token} has expired and no {.field refresh_token} available"
    )
  } else if ((cur_time + token[["refresh_token_expires_in"]]) < cur_time) {
    # if it has a refresh check to see if refresh hasn't expired
    cli::cli_abort("Token's {.field refresh_token} has expired.")
  }

  # should be able to refresh, go ahead.
  res <- httr2::oauth_flow_refresh(cln, token[["refresh_token"]])

  # set the host back into the token
  # Note we don't do this for username because only `auth_code()` provides
  # a refresh token
  res[["arcgis_host"]] <- host
  res
}

#' @rdname auth
#' @export
#' @param refresh_threshold default `10`. If token expiry is within this
#'  threshold (in seconds) the token will be refreshed only if a
#'  `refresh_token` is available. Token refreshing is only possible with
#'  `auth_code()` flow.
validate_or_refresh_token <- function(
  token,
  client = Sys.getenv("ARCGIS_CLIENT"),
  host = arc_host(),
  refresh_threshold = 10,
  call = rlang::caller_env()
) {
  # validate the object is a token
  obj_check_token(token, call = call)
  check_string(client, allow_empty = TRUE)
  check_string(host, allow_empty = FALSE)
  check_number_whole(refresh_threshold, min = 0, max = 3600)

  cur_time <- as.numeric(Sys.time())

  # if there isn't a client token value then we can return the token
  if (!nzchar(client)) {
    return(token)
  }

  # provide an error if there is an expired token
  if (token[["expires_at"]] <= cur_time) {
    cli::cli_abort("Provided token has expired", call = call)
  }

  # check if token is expired or expires within threshold
  # the idea being if the token is going to expire soon, refresh it

  if (
    token[["expires_at"]] - refresh_threshold <= cur_time &&
      !is.null(token[["refresh_token"]])
  ) {
    # if it is refresh the token
    token <- refresh_token(client, host, token)
  } else {
    # if not return token
    token
  }
}

#' Determines Portal Host
#'
#' Returns a scalar character indicating the host to make requests to.
#'
#' By default, the host is ArcGIS Online <`https://www.arcgis.com`>. If the
#' environment variable `ARCGIS_HOST` is set, it will be returned.
#'
#' @export
#' @examples
#' arc_host()
#' @returns
#' A scalar character, `"https://www.arcgis.com"` by default.
arc_host <- function() {
  host <- Sys.getenv("ARCGIS_HOST")
  check_string(host, allow_empty = TRUE)
  if (host == "") {
    host <- "https://www.arcgis.com"
  }

  host
}

test_that("fetch_layer_metadata() NULL and empty string", {
  skip()
  furl <- "https://services.arcgisonline.com/arcgis/rest/services/USA_Topo_Maps/MapServer"
  req <- httr2::request(furl)

  expect_equal(
    fetch_layer_metadata(req, NULL),
    fetch_layer_metadata(req, "")
  )
})

test_that("fetch_layer_metadata() known working MapServer", {
  skip()
  furl <- "https://sampleserver6.arcgisonline.com/arcgis/rest/services/Census/MapServer"
  req <- httr2::request(furl)

  # this works with no token provided as a null or nzchar
  expect_equal(
    fetch_layer_metadata(req, NULL),
    fetch_layer_metadata(req, "")
  )
})


test_that("fetch_layer_metadata() works for private content", {
  skip("must be done interactively")
  furl <- "https://services1.arcgis.com/hLJbHVT9ZrDIzK0I/arcgis/rest/services/North%20Carolina%20SIDS%20sample/FeatureServer"
  req <- httr2::request(furl)
  token <- auth_code()

  expect_no_error(fetch_layer_metadata(req, token$access_token))

})

#' Retrieve metadata
#'
#' Utility functions for feature service metadata.
#'
#' @details
#'
#' - `fetch_layer_metadata()` given a request, fetches the metadata by setting
#'    the query parameter `f=json`
#'
#' @param url the url of the item.
#' @param token an `httr2_token` from one of the provided `auth_` functions
#' @param call default [`rlang::caller_env()`]. The calling environment passed to `detect_errors()`.
#' @export
#' @examples
#' # url is broken into parts to fit within 100 characters to avoid CRAN notes
#' url_parts <- c(
#'   "https://services.arcgis.com/P3ePLMYs2RVChkJx/ArcGIS/rest/services",
#'   "/USA_Counties_Generalized_Boundaries/FeatureServer/0"
#' )
#'
#' furl <- paste0(url_parts, collapse = "")
#' meta <- fetch_layer_metadata(furl)
#' head(names(meta))
#' @returns returns a list object
fetch_layer_metadata <- function(
  url,
  token = NULL,
  call = rlang::caller_env()
) {
  req <- arc_base_req(url, token, error_call = call)

  # add f=json to the url for querying
  req <- httr2::req_url_query(req, f = "json")

  # process the request and capture the response string
  resp_string <- httr2::resp_body_string(
    httr2::req_perform(req, error_call = call)
  )

  # process the response string
  meta <- RcppSimdJson::fparse(resp_string)

  # check if any errors occurred
  detect_errors(meta, error_call = call)

  # return the list
  meta
}

#' Detect errors in parsed json response
#'
#' The requests responses from ArcGIS don't return the status code
#' in the response itself but rather from the body in the json.
#' This function checks for the existence of an error. If an error is found,
#' the contents of the error message are bubbled up.
#'
#' @param response for `detect_errors()`, a list typically from `RcppSimdJson::fparse(httr2::resp_body_string(resp))`. For `catch_error()`, the string from `httr2::resp_body_string(resp)`.
#' @param error_call default [`rlang::caller_env()`]. The environment from which
#'  to throw the error from.
#' @returns
#'
#' Nothing. Used for it's side effect. If an error code is encountered in the
#' response an error is thrown with the error code and the error message.
#' @export
#' @family requests
#' @examples
#' \dontrun{
#' response <- list(
#'   error = list(
#'     code = 400L,
#'     message = "Unable to generate token.",
#'     details = "Invalid username or password."
#'   )
#' )
#'
#' detect_errors(response)
#' }
detect_errors <- function(response, error_call = rlang::caller_env()) {
  e_msg <- capture_message(response)

  if (is.null(e_msg)) {
    return(response)
  }

  rlang::abort(
    e_msg,
    call = error_call
  )
}

#' @keywords internal
#' @noRd
report_errors <- function(response, error_call = rlang::caller_env()) {
  e_msg <- capture_message(response)

  if (is.null(e_msg)) {
    return(invisible(NULL))
  }

  rlang::warn(
    e_msg,
    call = error_call
  )
}

#' Capture the message from a parsed error JSON object
#' This will print the error status code, message and details if present
#' The JSON must have already been parsed. It might make more sense to
#' try and parse the JSON in a safe way—e.g. use tryCatch
#' @keywords internal
#' @noRd
capture_message <- function(error) {
  e <- error[["error"]]

  if (!is.null(e)) {
    err_msg <- strwrap(
      paste0("  Error", e$messageCode, ": ", e$message),
      prefix = "    ",
      initial = ""
    )

    full_msg <- c(
      "Status code: ",
      e[["code"]],
      "\n",
      paste0(err_msg, collapse = "\n")
    )

    c(paste0(full_msg, collapse = ""), "i" = e$details)
  } else {
    invisible(NULL)
  }
}

#' Catch error will parse the json for you
#' This is so that it can be done safely
#'
#' @rdname detect_errors
#' @export
catch_error <- function(response, error_call = rlang::caller_env()) {
  if (rlang::is_empty(response)) {
    return(invisible(NULL))
  }
  check_string(response, allow_null = TRUE, allow_na = FALSE, call = error_call)

  rlang::catch_cnd(
    report_errors(RcppSimdJson::fparse(response), error_call = error_call)
  )
}

#' Set user-agent for arcgisutils
#'
#' Override the default user-agent set by httr2 to indicate that a request
#' came from arcgisutils.
#'
#' @param req an httr2 request
#' @return an httr2 request object
#' @export
#' @examples
#' req <- httr2::request("http://example.com")
#' arc_agent(req)
arc_agent <- function(req) {
  ver <- utils::packageVersion("arcgisutils")
  httr2::req_user_agent(req, paste0("arcgisutils v", ver))
}

# Generated by extendr: Do not edit by hand

# nolint start

#
# This file was created with the following call:
#   .Call("wrap__make_arcgisutils_wrappers", use_symbols = TRUE, package_name = "arcgisutils")

#' @usage NULL
#' @useDynLib arcgisutils, .registration = TRUE
NULL

sfg_point_as_point <- function(x, sr) .Call(wrap__sfg_point_as_point, x, sr)

sfg_multipoint_as_multipoint <- function(x, sr) .Call(wrap__sfg_multipoint_as_multipoint, x, sr)

sfg_linestring_as_polyline <- function(x, sr) .Call(wrap__sfg_linestring_as_polyline, x, sr)

sfg_multilinestring_as_polyline <- function(x, sr) .Call(wrap__sfg_multilinestring_as_polyline, x, sr)

sfg_polygon_as_polygon <- function(x, sr) .Call(wrap__sfg_polygon_as_polygon, x, sr)

sfg_multipolygon_as_polygon <- function(x, sr) .Call(wrap__sfg_multipolygon_as_polygon, x, sr)

sf_as_features_2d_list <- function(attrs, geoms, n, sr) .Call(wrap__sf_as_features_2d_list, attrs, geoms, n, sr)

sf_as_features_2d_string <- function(attrs, geoms, n, sr) .Call(wrap__sf_as_features_2d_string, attrs, geoms, n, sr)

sf_as_features_3d_list <- function(attrs, geoms, n, sr) .Call(wrap__sf_as_features_3d_list, attrs, geoms, n, sr)

sf_as_features_3d_string <- function(attrs, geoms, n, sr) .Call(wrap__sf_as_features_3d_string, attrs, geoms, n, sr)

sfc_linestring_features_2d_list <- function(x, sr) .Call(wrap__sfc_linestring_features_2d_list, x, sr)

sfc_linestring_features_2d_string <- function(x, sr) .Call(wrap__sfc_linestring_features_2d_string, x, sr)

sfc_linestring_features_3d_list <- function(x, sr) .Call(wrap__sfc_linestring_features_3d_list, x, sr)

sfc_linestring_features_3d_string <- function(x, sr) .Call(wrap__sfc_linestring_features_3d_string, x, sr)

sfc_linestring_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_linestring_featureset_2d_list, x, sr)

sfc_linestring_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_linestring_featureset_2d_string, x, sr)

sfc_linestring_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_linestring_featureset_3d_list, x, sr)

sfc_linestring_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_linestring_featureset_3d_string, x, sr)

sfc_multilinestring_features_2d_string <- function(x, sr) .Call(wrap__sfc_multilinestring_features_2d_string, x, sr)

sfc_multilinestring_features_2d_list <- function(x, sr) .Call(wrap__sfc_multilinestring_features_2d_list, x, sr)

sfc_multilinestring_features_3d_string <- function(x, sr) .Call(wrap__sfc_multilinestring_features_3d_string, x, sr)

sfc_multilinestring_features_3d_list <- function(x, sr) .Call(wrap__sfc_multilinestring_features_3d_list, x, sr)

sfc_multilinestring_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_multilinestring_featureset_2d_string, x, sr)

sfc_multilinestring_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_multilinestring_featureset_2d_list, x, sr)

sfc_multilinestring_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_multilinestring_featureset_3d_string, x, sr)

sfc_multilinestring_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_multilinestring_featureset_3d_list, x, sr)

sfc_multipoint_features_2d_string <- function(x, sr) .Call(wrap__sfc_multipoint_features_2d_string, x, sr)

sfc_multipoint_features_2d_list <- function(x, sr) .Call(wrap__sfc_multipoint_features_2d_list, x, sr)

sfc_multipoint_features_3d_string <- function(x, sr) .Call(wrap__sfc_multipoint_features_3d_string, x, sr)

sfc_multipoint_features_3d_list <- function(x, sr) .Call(wrap__sfc_multipoint_features_3d_list, x, sr)

sfc_multipoint_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_multipoint_featureset_2d_string, x, sr)

sfc_multipoint_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_multipoint_featureset_2d_list, x, sr)

sfc_multipoint_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_multipoint_featureset_3d_string, x, sr)

sfc_multipoint_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_multipoint_featureset_3d_list, x, sr)

sfc_multipolygon_features_2d_string <- function(x, sr) .Call(wrap__sfc_multipolygon_features_2d_string, x, sr)

sfc_multipolygon_features_3d_string <- function(x, sr) .Call(wrap__sfc_multipolygon_features_3d_string, x, sr)

sfc_multipolygon_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_multipolygon_featureset_2d_string, x, sr)

sfc_multipolygon_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_multipolygon_featureset_3d_string, x, sr)

sfc_multipolygon_features_2d_list <- function(x, sr) .Call(wrap__sfc_multipolygon_features_2d_list, x, sr)

sfc_multipolygon_features_3d_list <- function(x, sr) .Call(wrap__sfc_multipolygon_features_3d_list, x, sr)

sfc_multipolygon_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_multipolygon_featureset_2d_list, x, sr)

sfc_multipolygon_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_multipolygon_featureset_3d_list, x, sr)

sfc_point_features_2d_string <- function(x, sr) .Call(wrap__sfc_point_features_2d_string, x, sr)

sfc_point_features_3d_string <- function(x, sr) .Call(wrap__sfc_point_features_3d_string, x, sr)

sfc_point_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_point_featureset_2d_string, x, sr)

sfc_point_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_point_featureset_3d_string, x, sr)

sfc_point_features_2d_list <- function(x, sr) .Call(wrap__sfc_point_features_2d_list, x, sr)

sfc_point_features_3d_list <- function(x, sr) .Call(wrap__sfc_point_features_3d_list, x, sr)

sfc_point_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_point_featureset_2d_list, x, sr)

sfc_point_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_point_featureset_3d_list, x, sr)

sfc_polygon_features_2d_string <- function(x, sr) .Call(wrap__sfc_polygon_features_2d_string, x, sr)

sfc_polygon_features_3d_string <- function(x, sr) .Call(wrap__sfc_polygon_features_3d_string, x, sr)

sfc_polygon_featureset_2d_string <- function(x, sr) .Call(wrap__sfc_polygon_featureset_2d_string, x, sr)

sfc_polygon_featureset_3d_string <- function(x, sr) .Call(wrap__sfc_polygon_featureset_3d_string, x, sr)

sfc_polygon_features_2d_list <- function(x, sr) .Call(wrap__sfc_polygon_features_2d_list, x, sr)

sfc_polygon_features_3d_list <- function(x, sr) .Call(wrap__sfc_polygon_features_3d_list, x, sr)

sfc_polygon_featureset_2d_list <- function(x, sr) .Call(wrap__sfc_polygon_featureset_2d_list, x, sr)

sfc_polygon_featureset_3d_list <- function(x, sr) .Call(wrap__sfc_polygon_featureset_3d_list, x, sr)

as_featureset_2d_list <- function(attrs, geoms, n, sr) .Call(wrap__as_featureset_2d_list, attrs, geoms, n, sr)

as_featureset_2d_string <- function(attrs, geoms, n, sr) .Call(wrap__as_featureset_2d_string, attrs, geoms, n, sr)

as_featureset_3d_list <- function(attrs, geoms, n, sr, has_z) .Call(wrap__as_featureset_3d_list, attrs, geoms, n, sr, has_z)

as_featureset_3d_string <- function(attrs, geoms, n, sr, has_z) .Call(wrap__as_featureset_3d_string, attrs, geoms, n, sr, has_z)


# nolint end

#' Parse Esri JSON
#'
#' Parses an Esri FeatureSet JSON object into an R object. If there is no
#' geometry present, a data.frame is returned. If there is geometry, an sf object is returned.
#'
#' @param string the raw Esri JSON string.
#' @param ... additional arguments passed to [`RcppSimdJson::fparse`]
#' @inheritParams cli::cli_abort
#'
#' @examples
#'
#' esri_json <- '{
#'     "geometryType": "esriGeometryPolygon",
#'     "spatialReference": {
#'         "wkid": 4326
#'     },
#'     "hasZ": false,
#'     "hasM": false,
#'     "features": [
#'         {
#'             "attributes": {
#'                 "id": 1
#'             },
#'             "geometry": {
#'                 "rings": [
#'                     [
#'                         [0.0, 0.0],
#'                         [1.0, 0.0],
#'                         [1.0, 1.0],
#'                         [0.0, 1.0],
#'                         [0.0, 0.0]
#'                     ]
#'                 ]
#'             }
#'         }
#'     ]
#' }'
#'
#' parse_esri_json(esri_json)
#'
#'@export
#'@returns
#' A data.frame. If geometry is found, returns an sf object.
parse_esri_json <- function(string, ..., call = rlang::caller_env()) {
  # parse the string
  # ensure any json nulls are NAs
  b_parsed <- RcppSimdJson::fparse(
    string,
    empty_object = NA,
    empty_array = NULL,
    single_null = NA,
    ...
  )

  # DEVELOPER NOTE: the rings are not reversed in accordance with
  # the SFA Standard. This doesn't matter to {sf} but it may be useful.
  # Winding can be fixed using `sf::st_sfc(check_winding = TRUE)` at the end

  # extract the geometry features
  fts_raw <- b_parsed[["features"]]

  if (is.null(fts_raw)) {
    report_errors(b_parsed, error_call = call)

    # return an empty data.frame with the appropriate fields
    return(ptype_tbl(b_parsed[["fields"]]))
  }

  # if this is a logical vector of length one we need to abort
  if (is.logical(fts_raw) && length(fts_raw) == 1L) {
    return(data.frame())
  }

  # bind all of them together into a single data frame
  # TODO do call strips class. So any integer64 classes need to be restored
  fields <- do.call(rbind.data.frame, fts_raw[["attributes"]])

  # check if fields are present in the featureset
  # they are optional
  if (!is.null(b_parsed[["fields"]])) {
    # sanitize dates if there are any
    field_types <- b_parsed[[c("fields", "type")]]
    are_dates <- which(field_types == "esriFieldTypeDate")
    # field position might not match must grab by name
    date_cols <- b_parsed[[c("fields", "name")]][are_dates]

    # handle dates if present
    if (length(are_dates) > 0) {
      for (col in date_cols) {
        fields[[col]] <- from_esri_date(fields[[col]])
      }
    }
  }

  # early return if geometry is missing skips parsing
  if (is.null(b_parsed$features$geometry)) {
    return(fields)
  }

  # extract geometry
  # this allows us to modify in place via for loop
  geo_raw <- fts_raw[["geometry"]]

  # identify what it's first element's first element's class is.
  # If it is a list then its a MULTI variant otherwise
  # its a singular linestring or polygon
  # this value is passed to `identify_class`
  list_ele_class <- class(geo_raw[[1]][[1]])

  sfg_class <- switch(
    b_parsed[["geometryType"]],
    esriGeometryPoint = "POINT",
    esriGeometryMultipoint = "MULTIPOINT",
    esriGeometryPolyline = identify_class("LINESTRING", list_ele_class),
    esriGeometryPolygon = identify_class("POLYGON", list_ele_class),
    # Error if unknown geometryType
    cli::cli_abort(
      "Unsupported geometry type {.val {b_parsed[[\"geometryType\"]]}}",
      call = call
    )
  )

  # TODO what about xyz and xyzm?
  obj_classes <- c("XY", sfg_class, "sfg")

  # manually apply the sfg class
  for (i in seq_along(geo_raw)) {
    if (rlang::is_scalar_integer(geo_raw[[i]])) {
      geo_raw[[i]] <- rlang::eval_bare(
        rlang::parse_expr(paste0("sf::st_", tolower(sfg_class), "()"))
      )
    } else if (is.null(geo_raw[[i]][[1]])) {
      # use empty geometry collection to represent NULL geometry
      geo_raw[[i]] <- sf::st_geometrycollection()
      next
    } else if (sfg_class == "POINT") {
      geo_raw[[i]] <- unlist(geo_raw[[i]])
    } else if (sfg_class %in% c("MULTILINESTRING", "MULTIPOINT")) {
      # weird edge case where we get a single matrix instead of a list
      # of matrix we need to put it into a list
      if (sfg_class == "MULTILINESTRING" && is.matrix(geo_raw[[i]])) {
        geo_raw[[i]] <- list(geo_raw[[i]])
      }

      geo_raw[[i]] <- geo_raw[[i]][[1]]
    } else if (sfg_class == "MULTIPOLYGON") {
      # remove spatial reference field from multipolygon
      geo_raw[[i]][["spatialReference"]] <- NULL
    }

    # add class to geometry in place
    class(geo_raw[[i]]) <- obj_classes
  }

  # extract spatial reference
  sr <- b_parsed[["spatialReference"]]

  # find the provided spatial reference
  crs_raw <- if (!all(is.na(sr))) {
    sr[["latestWkid"]] %||% sr[["wkid"]] %||% sr[["wkt"]]
  } else {
    NA
  }

  # create an sf object
  res <- sf::st_sf(
    fields,
    # create an sfc object
    geometry = sf::st_sfc(geo_raw, crs = sf::st_crs(crs_raw))
  )

  if (!is.null(b_parsed$exceededTransferLimit)) {
    attr(res, "exceededTransferLimit") <- b_parsed$exceededTransferLimit
  }
  res
}


# identifies if an esriGeometryPolyline or esriGeometryPolygon
# is a MULTILINESTRING or MULTIPOLYGON
# requires
identify_class <- function(object_type, inner_class) {
  object_type <- toupper(object_type)
  match.arg(object_type, c("POLYGON", "LINESTRING"))
  switch(
    inner_class[1],
    list = paste0("MULTI", object_type),
    object_type
  )
}

test_that("parse_esri_json(): 0 feat. results", {
  jsn <- '{"displayFieldName":"GNIS_NAME","fieldAliases":{"OBJECTID":"OBJECTID","permanent_identifier":"Permanent_Identifier","fdate":"FDate","resolution":"Resolution","gnis_id":"GNIS_ID","gnis_name":"GNIS_Name","lengthkm":"LengthKm","ftype":"FType","fcode":"FCode","visibilityfilter":"VisibilityFilter","nhdplusid":"NHDPlusID","vpuid":"VPUID","Shape_Length":"Shape_Length"},"geometryType":"esriGeometryPolyline","spatialReference":{"wkid":102100,"latestWkid":3857,"vcsWkid":5703,"latestVcsWkid":5703},"fields":[{"name":"OBJECTID","type":"esriFieldTypeOID","alias":"OBJECTID"},{"name":"permanent_identifier","type":"esriFieldTypeString","alias":"Permanent_Identifier","length":40},{"name":"fdate","type":"esriFieldTypeDate","alias":"FDate","length":8},{"name":"resolution","type":"esriFieldTypeInteger","alias":"Resolution"},{"name":"gnis_id","type":"esriFieldTypeString","alias":"GNIS_ID","length":10},{"name":"gnis_name","type":"esriFieldTypeString","alias":"GNIS_Name","length":65},{"name":"lengthkm","type":"esriFieldTypeDouble","alias":"LengthKm"},{"name":"ftype","type":"esriFieldTypeInteger","alias":"FType"},{"name":"fcode","type":"esriFieldTypeInteger","alias":"FCode"},{"name":"visibilityfilter","type":"esriFieldTypeInteger","alias":"VisibilityFilter"},{"name":"nhdplusid","type":"esriFieldTypeDouble","alias":"NHDPlusID"},{"name":"vpuid","type":"esriFieldTypeString","alias":"VPUID","length":8},{"name":"Shape_Length","type":"esriFieldTypeDouble","alias":"Shape_Length"}],"features":[]}'
  res <- parse_esri_json(jsn)
  expect_true(is.data.frame(res))
  expect_true(identical(nrow(res), 0L))

  known_names <- c("OBJECTID", "permanent_identifier", "fdate", "resolution",
    "gnis_id", "gnis_name", "lengthkm", "ftype", "fcode", "visibilityfilter",
    "nhdplusid", "vpuid", "Shape_Length")

  expect_identical(names(res), known_names)

})

#' Create Esri JSON Geometry Objects
#'
#' `as_esri_geometry()` converts an `sfg` object to a EsriJSON Geometry object as a string.
#'
#' See [`as_featureset()`] and [`as_features()`] for converting `sfc` and `sf` objects into EsriJSON.
#'
#' @param x an object of class `sfg`. Must be one of  `"POINT"`, `"MULTIPOINT"`, `"LINESTRING"`, `"MULTILINESTRING"`, `"POLYGON"`, or `"MULTIPOLYGON"`.
#' @param crs the coordinate reference system. It must be interpretable by [`sf::st_crs()`].
#' @inheritParams cli::cli_abort
#' @returns a scalar string
#' @references [API Reference](https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm)
#' @examples
#' library(sf)
#' # POINT
#' # create sfg points
#' xy <- st_point(c(1, 2))
#' xyz <- st_point(c(1, 2, 3))
#' xym <- st_point(c(1, 2, 3), dim = "XYM")
#' xyzm <- st_point(c(1, 2, 3, 4))
#'
#' as_esri_geometry(xy)
#' as_esri_geometry(xyz)
#' as_esri_geometry(xym)
#' as_esri_geometry(xyzm)
#'
#' # MULTIPOINT
#' # vector to create matrix points
#' set.seed(0)
#' x <- rnorm(12)
#'
#' xy <- st_multipoint(matrix(x, ncol = 2))
#' xyz <- st_multipoint(matrix(x, ncol = 3))
#' xym <- st_multipoint(matrix(x, ncol = 3), dim = "XYM")
#' xyzm <- st_multipoint(matrix(x, ncol = 4), dim = "XYM")
#'
#' as_esri_geometry(xy)
#' as_esri_geometry(xyz)
#' as_esri_geometry(xym)
#' as_esri_geometry(xyzm)
#'
#' # LINESTRING
#' xy <- st_linestring(matrix(x, ncol = 2))
#' xyz <- st_linestring(matrix(x, ncol = 3))
#' xym <- st_linestring(matrix(x, ncol = 3), dim = "XYM")
#' xyzm <- st_linestring(matrix(x, ncol = 4), dim = "XYM")
#'
#' as_esri_geometry(xy)
#' as_esri_geometry(xyz)
#' as_esri_geometry(xym)
#' as_esri_geometry(xyzm)
#'
#' # MULTILINESTRING
#' as_esri_geometry(st_multilinestring(list(xy, xy)))
#' as_esri_geometry(st_multilinestring(list(xyz, xyz)))
#' as_esri_geometry(st_multilinestring(list(xym, xym)))
#' as_esri_geometry(st_multilinestring(list(xyzm, xyzm)))
#'
#' # POLYGON
#' coords <- rbind(
#'   c(0, 0, 0, 1),
#'   c(0, 1, 0, 1),
#'   c(1, 1, 1, 1),
#'   c(1, 0, 1, 1),
#'   c(0, 0, 0, 1)
#' )
#'
#' xy <- st_polygon(list(coords[, 1:2]))
#' xyz <- st_polygon(list(coords[, 1:3]))
#' xym <- st_polygon(list(coords[, 1:3]), dim = "XYM")
#' xyzm <- st_polygon(list(coords))
#'
#' as_esri_geometry(xy)
#' as_esri_geometry(xyz)
#' as_esri_geometry(xym)
#' as_esri_geometry(xyzm)
#'
#' # MULTIPOLYGON
#' as_esri_geometry(st_multipolygon(list(xy, xy)))
#' as_esri_geometry(st_multipolygon(list(xyz, xyz)))
#' as_esri_geometry(st_multipolygon(list(xym, xym)))
#' as_esri_geometry(st_multipolygon(list(xyzm, xyzm)))
#' @export
as_esri_geometry <- function(x, crs = NULL, call = rlang::caller_env()) {
  sr <- validate_crs(crs)[[1]] %||% list()

  valid_sfg_classes <- c(
    "POINT",
    "MULTIPOINT",
    "LINESTRING",
    "MULTILINESTRING",
    "POLYGON",
    "MULTIPOLYGON"
  )

  if (!rlang::inherits_any(x, valid_sfg_classes)) {
    cli::cli_abort(
      "{.arg x} must inherit one of the following classes {.cls {valid_sfg_classes}}",
      call = call
    )
  }

  sfg_class <- inherits_which(x, valid_sfg_classes)
  switch(
    sfg_class,
    "POINT" = sfg_point_as_point(x, sr),
    "MULTIPOINT" = sfg_multipoint_as_multipoint(x, sr),
    "LINESTRING" = sfg_linestring_as_polyline(x, sr),
    "MULTILINESTRING" = sfg_multilinestring_as_polyline(x, sr),
    "POLYGON" = sfg_polygon_as_polygon(x, sr),
    "MULTIPOLYGON" = sfg_multipolygon_as_polygon(x, sr),
    cli::cli_abort()
  )
}

# Helper to determine which classes are inherited
inherits_which <- function(x, class) {
  inherited <- vapply(class, inherits, logical(1), x = x)
  class[inherited]
}

#' Discover Authenticated User Metadata
#'
#' Given an authentication token, return a list of user-specfic information such as the user ID, username, available credits, email, groups, last login date and more.
#'
#' @inheritParams arc_item
#' @inheritParams arc_base_req
#' @keywords portal self
#' @export
#' @references [API Reference](https://developers.arcgis.com/rest/users-groups-and-items/self/)
#' @returns a list of the authenticated user's metadata
#' @examples
#' \dontrun{
#' if (interactive()) {
#'   arc_user_self(token = auth_user())
#' }
#' }
arc_user_self <- function(
  host = arc_host(),
  token = arc_token(),
  error_call = rlang::caller_call()
) {
  req <- arc_base_req(
    host,
    token,
    path = "sharing/rest/community/self",
    query = c("f" = "json")
  )

  resp <- httr2::req_perform(
    req,
    error_call = error_call
  )

  res <- httr2::resp_body_string(resp) |>
    RcppSimdJson::fparse() |>
    detect_errors()

  date_fields <- c("created", "modified", "emailStatusDate", "lastLogin")

  for (field in date_fields) {
    res[[field]] <- from_esri_date(res[[field]])
  }

  # make into a tbl
  res[["groups"]] <- data_frame(res[["groups"]])
  # process dates

  for (field in c("modified", "created")) {
    col <- res[["groups"]][[field]]
    res[["groups"]][[field]] <- from_esri_date(col)
  }

  structure(res, class = c("PortalUser", "list"))
}

#' Access the Portal Self Resource
#'
#' The function returns the [`/self`](https://developers.arcgis.com/rest/users-groups-and-items/portal-self.htm) resource from the ArcGIS REST API. The `/self` endpoint
#' returns the view of the portal as seen by the current user, whether anonymous
#' or signed in.
#'
#' @details
#'
#' See the [endpoint documentation](https://developers.arcgis.com/rest/users-groups-and-items/portal-self.htm) for more details.
#'
#' The Portal Self response can vary based on whether it's called by a user, an app, or both.
#'
#' The response includes user and appinfo properties, and the variations in responses are primarily related to these two properties. As the names indicate, the user property includes information about the user making the call, and the appinfo property includes information pertaining to the app that made the call.
#'
#' @returns
#'
#' A named list.
#'
#' @inheritParams arc_base_req
#' @export
#' @family portal
#' @examples
#' \dontrun{
#' set_arc_token(auth_code())
#' self <- arc_self_meta()
#' names(self)
#' }
#' @name self
arc_self_meta <- function(
  token = arc_token(),
  error_call = rlang::current_call()
) {
  lifecycle::deprecate_warn(
    "0.3.4",
    "arc_self_meta()",
    "arc_portal_self()",
    always = TRUE
  )
  arc_portal_self(token, error_call)
}

#' @export
#' @name self
arc_portal_self <- function(
  token = arc_token(),
  error_call = rlang::current_call()
) {
  obj_check_token(token)

  burl <- file.path(
    # use the host from a token if set, otherwise default
    token[["arcgis_host"]] %||% arc_host(),
    "sharing",
    "rest",
    "portals",
    "self",
    fsep = "/"
  )

  b_req <- arc_base_req(burl, token)
  req <- httr2::req_body_form(b_req, f = "json")
  resp <- httr2::req_perform(req, error_call = error_call)

  res <- RcppSimdJson::fparse(
    httr2::resp_body_string(resp)
  )

  res
}

#' Paginate ArcGIS Requests
#'
#' Many API endpoints provide common [pagination properties](https://developers.arcgis.com/rest/users-groups-and-items/common-parameters/#paging-properties). `arc_paginate_request()` automatically applies pagination to an input request.
#'
#' @returns a list of `httr2_response`.
#'
#' @param req an `httr2_request` ideally created with `arc_base_req`
#' @param page_size a scalar integer between 1 and 100 indicating the number of responses per page.
#' @param max_pages the maximum number of pages to fetch. By default fetches all pages.
#' @param .progress default `TRUE`. Whether to display a progress bar for requests.
#'
#' @export
#' @references [API Documentation](https://developers.arcgis.com/rest/users-groups-and-items/common-parameters/#paging-properties)
#' @seealso [arc_base_req()]
arc_paginate_req <- function(
  req,
  page_size = 10,
  max_pages = Inf,
  .progress = TRUE
) {
  if (!inherits(req, "httr2_request")) {
    cli::cli_abort(
      c(
        "{.arg req} must be an {.cls httr2_request}",
        "i" = "Create a new request with {.fn arc_base_req}"
      )
    )
  }
  check_bool(.progress)
  check_number_whole(page_size, 1, 100)
  check_number_whole(max_pages, min = 1, allow_infinite = TRUE)
  httr2::req_perform_iterative(
    httr2::req_url_query(req, num = page_size),
    arc_next_req,
    max_reqs = max_pages,
    on_error = "return",
    progress = .progress
  )
}

# Internal function that is applied to req_perform_iterative
arc_next_req <- function(resp, req) {
  if (httr2::resp_is_error(resp)) {
    return(NULL)
  }

  resp_type <- httr2::resp_content_type(resp)

  if (!resp_type %in% c("text/plain", "application/json")) {
    cli::cli_warn("Response returned {.val {resp_type}} but expected json")
    return(NULL)
  }

  # extract next page
  next_start <- rlang::try_fetch(
    {
      RcppSimdJson::fparse(
        httr2::resp_body_string(resp),
        query = "/nextStart"
      )
    },
    error = function(cnd) {
      cli::cli_warn("Failed to extract next page-ending pagination")
      NULL
    }
  )

  # if we didnt get an integer warn and return NULL
  if (!rlang::is_integerish(next_start)) {
    cli::cli_warn(c(
      "Error extracting next page.",
      "i" = "Expected an integer found {obj_type_friendly(next_start)}"
    ))
    return(NULL)
  }

  # -1 when pagination is over
  if (next_start < 1) {
    return(NULL)
  }

  # next response
  httr2::req_url_query(req, start = next_start)
}

#' Portal Item Types
#'
#' Every portal item has an associated item type. Each of those item types have keywords which cna be used to help narrow down search further.
#'
#' @references [REST API Documentation](https://developers.arcgis.com/rest/users-groups-and-items/items-and-item-types)
#' @export
#' @name portal_types
#' @param item_type a scalar character of the item type. See [`portal_item_types()`] for valid item types.
item_type <- S7::new_class(
  "ItemType",
  package = "arcgisutils",
  properties = list(item_type = S7::class_character),
  validator = function(self) {
    check_string(self@item_type, allow_empty = FALSE)

    if (!tolower(self@item_type) %in% tolower(portal_item_types())) {
      cli::cli_abort(
        "{.arg item_type} is not a valid item type. See {.fn portal_item_types}"
      )
    }
  }
)

#' @export
#' @name portal_types
#' @param keyword a scalar character of the item type keyword. See [`portal_item_keywords()`].
item_keyword <- S7::new_class(
  "ItemKeyword",
  package = "arcgisutils",
  properties = list(keyword = S7::class_character),
  validator = function(self) {
    check_string(self@keyword, allow_empty = FALSE)

    if (!tolower(self@keyword) %in% tolower(portal_item_keywords())) {
      cli::cli_abort(
        c(
          "{.arg keyword} is not a valid keyword.",
          "i" = "See {.url https://developers.arcgis.com/rest/users-groups-and-items/items-and-item-types}"
        )
      )
    }
  }
)

#' @export
#' @name portal_types
portal_item_keywords <- function() {
  c(
    ".NET-Windows Desktop",
    "2D",
    "360",
    "3D",
    "3dd",
    "3DObject",
    "3DTiles",
    "3DTiles Package",
    "3DTiles Service",
    "Add In",
    "Add-in",
    "Administrative Report",
    "Analysis",
    "App",
    "App Builder",
    "AppBuilder",
    "Application",
    "ArcGIS",
    "ArcGIS Desktop",
    "ArcGIS Earth",
    "ArcGIS Explorer",
    "ArcGIS Image Server",
    "ArcGIS Knowledge",
    "ArcGIS Mission Management",
    "ArcGIS Online",
    "ArcGIS Pro",
    "ArcGIS Pro Add In",
    "ArcGIS Professional",
    "ArcGIS Project Template",
    "ArcGIS Server",
    "ArcGIS Workflow Manager",
    "ArcGlobe",
    "ArcMap",
    "ArcPad",
    "ArcPad Project",
    "ArcPad Template",
    "ArcReader",
    "ArcScene",
    "Basemap",
    "BIM",
    "bpk",
    "Building",
    "Business Analyst",
    "C",
    "C#",
    "C++",
    "CAD",
    "City",
    "Classification",
    "Classifier",
    "Classify Feature",
    "Classify Pixel",
    "Code",
    "Code Sample",
    "Collector",
    "Compact Tile Package",
    "Compare",
    "Configurable",
    "Configuration",
    "Connection",
    "Coordinates Type",
    "CRS",
    "CSV",
    "Cube Map",
    "Dashboard",
    "Data",
    "Data Engineering",
    "Data Engineering Model",
    "Data Engineering Workbook",
    "Data interoperability",
    "Data Pipeline",
    "Data Store",
    "Database",
    "Deep Learning",
    "Deep Learning Package",
    "Deep Learning Studio",
    "Deep Learning Studio Project",
    "Desktop",
    "Desktop Add In",
    "Detect Objects",
    "dlpk",
    "Document",
    "Earth",
    "Earth Configuration",
    "eaz",
    "ecd",
    "epk",
    "Esri Classification Schema",
    "Esri Classifier Definition",
    "esriaddin",
    "esriaddinx",
    "esriTasks",
    "Experience",
    "Experience Builder",
    "Experience Builder Widget",
    "Explorer",
    "Explorer Add In",
    "Explorer Document",
    "Explorer Layer",
    "Explorer Map",
    "Explorer Web Map",
    "Export Package",
    "Feature",
    "Feature Access",
    "Feature Collection",
    "Feature Service",
    "Feature Service Template",
    "FeatureCollection",
    "File",
    "Flex",
    "Form",
    "Function Template",
    "Functions",
    "gcpk",
    "GeoBIM",
    "Geocoding",
    "Geodata Service",
    "Geodatabase",
    "Geoenrichment",
    "Geometry",
    "GeometryCollection",
    "GeoPackage",
    "Geoprocessing",
    "Geoprocessing Package",
    "Globe Document",
    "gpk",
    "gpkg",
    "gpkx",
    "Graph",
    "Hosted Service",
    "HTML3D/HTML",
    "Hub",
    "Hub Event",
    "Hub Initiative Template",
    "Hub Project",
    "Image",
    "Image Collection",
    "Image Service",
    "Imagery",
    "Inference",
    "Insights",
    "Insights Data Engineering",
    "Insights Data Engineering Model",
    "Insights Data Engineering Workbook",
    "Insights Model",
    "Insights Page",
    "Insights Theme",
    "Insights Workbook",
    "Insights Workbook Package",
    "Installer",
    "IntegratedMesh",
    "Investigation",
    "iPhone",
    "Java",
    "JavaScript",
    "KML",
    "Knowledge",
    "Knowledge Studio",
    "Knowledge Studio Project",
    "Label",
    "Layer",
    "Layer File",
    "Layer Package",
    "Layout",
    "LayoutFile",
    "Library",
    "Link Chart",
    "Locator",
    "Locator Package",
    "lpk",
    "lyr",
    "lyrx",
    "Mac",
    "map",
    "Map",
    "Map Area",
    "Map Document",
    "Map File",
    "Map Notes Template",
    "Map Package",
    "Map Service",
    "Mapping Site",
    "mapx",
    "Microsoft Excel",
    "Microsoft Powerpoint",
    "Microsoft Word",
    "Mission Command",
    "mmpk",
    "Mobile",
    "Mobile Base Package",
    "Mobile Map Package",
    "Mobile Scene Package",
    "model",
    "Model",
    "mpk",
    "msd",
    "mspk",
    "mxd",
    "Native",
    "nmc",
    "nmf",
    "Notebook",
    "Object Detection",
    "Offline",
    "OGC",
    "OGC Feature Service",
    "OIC",
    "Online Map",
    "OpenData",
    "Operations Dashboard",
    "Oriented Imagery Catalog",
    "Ortho",
    "Ortho Maker",
    "Ortho Mapping Project",
    "Ortho Mapping Template",
    "Orthomosaic",
    "Orthophoto",
    "OS X",
    "Package Widget",
    "pagx",
    "Panorama",
    "Parquet",
    "PDF",
    "Photo",
    "Pixel Classification",
    "Pixel Segmentation",
    "Platform",
    "pmf",
    "Point",
    "PointCloud",
    "Presentation",
    "Presentation File",
    "Pro Presentation File",
    "Pro Report",
    "Pro Report Template",
    "Processing",
    "proConfigX",
    "Project",
    "Project Package",
    "Project Template",
    "Proposals",
    "prsx",
    "Published Map",
    "Python",
    "Qt",
    "Raster",
    "Ready to Use",
    "Ready To Use",
    "Regression",
    "Relational",
    "Report",
    "Report Template",
    "Reporting",
    "Result",
    "rft",
    "Route Layer",
    "rpk",
    "rptt",
    "rptx",
    "Rule Package",
    "Sample",
    "Scenarios",
    "Scene",
    "Scene Document",
    "Scene Package",
    "Scene Service",
    "script",
    "Script",
    "Service",
    "Service Definition",
    "Shapefile",
    "Sharing",
    "Silverlight",
    "slpk",
    "Smart Cities",
    "Snippet",
    "Spherical",
    "spk",
    "Statistical Data Collection",
    "StatusActive",
    "StatusComplete",
    "StatusDraft",
    "StoryMap",
    "StoryMap Theme",
    "Stream Service",
    "Streaming",
    "Street View",
    "Style",
    "Survey123",
    "Survey123 Add In",
    "sxd",
    "Symbology",
    "Symbols",
    "Symbols (either 2d or 3d)",
    "Task Assistant",
    "template",
    "Template",
    "Templates",
    "Text",
    "Tile Package",
    "Tool",
    "Toolbox",
    "tpk",
    "tpkx",
    "Training Sample",
    "Transportation Network",
    "Urban",
    "Urban Project",
    "Vector Tile Package",
    "Vector Tile Service",
    "Virtual Reality",
    "Visio Document",
    "Voxel",
    "VR",
    "VR360",
    "vtpk",
    "Web",
    "Web ADF",
    "Web AppBuilder",
    "Web Application",
    "Web Coverage Service",
    "Web Experience",
    "Web Feature Service",
    "Web Map",
    "Web Map Service",
    "Web Map Tile Service",
    "Web Mapping Application",
    "Web Page",
    "Web Scene",
    "Web Site",
    "Widget",
    "Workbook",
    "WPF",
    "wpk",
    "xForm"
  )
}

#' @export
#' @name portal_types
portal_item_types <- function() {
  sort(c(
    "360 VR Experience",
    "CityEngine Web Scene",
    "Map Area",
    "Pro Map",
    "Web Map",
    "Web Scene",
    "Feature Collection",
    "Feature Collection Template",
    "Feature Service",
    "Geodata Service",
    "Group Layer",
    "Media Layer",
    "Image Service",
    "KML",
    "KML Collection",
    "Map Service",
    "OGCFeatureServer",
    "Oriented Imagery Catalog",
    "Relational Database Connection",
    "3DTilesService",
    "Scene Service",
    "Stream Service",
    "Vector Tile Service",
    "WCS",
    "WFS",
    "WMS",
    "WMTS",
    "Geometry Service",
    "Geocoding Service",
    "Geoprocessing Service",
    "Network Analysis Service",
    "Workflow Manager Service",
    "AppBuilder Extension",
    "AppBuilder Widget Package",
    "Analysis Model",
    "Code Attachment",
    "Dashboard",
    "Data Pipeline",
    "Deep Learning Studio Project",
    "Esri Classification Schema",
    "Excalibur Imagery Project",
    "Experience Builder Widget",
    "Experience Builder Widget Package",
    "Form",
    "GeoBIM Application",
    "GeoBIM Project",
    "Hub Event",
    "Hub Initiative",
    "Hub Initiative Template",
    "Hub Page",
    "Hub Project",
    "Hub Site Application",
    "Insights Workbook",
    "Insights Workbook Package",
    "Insights Model",
    "Insights Page",
    "Insights Theme",
    "Insights Data Engineering Workbook",
    "Insights Data Engineering Model",
    "Investigation",
    "Knowledge Studio Project",
    "Mission",
    "Mobile Application",
    "Notebook",
    "Notebook Code Snippet Library",
    "Native Application",
    "Native Application Installer",
    "Ortho Mapping Project",
    "Ortho Mapping Template",
    "Pro Presentation",
    "Solution",
    "StoryMap",
    "Urban Project",
    "Web AppBuilder Widget",
    "Web Experience",
    "Web Experience Template",
    "Web Mapping Application",
    "Workforce Project",
    "Administrative Report",
    "Apache Parquet",
    "CAD Drawing",
    "Color Set",
    "Content Category Set",
    "CSV",
    "CSV Collection",
    "Document Link",
    "Earth configuration",
    "Esri Classifier Definition",
    "Export Package",
    "File Geodatabase",
    "GeoJson",
    "GeoPackage",
    "GML",
    "Image",
    "iWork Keynote",
    "iWork Numbers",
    "iWork Pages",
    "Microsoft Excel",
    "Microsoft Powerpoint",
    "Microsoft Word",
    "PDF",
    "Report Template",
    "Service Definition",
    "Shapefile",
    "SQLite Geodatabase",
    "Statistical Data Collection",
    "StoryMap Theme",
    "Style",
    "Symbol Set",
    "Visio Document",
    "ArcPad Package",
    "Compact Tile Package",
    "Explorer Map",
    "Globe Document",
    "Layout",
    "Map Document",
    "Map Package",
    "Map Template",
    "Mobile Basemap Package",
    "Mobile Map Package",
    "Mobile Scene Package",
    "Pro Map",
    "Project Package",
    "Project Template",
    "Published Map",
    "Scene Document",
    "Task File",
    "Tile Package",
    "Vector Tile Package",
    "Explorer Layer",
    "Image Collection",
    "Layer",
    "Layer",
    "Layer Package",
    "Pro Report",
    "Scene Package",
    "3DTilesPackage",
    "Desktop Style",
    "ArcGIS Pro Configuration",
    "Deep Learning Package",
    "Geoprocessing Package",
    "Geoprocessing Package (Pro version)",
    "Geoprocessing Sample",
    "Locator Package",
    "Raster function template",
    "Rule Package",
    "Pro Report Template",
    "ArcGIS Pro Add In",
    "Code Sample",
    "Desktop Add In",
    "Desktop Application",
    "Desktop Application Template",
    "Explorer Add In",
    "Survey123 Add In",
    "Workflow Manager Package"
  ))
}

# Item types were fetched from the documentation automatically using:
# library(rvest)
# info <- read_html("https://developers.arcgis.com/rest/users-groups-and-items/items-and-item-types/#web-content") |>
#   html_nodes("table") |>
#   html_table() |>
#   dplyr::bind_rows()
#   dplyr::pull(Type) |>
#   dput()
# type_keywords <- read_html("https://developers.arcgis.com/rest/users-groups-and-items/items-and-item-types/#web-content") |>
#   html_nodes("table tbody tr td:nth-child(3) li") |>
#   html_text() |>
#   unique() |>
#   sort()

#' Esri Field Type Mapping
#'
#' Infers Esri field types from R objects. Use [`as_fields()`] to create a
#' data.frame of valid [Esri Field Types](https://developers.arcgis.com/web-map-specification/objects/field/)
#' from an `sf` object or `data.frame`.
#'
#' @details
#' `r lifecycle::badge("experimental")`
#'
#' - `as_fields()` takes a data frame-like object and infers the Esri field type from it.
#' - `fields_as_pytpe_df()` takes a list with `type` and `name` and creates an empty `data.frame` with the corresponding column names and types.

#' - `get_ptype()` takes a scalar character containing the Esri field type and returns a prototype of the pertinent R type

#'
#' ### Field type mapping:
#'
#' Esri field types are mapped as
#'
#' - `esriFieldTypeSmallInteger`: integer
#' - `esriFieldTypeSingle`: double
#' - `esriFieldTypeGUID`: integer
#' - `esriFieldTypeOID`: integer
#' - `esriFieldTypeInteger`: integer
#' - `esriFieldTypeBigInteger`: double
#' - `esriFieldTypeDouble`: double
#' - `esriFieldTypeString`: character
#' - `esriFieldTypeDate`: date
#'
#' R types are mapped as
#'
#' - `double`: esriFieldTypeDouble
#' - `integer`: esriFieldTypeInteger
#' - `character`: esriFieldTypeString
#' - `date`: esriFieldTypeDate
#' - `raw`: esriFieldTypeBlob
#'
#' @examples
#' inferred <- as_fields(iris)
#' inferred
#'
#' fields_as_ptype_df(inferred)
#'
#' @returns
#'
#' - `fields_as_ptype_df()` takes a `data.frame` with columns `name` and `type` and creates an empty `data.frame` with the corresponding columns and R types
#' - `as_fields()` returns a `data.frame` with columns `name`, `type`, `alias`, `nullable`, and `editable` columns
#'   - This resembles that of the `fields` returned by a FeatureService
#' @export
#' @rdname field_mapping
#' @param .data an object of class `data.frame`.
#' @inheritParams cli::cli_abort
#' @inheritParams rlang::caller_arg
as_fields <- function(
  .data,
  arg = rlang::caller_arg(.data),
  call = rlang::caller_env()
) {
  if (!inherits(.data, "data.frame")) {
    cli::cli_abort(
      "Expected {.cls data.frame} found {.obj_type_friendly {(.data)}}.",
      call = call,
      arg = arg
    )
  }

  if (inherits(.data, "sf")) {
    .data <- sf::st_drop_geometry(.data)
  }

  if (nrow(.data) == 0) {
    empty_fields <- data.frame(
      name = character(),
      type = character(),
      alias = character(),
      nullable = logical(),
      editable = logical()
    )

    return(empty_fields)
  }

  # field mappings
  field_base_types <- vapply(.data, typeof, character(1))
  date_check <- vapply(.data, is_date, logical(1))
  factor_check <- vapply(.data, is.factor, logical(1))

  field_base_types[date_check] <- "date"
  field_base_types[factor_check] <- "factor"

  data.frame(
    name = colnames(.data),
    type = vec_mapping[field_base_types],
    alias = colnames(.data),
    length = ifelse(factor_check, 255L, NA),
    nullable = TRUE,
    editable = TRUE
  )
}

#' @export
#' @rdname field_mapping
infer_esri_type <- function(
  .data,
  arg = rlang::caller_arg(.data),
  call = rlang::caller_env()
) {
  lifecycle::deprecate_soft("0.4.0", "infer_esri_type()", "as_fields()")
  as_fields(.data, arg, call)
}


get_ptype <- function(field_type, n = 1, call = rlang::caller_env()) {
  res <- switch(
    field_type,
    "esriFieldTypeSmallInteger" = integer(n),
    "esriFieldTypeSingle" = double(n),
    "esriFieldTypeGUID" = integer(n),
    "esriFieldTypeGlobalID" = character(n),
    "esriFieldTypeOID" = integer(n),
    "esriFieldTypeInteger" = integer(n),
    "esriFieldTypeBigInteger" = double(n),
    "esriFieldTypeDouble" = double(n),
    "esriFieldTypeString" = character(n),
    "esriFieldTypeDate" = rep(Sys.Date(), n),
    "esriFieldTypeGeometry" = numeric(n),
    "esriFieldTypeBlob" = {
      cli::cli_alert_warning(
        "{.code esriFieldTypeBlob} encountered. If an issue occurs please report it at {.url https://github.com/r-arcgis/arcgisutils/issues/new}"
      )
      vector("list", length = n)
    }
  )

  if (is.null(res)) {
    cli::cli_abort(
      "Column of type {.var {field_type}} cannot be mapped to an R vector",
      call = call
    )
  }

  res
}

#' @param fields a list or data.frame of field types. Requires the fields `type` and `name` to be present.
#' @export
#' @rdname field_mapping
#' @param n the number of rows to create in the prototype table
fields_as_ptype_df <- function(fields, n = 0, call = rlang::caller_env()) {
  if (!rlang::is_list(fields)) {
    cli::cli_abort(
      "{.arg fields} must be a list with fields `type` and `name`",
      call = call
    )
  }

  ftype <- fields[["type"]]
  fname <- fields[["name"]]

  if (rlang::is_null(ftype)) {
    cli::cli_abort("{.field type} is missing from {.arg fields}", call = call)
  }

  if (rlang::is_null(fname)) {
    cli::cli_abort("{.field name} is missing from {.arg fields}", call = call)
  }

  data_frame(
    list2DF(
      lapply(
        rlang::set_names(ftype, fname),
        get_ptype,
        n = n,
        call = call
      )
    )
  )
}

#' @export
#' @rdname field_mapping
ptype_tbl <- function(fields, n = 0, call = rlang::caller_env()) {
  lifecycle::deprecate_soft("0.4.0", "ptype_tbl()", "fields_as_ptype_df()")
  fields_as_ptype_df(fields = fields, n = n, call = call)
}


vec_mapping <- c(
  "double" = "esriFieldTypeDouble",
  "integer" = "esriFieldTypeInteger",
  "character" = "esriFieldTypeString",
  "factor" = "esriFieldTypeString",
  # date will be manually defined as being Date or POSIX
  "date" = "esriFieldTypeDate",
  # FIXME actually should be `blob::blob.`
  "raw" = "esriFieldTypeBlob"
)

test_that("field types are appropriately mapped", {
  fields <- data.frame(
    name = c(
      "ObjectID",
      "Name",
      "Locations",
      "BarrierType",
      "Attr_Minutes",
      "Attr_TimeAt1KPH",
      "Attr_TravelTime",
      "Attr_TruckMinutes",
      "Attr_TruckTravelTime",
      "Attr_WalkTime",
      "Attr_Kilometers",
      "Attr_Miles",
      "Shape_Length",
      "Shape_Area"
    ),
    type = rep(
      c(
        "esriFieldTypeOID",
        "esriFieldTypeString",
        "esriFieldTypeBlob",
        "esriFieldTypeInteger",
        "esriFieldTypeDouble"
      ),
      rep(c(1L, 10L), c(4L, 1L))
    ),
    alias = c(
      "ObjectID",
      "Name",
      "Locations",
      "BarrierType",
      "Attr_Minutes",
      "Attr_TimeAt1KPH",
      "Attr_TravelTime",
      "Attr_TruckMinutes",
      "Attr_TruckTravelTime",
      "Attr_WalkTime",
      "Attr_Kilometers",
      "Attr_Miles",
      "Shape_Length",
      "Shape_Area"
    ),
    length = rep(c(NA, 500L, NA), c(1L, 1L, 12L))
  )

  expected <- structure(
    list(
      ObjectID = integer(0),
      Name = character(0),
      Locations = list(),
      BarrierType = integer(0),
      Attr_Minutes = numeric(0),
      Attr_TimeAt1KPH = numeric(0),
      Attr_TravelTime = numeric(0),
      Attr_TruckMinutes = numeric(0),
      Attr_TruckTravelTime = numeric(0),
      Attr_WalkTime = numeric(0),
      Attr_Kilometers = numeric(0),
      Attr_Miles = numeric(0),
      Shape_Length = numeric(0),
      Shape_Area = numeric(0)
    ),
    class = c("tbl", "data.frame"),
    row.names = c(NA, 0L)
  )

  expect_identical(fields_as_ptype_df(fields), expected)
})

#' Create Esri layer objects
#'
#' These functions are used to generate list objects that can be converted into
#' json objects that are used in REST API requests. Notably they are used for adding
#' R objects as items to a portal.
#'
#' A `featureCollection` defines a layer of features that will be stored on a web map.
#' It consists of an array of `layer`s. The `layer` contains the features
#' (attributes and geometries) as a `featureSet` (see [as_esri_featureset()]) and
#' additional metadata which is stored in the `layerDefinition`object. The
#' `layerDefinition` most importantly documents the fields in the object, the object ID,
#' and additional metadata such as name, title, and display scale.
#'
#' Additional documentation for these json object:
#'
#' - [`layer`](https://developers.arcgis.com/rest/services-reference/enterprise/layer/)
#' - [`layerDefinition`](https://developers.arcgis.com/web-map-specification/objects/layerDefinition/)
#' - [`featureCollection`](https://developers.arcgis.com/web-map-specification/objects/featureCollection/)
#'
#' @export
#' @rdname layer_json
#' @param x an object of class `data.frame`. This can be an `sf` object or `tibble` or
#'   any other subclass of `data.frame`.
#' @param name a scalar character of the name of the layer. Must be unique.
#' @param title A user-friendly string title for the layer that can be used in
#'   a table of contents.
#' @param layer_definition a layer definition list as created by `as_layer_definition()`.
#'   A default is derived from `x` and the `name` object.
#' @param id A number indicating the index position of the layer in the WMS or map service.
#' @param layer_url default `NULL`. A string URL to a service that should be used for all queries
#'   against the layer. Used with hosted tiled map services on ArcGIS Online when
#'   there is an associated feature service that allows for queries.
#' @param legend_url default `NULL`. A string URL to a legend graphic for the layer.
#'   Used with WMS layers. The URL usually contains a GetLegendGraphic request.
#' @param popup_info default `NULL`. A list that can be converted into a
#'   [popupInfo](https://developers.arcgis.com/web-map-specification/objects/popupInfo/)
#'   object defining the pop-up window content for the layer. There is no helper for
#'   popupInfo objects.
#' @inheritParams cli::cli_abort
#'
#' @returns
#'
#' A list object containing the required fields for each respective json type.
#' The results can be converted to json using `jsonify::to_json(x, unbox = TRUE)`
#' @examples
#' ld <- as_layer_definition(iris, "iris", "objectID")
#' l <- as_layer(iris, "iris name", "Iris Title")
#' fc <- as_feature_collection(layers = list(l))
as_layer <- function(
  x,
  name,
  title,
  layer_definition = as_layer_definition(
    x,
    name,
    "object_id",
    infer_esri_type(x)
  ),
  id = NULL,
  layer_url = NULL,
  legend_url = NULL,
  popup_info = NULL,
  call = rlang::caller_env()
) {
  if (!rlang::inherits_any(x, "data.frame")) {
    cli::cli_abort(
      "Expected {.cls data.frame} found {.obj_type_friendly {x}}",
      call = call
    )
  } else if (!rlang::is_scalar_character(name)) {
    cli::cli_abort(
      "{.arg name} must be a scalar character vector.",
      call = call
    )
  } else if (!rlang::is_scalar_character(title)) {
    cli::cli_abort(
      "{.arg title} must be a scalar character vector.",
      call = call
    )
  } else if (!nzchar(name)) {
    cli::cli_abort(
      "{.arg name} must not be empty.",
      call = call
    )
  } else if (!nzchar(title)) {
    cli::cli_abort(
      "{.arg name} must not be empty.",
      call = call
    )
  }

  oid_field <- layer_definition[["objectIdField"]]
  # check that the OID field is present in x if not, create it.
  # if not numeric error
  # if numeric & not integer cast to int
  if (!oid_field %in% colnames(x)) {
    x[[oid_field]] <- 1:nrow(x)
  } else if (!is.numeric(x[[oid_field]])) {
    cli::cli_abort("{.arg x} must have a numeric Object ID field", call = call)
  } else if (!is.integer(x[[oid_field]])) {
    x[[oid_field]] <- as.integer(x[[oid_field]])
  }

  compact(
    list(
      featureSet = as_featureset(x),
      id = id,
      layerDefinition = layer_definition,
      layerUrl = layer_url,
      legendUrl = legend_url,
      name = name,
      title = title,
      popupInfo = popup_info
    )
  )
}


#' @param object_id_field a scalar character vector indicating the name of the
#'   object ID field in the dataset.
#' @param fields a data.frame describing the fields in `x`. These values are inferred
#'  by default via [`infer_esri_type()`].
#' @param display_field default `NULL`. A scalar character containing the name of the field that
#'   best summarizes the feature. Values from this field are used by default as
#'   the titles for pop-up windows.
#' @param has_attachments default `FALSE`.
#' @param drawing_info default `NULL`. See REST documentation in details for more.
#'   There are no helpers or validators for `drawingInfo` objects.
#' @param max_scale default `NULL`. A number representing the maximum scale at
#'   which the layer definition will be applied. The number is the scale's
#'   denominator; thus, a value of 2400 represents a scale of 1/2,400. A value
#'   of 0 indicates that the layer definition will be applied regardless of
#'   how far you zoom in.
#' @param min_scale default `NULL`. A number representing the minimum scale at which the layer definition will be applied.
#' @param templates default `NULL`. See REST documentation in details for more.
#' @param type_id_field default `NULL`. See REST documentation in details for more.
#' @param types An array of type objects available for the dataset.
#'   This is used when the `type_id_field` is populated. NOTE there are no
#'   helper functions to create type objects. Any type list objects must match the
#'   json structure when passed to `jsonify::to_json(x, unbox = TRUE)`.
#' @export
#' @rdname layer_json
as_layer_definition <- function(
  x,
  name,
  object_id_field,
  fields = infer_esri_type(x),
  display_field = NULL,
  drawing_info = NULL,
  has_attachments = FALSE,
  max_scale = 0,
  min_scale = 0,
  templates = NULL,
  type_id_field = NULL,
  types = NULL,
  call = rlang::caller_env()
) {
  if (!rlang::inherits_any(x, "data.frame")) {
    cli::cli_abort(
      "Expected {.cls data.frame} found {.obj_type_friendly {x}}",
      call = call
    )
  } else if (!rlang::is_scalar_character(name)) {
    cli::cli_abort(
      "{.arg name} must be a scalar character vector.",
      call = call
    )
  } else if (!rlang::is_scalar_character(object_id_field)) {
    cli::cli_abort(
      "{.arg object_id_field} must be a scalar character vector.",
      call = call
    )
  } else if (!nzchar(name)) {
    cli::cli_abort(
      "{.arg name} must not be empty.",
      call = call
    )
  } else if (!nzchar(object_id_field)) {
    cli::cli_abort(
      "{.arg object_id_field} must not be empty.",
      call = call
    )
  }

  geo_type <- determine_esri_geo_type(x, call = call)
  # get geo-type. If NULL `Table` else `Feature Layer`
  type <- if (is.null(geo_type)) {
    "Table"
  } else {
    "Feature Layer"
  }

  # check display field
  if (!is.null(display_field) && !(display_field %in% colnames(x))) {
    cli::cli_abort(
      "{.arg display_field} must be a column in {.arg x}",
      call = call
    )
  }

  # check OID / create if needed
  if (!object_id_field %in% fields[["name"]]) {
    oid_fields <- data.frame(
      name = object_id_field,
      type = "esriFieldTypeOID",
      alias = object_id_field,
      length = NA,
      editable = FALSE,
      nullable = FALSE
    )

    fields <- rbind(oid_fields, fields)
  } else {
    oid_position <- which(object_id_field %in% fields[["name"]])
    fields[["type"]][oid_position] <- "esriFieldTypeOID"
    # make sure its not editable or nullable
    fields[["editable"]][oid_position] <- FALSE
    fields[["nullable"]][oid_position] <- FALSE
    # remove length if present
    fields[["length"]][oid_position] <- NA
  }

  layer_def_body <- list(
    name = name,
    displayField = display_field,
    # https://developers.arcgis.com/documentation/common-data-types/drawinginfo.htm
    drawingInfo = drawing_info,
    # https://developers.arcgis.com/documentation/common-data-types/field.htm
    objectIdField = object_id_field,
    geometryType = geo_type,
    fields = fields,
    hasAttachments = FALSE,
    maxScale = max_scale,
    minScale = min_scale,
    # https://developers.arcgis.com/documentation/common-data-types/template.htm
    templates = templates,
    # Must be "Feature Layer" or "Table"
    type = type,
    # field that determine the geometry type
    typeIdField = type_id_field,
    # https://developers.arcgis.com/documentation/common-data-types/type.htm
    types = types,
    # this is not documented anywhere but is required from what i can tell
    extent = as_extent(x)
  )

  compact(layer_def_body)
}


#' @export
#' @rdname layer_json
#'
#' @param layers a list of layers as created by `as_layer()`.
#' @param show_legend default `FALSE`. Logical scalar indicating if this layer
#'   should be shown in the legend in client applications.
as_feature_collection <- function(
  layers = list(),
  show_legend = TRUE,
  call = rlang::caller_env()
) {
  if (!rlang::is_list(layers)) {
    cli::cli_abort(
      c(
        "Invalid {.arg layers} object.",
        "i" = "expected {.cls list}, found {.obj_type_friendly {layers}}"
      ),
      call = call
    )
  }
  c(list(layers = layers), showLegend = show_legend)
}

# Item search fields
#   - title
#   - tags
#   - snippet
#   - description
#   - type
#   - typekeywords

# Groups search fields
#  - id
#  - title
#  - description
#  - snippet
#  - tags
#  - owner

# Filter

# Parameters:

# q
# bbox
# filter
# categories
# categoryFilter
# sortField: (title, created, type, owner, avgrating, numratings, numcomments, numviews, scorecompleteness)
# sortOrder: asc, desc
# countFields: type, access, contentstatus, categories,
# countSize: range 10:200
# exclude: max len is 20. comma separated
# displaySublayers: true / false
# MeanPixelSize: integer
# BandCount: integer
# displayServiceProperties: bool

search_term <- function(term, value, error_call = rlang::caller_call()) {
  check_string(term)
  if (!rlang::has_length(value, 1)) {
    cli::cli_abort("{.arg value} must be a scalar")
  }

  sprintf('%s:"%s"', term, value)
}

search_bool <- function(query, bool = c("AND", "OR", "NOT"), term, value) {
  bool <- rlang::arg_match(bool)
  sprintf("%s %s %s", query, bool, search_term(term, value))
}

search_and <- function(query, term, value) {
  search_bool(query, "AND", term, value)
}

search_or <- function(query, term, value) {
  search_bool(query, "OR", term, value)
}

# Helper to combine many items into one OR term
search_vals_or <- function(term, vals) {
  if (length(vals) == 0) {
    return("")
  }
  if (length(vals) == 1) {
    return(search_term(term, vals[1]))
  }

  first_term <- search_term(term, vals[1])
  remaining_vals <- vals[-1]

  Reduce(
    function(.prev, .next) {
      search_or(.prev, term, .next)
    },
    remaining_vals,
    init = first_term
  )
}

# Helper to combine many items into one AND term
search_vals_and <- function(term, vals) {
  if (length(vals) == 0) {
    return("")
  }
  if (length(vals) == 1) {
    return(search_term(term, vals[1]))
  }

  first_term <- search_term(term, vals[1])
  remaining_vals <- vals[-1]

  Reduce(
    function(.prev, .next) {
      search_and(.prev, term, .next)
    },
    remaining_vals,
    init = first_term
  )
}

search_not <- function(query, term, value) {
  search_bool(query, "NOT", term, value)
}


item_search_terms <- c(
  "id",
  "owner",
  "created",
  "modified",
  "title",
  "type",
  "typekeywords",
  "description",
  "tags",
  "snippet",
  "accessinformation",
  "access",
  "group",
  "numratings",
  "numcomments",
  "avgrating",
  "culture",
  "orgid",
  "categories",
  "contentstatus",
  "classification"
)

search_range <- function(
  param,
  from,
  to = Sys.Date(),
  error_call = rlang::caller_call()
) {
  check_string(param, call = error_call)

  from <- rlang::try_fetch(as.POSIXct(from), error = function(cnd) {
    cli::cli_abort(
      "Unable to coerce {.arg {param}} to a date time",
      call = error_call
    )
  })

  to <- rlang::try_fetch(as.POSIXct(to), error = function(cnd) {
    cli::cli_abort(
      "Unable to coerce {.arg {param}} to a date time",
      call = error_call
    )
  })

  sprintf(
    "%s: [%s TO %s]",
    param,
    format(date_to_ms(from), scientific = FALSE),
    format(date_to_ms(to), scientific = FALSE)
  )
}

#' Search for Portal Items
#'
#' Perform full text search or use parameters to programamatically query your portal for content items.
#'
#' @param query a scalar character for free text search or a valid query string as defined by the REST API.
#' @param filter a scalar character. If provided all other arguments except query are ignored.
#' @param title optional character vector of content item titles.
#' @param tags optional character vector of tags to search for.
#' @param owner optional character vector of owner usernames to search for.
#' @param orgid optional character vector of organization IDs to search for.
#' @param item_type optional character vector of content item types. Validated with [`item_type()`].
#' @param type_keywords optional character vector of content tpye keywords. Validated with [`item_keyword()`].
#' @param created,modified optional length two vector which must be coercible to a date time vector. Converted using [`as.POSIXct()`]. Returns only items within this range.
#' @param description,snippet optional scalar character of text to check for.
#' @param categories optional character vector of up to 8 organization content categories.
#' @param category_filters optional character vector of up to 3 category terms. Items that have matching categories are returned. Exclusive with `categories`.
#' @param sort_field optional character vector of fields to sort by. Can sort by `title`, `created`, `type`, `owner`, `modified`, `avgrating`, `numratings`, `numcomments`, `numviews`, and `scorecompleteness`.
#' @param sort_order optional string. One of either `asc` or `desc` for ascending or descending order respectively.
#' @param count_fields optional character vector of up to 3 fields to count. Must be one of `c("type", "access", "contentstatus", "categories")`.
#' @param count_size optional integer determines the maximum number of field values to count for each counted field in `count_fields`. Maximum of 200.
#' @param display_sublayers default `FALSE`. Returns feature layers inside of feature services.
#' @param filter_logic default `"and"` must be one of `c("and", "or", "not")`. Determines if parameters
#' @param bbox unimplemented.
#' @inheritParams arc_paginate_req
#' @inheritParams arc_base_req
#' @inheritParams arc_item
#' @details
#'`r lifecycle::badge("experimental")`
#'
#' Search is quite nuanced and should be handled with care as you may get unexpected results.
#'
#' - Most arguments are passed as `filter` parameters to the API endpoint.
#' - If multiple values are passed to an argument such as `tags`, the search will use an `"OR"` statement.
#' - When multiple arguments are provided, for example `tags`, `owner`, and `item_type`, the search will use `"AND"` logic—i.e. results shown match the `tags` **and** `owner` **and** `item_type`.
#'   - Note: you can change this to `"OR"` behavior by setting `filter_logic = "or"`
#' - If the `filter` argument is provided, all other arguments except `query` are ignored.
#' @export
#' @references [API Documentation](https://developers.arcgis.com/rest/users-groups-and-items/search)
#' @returns a data.frame.
#' @examplesIf curl::has_internet()
#' crime_items <- search_items(
#'   query = "crime",
#'   item_type = "Feature Service",
#'   max_pages = 1
#' )
#' crime_items
search_items <- function(
  query = NULL,
  filter = NULL,
  title = NULL,
  description = NULL,
  snippet = NULL,
  tags = NULL,
  owner = NULL,
  orgid = NULL,
  item_type = NULL,
  type_keywords = NULL,
  created = NULL,
  modified = NULL,
  categories = NULL,
  category_filters = NULL,
  sort_field = NULL,
  sort_order = NULL,
  count_fields = NULL,
  count_size = NULL,
  display_sublayers = FALSE,
  filter_logic = "and",
  bbox = NULL,
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
) {
  # scalars
  check_bool(display_sublayers)
  check_string(query, allow_null = TRUE)
  check_string(filter, allow_null = TRUE)
  check_string(snippet, allow_null = TRUE)
  check_string(description, allow_null = TRUE)

  # vectors
  check_character(tags, allow_null = TRUE)
  check_character(owner, allow_null = TRUE)
  check_character(orgid, allow_null = TRUE)
  check_character(title, allow_null = TRUE)
  check_character(item_type, allow_null = TRUE)
  check_character(categories, allow_null = TRUE)
  check_character(type_keywords, allow_null = TRUE)

  # check the filter logic variable
  filter_logic <- rlang::arg_match0(
    tolower(filter_logic),
    c("and", "or", "not")
  )

  # make title filter
  if (!is.null(title)) {
    title <- search_vals_or("title", title)
  }

  # make tags filter
  if (!is.null(tags)) {
    tags <- search_vals_or("tag", tags)
  }

  # make owner filter query
  if (!is.null(owner)) {
    owner <- search_vals_or("owner", owner)
  }

  # make org filter query
  if (!is.null(orgid)) {
    orgid <- search_vals_or("orgid", orgid)
  }

  # make item type filter
  if (!is.null(item_type)) {
    # validate the item types
    for (ty in item_type) {
      item_type(ty)
    }
    item_type <- search_vals_or("type", item_type)
  }

  # make typekeyword filter
  if (!is.null(type_keywords)) {
    # validate the item types
    for (ty in type_keywords) {
      item_keyword(ty)
    }

    type_keywords <- search_vals_or("typekeywords", type_keywords)
  }

  # handle created & modified ranges
  if (!is.null(created)) {
    if (!length(created) %in% c(1, 2)) {
      cli::cli_abort("{.arg created} must be a length 2 vector of dates")
    }
    created <- search_range(
      "created",
      created[[1]],
      created[[2]] %||% Sys.time()
    )
  }

  # handle created & modified ranges
  if (!is.null(modified)) {
    if (!length(modified) %in% c(1, 2)) {
      cli::cli_abort("{.arg modified} must be a length 2 vector of dates")
    }

    # we will allow a lenght 1 vector here and let it be the current time
    modified <- search_range(
      "modified",
      modified[[1]],
      modified[[2]] %||% Sys.time()
    )
  }

  if (!is.null(categories)) {
    if (length(categories) > 8) {
      cli::cli_abort("Only up to {.num 8} {.arg categories} are supported")
    }
    categories <- yyjsonr::write_json_str(categories, auto_unbox = TRUE)
  }

  if (!is.null(category_filters)) {
    if (length(category_filters) > 2) {
      cli::cli_abort(
        "Only up to {.num 2} {.arg category_filters} are supported"
      )
    }
    yyjsonr::write_json_str(category_filters, auto_unbox = TRUE)
  }

  if (!is.null(categories) && !is.null(category_filters)) {
    cli::cli_abort(c(
      "Only {.arg categories} or {.arg category_filters} can be set.",
      "i" = "They are mutually exclusive."
    ))
  }

  check_character(sort_field, allow_null = TRUE)

  if (!is.null(sort_field)) {
    sort_field <- rlang::arg_match(
      sort_field,
      c(
        "title",
        "created",
        "type",
        "owner",
        "modified",
        "avgrating",
        "numratings",
        "numcomments",
        "numviews",
        "scorecompleteness"
      ),
      multiple = TRUE
    )
    # sort_field <- yyjsonr::write_json_str(sort_field, auto_unbox = TRUE)
  }

  check_string(sort_order, allow_null = TRUE, allow_empty = FALSE)

  if (!is.null(sort_order)) {
    sort_order <- rlang::arg_match(sort_order, c("asc", "desc"))
  }

  check_character(count_fields, allow_null = TRUE)

  if (!is.null(count_fields)) {
    if (length(count_fields) > 3) {
      cli::cli_abort("Only up to 3 {.arg count_fields} are permitted")
    }
    count_fields <- toString(count_fields)
  }

  check_number_whole(count_size, allow_null = TRUE, min = 10, max = 200)
  if (!is.null(count_size)) {
    count_size <- as.integer(count_size)
  }

  all_params <- compact(
    list(
      title,
      tags,
      owner,
      orgid,
      item_type,
      type_keywords,
      created,
      modified
    )
  )

  query_params <- compact(list(
    sortField = sort_field,
    sortOrder = sort_order,
    countFields = count_fields,
    countSize = count_size,
    displaySublayers = display_sublayers,
    categoryFilters = category_filters,
    categories = categories
  ))

  filter_query <- Reduce(
    \(.prev, .next) sprintf("%s %s %s", .prev, filter_logic, .next),
    all_params
  )

  # when filter is provided, it is always preferred
  if (!is.null(filter)) {
    filter_query <- filter
  }

  req <- arc_base_req(
    host,
    token,
    path = c("sharing", "rest", "search"),
    query = c(q = query, filter = filter_query, f = "json")
  ) |>
    httr2::req_url_query(!!!query_params)

  all_resps <- arc_paginate_req(
    req,
    page_size = page_size,
    max_pages = max_pages,
    .progress = .progress
  )

  results <- lapply(all_resps, function(.resp) {
    res <- httr2::resp_body_string(.resp) |>
      RcppSimdJson::fparse() |>
      detect_errors()

    res[["results"]]
  })

  res <- data_frame(rbind_results(results))
  for (col in c("created", "modified", "lastViewed")) {
    res[[col]] <- from_esri_date(res[[col]])
  }

  res
}

# search_items("", title = "Seven Natural Wonders of the World")

# q <- r"{title:"Seven Natural Wonders of the World" OR description:"Seven Natural Wonders of the World"}"

# search_items(q)

# host = arc_host()
# token = arc_token()

# resp <- arc_base_req(
#   host,
#   token = token,
#   path = c("sharing", "rest", "search"),
#   query = c("f" = "json", q = 'owner:jparry_ANGP OR owner:NGiner_geosaurus')
# ) |>
#   httr2::req_perform()

# search_res <- resp |>
#   httr2::resp_body_string() |>
#   RcppSimdJson::fparse() |>
#   detect_errors()

# str(search_res, 1)
# res <- data_frame(search_res[["results"]])

# res$title

# arc_paginate_req()

test_that("multiplication works", {
  correct <- r"([{"attributes":{"OBJECTID":100.0,"RelativeDepthTypeCode":null}}])"

  created <- as_esri_features(
    data.frame(
      OBJECTID = 100,
      RelativeDepthTypeCode = NA_character_
    )
  )
  expect_identical(created, correct)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_items}
\alias{search_items}
\title{Search for Portal Items}
\usage{
search_items(
  query = NULL,
  filter = NULL,
  title = NULL,
  description = NULL,
  snippet = NULL,
  tags = NULL,
  owner = NULL,
  orgid = NULL,
  item_type = NULL,
  type_keywords = NULL,
  created = NULL,
  modified = NULL,
  categories = NULL,
  category_filters = NULL,
  sort_field = NULL,
  sort_order = NULL,
  count_fields = NULL,
  count_size = NULL,
  display_sublayers = FALSE,
  filter_logic = "and",
  bbox = NULL,
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{query}{a scalar character for free text search or a valid query string as defined by the REST API.}

\item{filter}{a scalar character. If provided all other arguments except query are ignored.}

\item{title}{optional character vector of content item titles.}

\item{description, snippet}{optional scalar character of text to check for.}

\item{tags}{optional character vector of tags to search for.}

\item{owner}{optional character vector of owner usernames to search for.}

\item{orgid}{optional character vector of organization IDs to search for.}

\item{item_type}{optional character vector of content item types. Validated with \code{\link[=item_type]{item_type()}}.}

\item{type_keywords}{optional character vector of content tpye keywords. Validated with \code{\link[=item_keyword]{item_keyword()}}.}

\item{created, modified}{optional length two vector which must be coercible to a date time vector. Converted using \code{\link[=as.POSIXct]{as.POSIXct()}}. Returns only items within this range.}

\item{categories}{optional character vector of up to 8 organization content categories.}

\item{category_filters}{optional character vector of up to 3 category terms. Items that have matching categories are returned. Exclusive with \code{categories}.}

\item{sort_field}{optional character vector of fields to sort by. Can sort by \code{title}, \code{created}, \code{type}, \code{owner}, \code{modified}, \code{avgrating}, \code{numratings}, \code{numcomments}, \code{numviews}, and \code{scorecompleteness}.}

\item{sort_order}{optional string. One of either \code{asc} or \code{desc} for ascending or descending order respectively.}

\item{count_fields}{optional character vector of up to 3 fields to count. Must be one of \code{c("type", "access", "contentstatus", "categories")}.}

\item{count_size}{optional integer determines the maximum number of field values to count for each counted field in \code{count_fields}. Maximum of 200.}

\item{display_sublayers}{default \code{FALSE}. Returns feature layers inside of feature services.}

\item{filter_logic}{default \code{"and"} must be one of \code{c("and", "or", "not")}. Determines if parameters}

\item{bbox}{unimplemented.}

\item{page_size}{a scalar integer between 1 and 100 indicating the number of responses per page.}

\item{max_pages}{the maximum number of pages to fetch. By default fetches all pages.}

\item{.progress}{default \code{TRUE}. Whether to display a progress bar for requests.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}
}
\value{
a data.frame.
}
\description{
Perform full text search or use parameters to programamatically query your portal for content items.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Search is quite nuanced and should be handled with care as you may get unexpected results.
\itemize{
\item Most arguments are passed as \code{filter} parameters to the API endpoint.
\item If multiple values are passed to an argument such as \code{tags}, the search will use an \code{"OR"} statement.
\item When multiple arguments are provided, for example \code{tags}, \code{owner}, and \code{item_type}, the search will use \code{"AND"} logic—i.e. results shown match the \code{tags} \strong{and} \code{owner} \strong{and} \code{item_type}.
\itemize{
\item Note: you can change this to \code{"OR"} behavior by setting \code{filter_logic = "or"}
}
\item If the \code{filter} argument is provided, all other arguments except \code{query} are ignored.
}
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
crime_items <- search_items(
  query = "crime",
  item_type = "Feature Service",
  max_pages = 1
)
crime_items
\dontshow{\}) # examplesIf}
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/search}{API Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-self.R
\name{self}
\alias{self}
\alias{arc_self_meta}
\alias{arc_portal_self}
\title{Access the Portal Self Resource}
\usage{
arc_self_meta(token = arc_token(), error_call = rlang::current_call())

arc_portal_self(token = arc_token(), error_call = rlang::current_call())
}
\arguments{
\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}

\item{error_call}{the caller environment to be used when propagating errors.}
}
\value{
A named list.
}
\description{
The function returns the \href{https://developers.arcgis.com/rest/users-groups-and-items/portal-self.htm}{\verb{/self}} resource from the ArcGIS REST API. The \verb{/self} endpoint
returns the view of the portal as seen by the current user, whether anonymous
or signed in.
}
\details{
See the \href{https://developers.arcgis.com/rest/users-groups-and-items/portal-self.htm}{endpoint documentation} for more details.

The Portal Self response can vary based on whether it's called by a user, an app, or both.

The response includes user and appinfo properties, and the variations in responses are primarily related to these two properties. As the names indicate, the user property includes information about the user making the call, and the appinfo property includes information pertaining to the app that made the call.
}
\examples{
\dontrun{
set_arc_token(auth_code())
self <- arc_self_meta()
names(self)
}
}
\seealso{
Other portal: 
\code{\link{arc_portal_resources}()},
\code{\link{arc_portal_urls}()},
\code{\link{arc_portal_users}()}
}
\concept{portal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geoprocessing-class.R
\name{arc_form_params}
\alias{arc_form_params}
\alias{as_form_params}
\title{Form request parameters}
\usage{
arc_form_params(params = list())

as_form_params(x)
}
\arguments{
\item{params}{a named list with scalar character elements}

\item{x}{for \code{as_form_params()}, a named list to convert to form parameters}
}
\value{
an object of class \code{arc_form_params}
}
\description{
ArcGIS endpoints make extensive use of form encoded data for the body
of http requests. Form requests require that each element has a name
and is encoded as a single string—often as json.
}
\details{
The \code{arc_form_params} class provides validation of form body parameters
ensuring that each element is a scalar string. It uses a named list
internally to store the parameters.

The helper function \code{as_form_params()} converts a named list to form
parameters by automatically JSON-encoding each element using
\code{yyjsonr::write_json_str()} with \code{auto_unbox = TRUE}.
}
\examples{
arc_form_params(
  list(f = "json", outFields = "*", where = "1 = 1")
)
}
\seealso{
Other geoprocessing: 
\code{\link{arc_gp_job}},
\code{\link{arc_job_status}()},
\code{\link{gp_job_from_url}()},
\code{\link{gp_params}}

Other geoprocessing: 
\code{\link{arc_gp_job}},
\code{\link{arc_job_status}()},
\code{\link{gp_job_from_url}()},
\code{\link{gp_params}}
}
\concept{geoprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharing.R
\name{arc_portal_urls}
\alias{arc_portal_urls}
\title{Organization's URLs}
\usage{
arc_portal_urls(host = arc_host(), token = arc_token())
}
\arguments{
\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}
}
\value{
a data.frame
}
\description{
Returns the URLs of an organizations services.
}
\details{
See \href{https://developers.arcgis.com/rest/users-groups-and-items/urls/}{API Reference} for more information.
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
arc_portal_urls()
\dontshow{\}) # examplesIf}
}
\seealso{
Other portal: 
\code{\link{arc_portal_resources}()},
\code{\link{arc_portal_users}()},
\code{\link{self}}
}
\concept{portal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-users.R
\name{arc_portal_users}
\alias{arc_portal_users}
\title{Portal Users}
\usage{
arc_portal_users(
  id = arc_portal_self(token)[["id"]],
  sort_field = NULL,
  provider = NULL,
  sort_order = NULL,
  role = NULL,
  fullname = NULL,
  username = NULL,
  firstname = NULL,
  lastname = NULL,
  filter_intersection = NULL,
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{id}{the portal ID. By default it fetches the \code{id} from \code{\link[=arc_portal_self]{arc_portal_self()}}.}

\item{sort_field}{optional field to sort by. It must be one of "username", "fullname", "created", "lastlogin", "mfaenabled", "level", "role".}

\item{provider}{optional filter users based on their identity provider. Must be one of "arcgis", "enterprise", "facebook", "google", "apple", or "github".}

\item{sort_order}{optional order to sort by. It must be one of \code{"asc"} or \code{"desc"}.}

\item{role}{optional role to filter down to. It must be one of "org_admin", "org_publisher", "org_user".}

\item{fullname}{optional string of the user's fullanme to search for.}

\item{username}{optional string of the user's user name to search for.}

\item{firstname}{optional string of the user's first name to search for.}

\item{lastname}{optional string of the user's last name to search for.}

\item{filter_intersection}{optional boolean value. If \code{TRUE} mutliple filters are treated as an \code{"and"} condition. If \code{FALSE}, treated as an \code{"or"}.}

\item{page_size}{a scalar integer between 1 and 100 indicating the number of responses per page.}

\item{max_pages}{the maximum number of pages to fetch. By default fetches all pages.}

\item{.progress}{default \code{TRUE}. Whether to display a progress bar for requests.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}
}
\value{
a data.frame of users.
}
\description{
This function lists all users in a portal.
}
\examples{
\dontrun{
set_arc_token(auth_user())
arc_portal_users()
}
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/users/}{API Reference}
}
\seealso{
Other portal: 
\code{\link{arc_portal_resources}()},
\code{\link{arc_portal_urls}()},
\code{\link{self}}
}
\concept{portal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-crs.R
\name{validate_crs}
\alias{validate_crs}
\title{Validate CRS object}
\usage{
validate_crs(crs, arg = rlang::caller_arg(crs), call = rlang::caller_env())
}
\arguments{
\item{crs}{a representation of a coordinate reference system.}

\item{arg}{An argument name in the current function.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
Returns a list of length 1 with an element named \code{spatialReference} which is itself
a named list.

If the provided CRS returns a valid well-known ID (WKID) \code{spatialReference} contains
a named element called \code{wkid} which is the integer value of the WKID. If the WKID
is not known but the CRS returned is a valid well-known text representation the \code{wkid} field
is \code{NA} and another field \code{wkt} contains the valid wkt.
}
\description{
Takes a representation of a CRS and ensures that it is a valid one. The CRS
is validated using \code{sf::st_crs()} if it cannot be validated, a null CRS is returned.
}
\details{
See \code{\link[sf:st_crs]{sf::st_crs()}} for more details on valid representations.
}
\examples{

# using epsg code integer or string representation
validate_crs(3857)
validate_crs("EPSG:4326")

# using a custom proj4 string
proj4string <- "+proj=longlat +datum=WGS84 +no_defs"

crs <- validate_crs(proj4string)

# using wkt2 (from above result)
crs <- validate_crs(crs$spatialReference$wkt)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-url.R
\name{url}
\alias{url}
\alias{arc_url_parse}
\alias{arc_url_type}
\alias{is_url}
\title{Parse an ArcGIS service or content URL into its components}
\usage{
arc_url_parse(url, base_url = NULL, error_call = rlang::caller_call())

arc_url_type(url, error_call = rlang::caller_call())

is_url(url, error_call = rlang::caller_call())
}
\arguments{
\item{url}{A string containing the URL to parse.}

\item{base_url}{Use this as a parent, if \code{url} is a relative URL.}

\item{error_call}{the caller environment to be used when propagating errors.}
}
\value{
A named list with the following components: scheme, hostname, username, password, port, path, query, fragment, url, type, and layer.
}
\description{
\code{\link[=arc_url_parse]{arc_url_parse()}} uses \code{\link[httr2:url_parse]{httr2::url_parse()}} to parse URL components and
combine the components with a service or content URL \code{type} and a \code{layer}
number if applicable. A \code{layer} component is only included if the \code{type} is \code{"MapServer"} or
\code{"FeatureServer"} and the URL includes a trailing digit. A full \code{url} value is
also included in the returned list. The \code{url}, \code{type}, and \code{layer} components
are not part of the \code{httr2_url} class object returned by
\code{\link[httr2:url_parse]{httr2::url_parse()}}.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
arc_url_parse(
  "https://services.arcgisonline.com/arcgis/rest/services/USA_Topo_Maps/MapServer/0"
)
arc_url_parse(
  "https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer"
)
arc_url_parse(
  "https://services.arcgisonline.com/arcgis/rest/services/WorldElevation3D/Terrain3D/ImageServer"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature-collection.R
\name{as_layer}
\alias{as_layer}
\alias{as_layer_definition}
\alias{as_feature_collection}
\title{Create Esri layer objects}
\usage{
as_layer(
  x,
  name,
  title,
  layer_definition = as_layer_definition(x, name, "object_id", infer_esri_type(x)),
  id = NULL,
  layer_url = NULL,
  legend_url = NULL,
  popup_info = NULL,
  call = rlang::caller_env()
)

as_layer_definition(
  x,
  name,
  object_id_field,
  fields = infer_esri_type(x),
  display_field = NULL,
  drawing_info = NULL,
  has_attachments = FALSE,
  max_scale = 0,
  min_scale = 0,
  templates = NULL,
  type_id_field = NULL,
  types = NULL,
  call = rlang::caller_env()
)

as_feature_collection(
  layers = list(),
  show_legend = TRUE,
  call = rlang::caller_env()
)
}
\arguments{
\item{x}{an object of class \code{data.frame}. This can be an \code{sf} object or \code{tibble} or
any other subclass of \code{data.frame}.}

\item{name}{a scalar character of the name of the layer. Must be unique.}

\item{title}{A user-friendly string title for the layer that can be used in
a table of contents.}

\item{layer_definition}{a layer definition list as created by \code{as_layer_definition()}.
A default is derived from \code{x} and the \code{name} object.}

\item{id}{A number indicating the index position of the layer in the WMS or map service.}

\item{layer_url}{default \code{NULL}. A string URL to a service that should be used for all queries
against the layer. Used with hosted tiled map services on ArcGIS Online when
there is an associated feature service that allows for queries.}

\item{legend_url}{default \code{NULL}. A string URL to a legend graphic for the layer.
Used with WMS layers. The URL usually contains a GetLegendGraphic request.}

\item{popup_info}{default \code{NULL}. A list that can be converted into a
\href{https://developers.arcgis.com/web-map-specification/objects/popupInfo/}{popupInfo}
object defining the pop-up window content for the layer. There is no helper for
popupInfo objects.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}

\item{object_id_field}{a scalar character vector indicating the name of the
object ID field in the dataset.}

\item{fields}{a data.frame describing the fields in \code{x}. These values are inferred
by default via \code{\link[=infer_esri_type]{infer_esri_type()}}.}

\item{display_field}{default \code{NULL}. A scalar character containing the name of the field that
best summarizes the feature. Values from this field are used by default as
the titles for pop-up windows.}

\item{drawing_info}{default \code{NULL}. See REST documentation in details for more.
There are no helpers or validators for \code{drawingInfo} objects.}

\item{has_attachments}{default \code{FALSE}.}

\item{max_scale}{default \code{NULL}. A number representing the maximum scale at
which the layer definition will be applied. The number is the scale's
denominator; thus, a value of 2400 represents a scale of 1/2,400. A value
of 0 indicates that the layer definition will be applied regardless of
how far you zoom in.}

\item{min_scale}{default \code{NULL}. A number representing the minimum scale at which the layer definition will be applied.}

\item{templates}{default \code{NULL}. See REST documentation in details for more.}

\item{type_id_field}{default \code{NULL}. See REST documentation in details for more.}

\item{types}{An array of type objects available for the dataset.
This is used when the \code{type_id_field} is populated. NOTE there are no
helper functions to create type objects. Any type list objects must match the
json structure when passed to \code{jsonify::to_json(x, unbox = TRUE)}.}

\item{layers}{a list of layers as created by \code{as_layer()}.}

\item{show_legend}{default \code{FALSE}. Logical scalar indicating if this layer
should be shown in the legend in client applications.}
}
\value{
A list object containing the required fields for each respective json type.
The results can be converted to json using \code{jsonify::to_json(x, unbox = TRUE)}
}
\description{
These functions are used to generate list objects that can be converted into
json objects that are used in REST API requests. Notably they are used for adding
R objects as items to a portal.
}
\details{
A \code{featureCollection} defines a layer of features that will be stored on a web map.
It consists of an array of \code{layer}s. The \code{layer} contains the features
(attributes and geometries) as a \code{featureSet} (see \code{\link[=as_esri_featureset]{as_esri_featureset()}}) and
additional metadata which is stored in the \code{layerDefinition}object. The
\code{layerDefinition} most importantly documents the fields in the object, the object ID,
and additional metadata such as name, title, and display scale.

Additional documentation for these json object:
\itemize{
\item \href{https://developers.arcgis.com/rest/services-reference/enterprise/layer/}{\code{layer}}
\item \href{https://developers.arcgis.com/web-map-specification/objects/layerDefinition/}{\code{layerDefinition}}
\item \href{https://developers.arcgis.com/web-map-specification/objects/featureCollection/}{\code{featureCollection}}
}
}
\examples{
ld <- as_layer_definition(iris, "iris", "objectID")
l <- as_layer(iris, "iris name", "Iris Title")
fc <- as_feature_collection(layers = list(l))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/esri-geometry.R
\name{as_esri_geometry}
\alias{as_esri_geometry}
\title{Create Esri JSON Geometry Objects}
\usage{
as_esri_geometry(x, crs = NULL, call = rlang::caller_env())
}
\arguments{
\item{x}{an object of class \code{sfg}. Must be one of  \code{"POINT"}, \code{"MULTIPOINT"}, \code{"LINESTRING"}, \code{"MULTILINESTRING"}, \code{"POLYGON"}, or \code{"MULTIPOLYGON"}.}

\item{crs}{the coordinate reference system. It must be interpretable by \code{\link[sf:st_crs]{sf::st_crs()}}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
a scalar string
}
\description{
\code{as_esri_geometry()} converts an \code{sfg} object to a EsriJSON Geometry object as a string.
}
\details{
See \code{\link[=as_featureset]{as_featureset()}} and \code{\link[=as_features]{as_features()}} for converting \code{sfc} and \code{sf} objects into EsriJSON.
}
\examples{
library(sf)
# POINT
# create sfg points
xy <- st_point(c(1, 2))
xyz <- st_point(c(1, 2, 3))
xym <- st_point(c(1, 2, 3), dim = "XYM")
xyzm <- st_point(c(1, 2, 3, 4))

as_esri_geometry(xy)
as_esri_geometry(xyz)
as_esri_geometry(xym)
as_esri_geometry(xyzm)

# MULTIPOINT
# vector to create matrix points
set.seed(0)
x <- rnorm(12)

xy <- st_multipoint(matrix(x, ncol = 2))
xyz <- st_multipoint(matrix(x, ncol = 3))
xym <- st_multipoint(matrix(x, ncol = 3), dim = "XYM")
xyzm <- st_multipoint(matrix(x, ncol = 4), dim = "XYM")

as_esri_geometry(xy)
as_esri_geometry(xyz)
as_esri_geometry(xym)
as_esri_geometry(xyzm)

# LINESTRING
xy <- st_linestring(matrix(x, ncol = 2))
xyz <- st_linestring(matrix(x, ncol = 3))
xym <- st_linestring(matrix(x, ncol = 3), dim = "XYM")
xyzm <- st_linestring(matrix(x, ncol = 4), dim = "XYM")

as_esri_geometry(xy)
as_esri_geometry(xyz)
as_esri_geometry(xym)
as_esri_geometry(xyzm)

# MULTILINESTRING
as_esri_geometry(st_multilinestring(list(xy, xy)))
as_esri_geometry(st_multilinestring(list(xyz, xyz)))
as_esri_geometry(st_multilinestring(list(xym, xym)))
as_esri_geometry(st_multilinestring(list(xyzm, xyzm)))

# POLYGON
coords <- rbind(
  c(0, 0, 0, 1),
  c(0, 1, 0, 1),
  c(1, 1, 1, 1),
  c(1, 0, 1, 1),
  c(0, 0, 0, 1)
)

xy <- st_polygon(list(coords[, 1:2]))
xyz <- st_polygon(list(coords[, 1:3]))
xym <- st_polygon(list(coords[, 1:3]), dim = "XYM")
xyzm <- st_polygon(list(coords))

as_esri_geometry(xy)
as_esri_geometry(xyz)
as_esri_geometry(xym)
as_esri_geometry(xyzm)

# MULTIPOLYGON
as_esri_geometry(st_multipolygon(list(xy, xy)))
as_esri_geometry(st_multipolygon(list(xyz, xyz)))
as_esri_geometry(st_multipolygon(list(xym, xym)))
as_esri_geometry(st_multipolygon(list(xyzm, xyzm)))
}
\references{
\href{https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm}{API Reference}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharing.R
\name{arc_item_data}
\alias{arc_item_data}
\title{Download an Item's Data}
\usage{
arc_item_data(item, host = arc_host(), token = arc_token())
}
\arguments{
\item{item}{the item ID or the result of \code{arc_item()}.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}
}
\value{
a raw vector containing the bytes of the data associated with the item. If the response is \code{application/json} then the json string is returned without parsing.
}
\description{
Download the data backing a portal item. This function always returns
a raw vector as the type of the data that is downloaded cannot always be known.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
arc_item_data("9df5e769bfe8412b8de36a2e618c7672")
\dontshow{\}) # examplesIf}
}
\seealso{
Other portal item: 
\code{\link{arc_item}()}
}
\concept{portal item}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geoprocessing-class.R
\name{arc_gp_job}
\alias{arc_gp_job}
\alias{new_gp_job}
\title{Create a Geoprocessing Service Job}
\usage{
new_gp_job(base_url, params = list(), token = arc_token())
}
\arguments{
\item{base_url}{the URL of the job service (without \verb{/submitJob})}

\item{params}{a named list where each element is a scalar character}

\item{token}{default \code{\link[=arc_token]{arc_token()}}. The token to be used with the job.}
}
\value{
An object of class \code{arc_gp_job}.
}
\description{
The \code{arc_gp_job} class is used to interact with Geoprocessing Services in
ArcGIS Online and Enterprise.
}
\details{
The \code{arc_gp_job} uses S7 classes for the job request parameters and job status
via \code{\link[=arc_form_params]{arc_form_params()}} and \code{\link[=arc_job_status]{arc_job_status()}} respectively. Importantly,
\code{\link[=arc_form_params]{arc_form_params()}} ensures that parameters provided to a geoprocessing
service are all character scalars as required by the form body.
}
\section{Associated Functions}{

\describe{
\item{\code{from_url(url, token = arc_token())}}{Create a GP Job object from an existing job URL}
}
}

\examples{
url <- paste0(
  "https://logistics.arcgis.com/arcgis/",
  "rest/services/World/ServiceAreas/",
  "GPServer/GenerateServiceAreas"
)
job <- new_gp_job(url, list(f = "json"))
job

# extract params S7 class
params <- job$params
params

# view underlying list
params@params
}
\seealso{
Other geoprocessing: 
\code{\link{arc_form_params}()},
\code{\link{arc_job_status}()},
\code{\link{gp_job_from_url}()},
\code{\link{gp_params}}
}
\concept{geoprocessing}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{base_url}}{the URL of the job service (without \verb{/submitJob})}

\item{\code{id}}{the ID of the started job. \code{NULL} \code{self$start()} has not been called.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{params}}{returns an S7 object of class \code{arc_form_params} (see \code{\link[=arc_form_params]{arc_form_params()}}) the list can be accessed via \code{self$params@params}.}

\item{\code{status}}{returns the status of the geoprocessing job as an S7 object of class \code{gp_job_status} (see \code{\link[=arc_job_status]{arc_job_status()}}) by querying the \verb{/jobs/\{job-id\}} endpoint.}

\item{\code{results}}{returns the current results of the job by querying the \verb{/jobs/\{job-id\}/results} endpoint.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-arc_gp_job-new}{\code{arc_gp_job$new()}}
\item \href{#method-arc_gp_job-start}{\code{arc_gp_job$start()}}
\item \href{#method-arc_gp_job-cancel}{\code{arc_gp_job$cancel()}}
\item \href{#method-arc_gp_job-await}{\code{arc_gp_job$await()}}
\item \href{#method-arc_gp_job-messages}{\code{arc_gp_job$messages()}}
\item \href{#method-arc_gp_job-clone}{\code{arc_gp_job$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-new"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$new(
  base_url,
  params = list(),
  result_fn = NULL,
  token = arc_token(),
  error_call = rlang::caller_call()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{base_url}}{the URL of the job service (without \verb{/submitJob})}

\item{\code{params}}{a named list where each element is a scalar character}

\item{\code{result_fn}}{Default \code{NULL}. An optional function to apply to the results JSON. By default parses results using \code{RcppSimdJson::fparse()}.}

\item{\code{token}}{default \code{\link[=arc_token]{arc_token()}}. The token to be used with the job.}

\item{\code{error_call}}{default \code{rlang::caller_call()} the calling environment.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-start"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-start}{}}}
\subsection{Method \code{start()}}{
Starts the job by calling the \verb{/submitJob} endpoint. This also sets the public field \code{id}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$start()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-cancel"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-cancel}{}}}
\subsection{Method \code{cancel()}}{
Cancels a job by calling the \verb{/cancel} endpoint.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$cancel()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-await"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-await}{}}}
\subsection{Method \code{await()}}{
Waits for job completion and returns results.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$await(interval = 0.1, verbose = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{interval}}{polling interval in seconds (default 0.1)}

\item{\code{verbose}}{whether to print status messages (default FALSE)}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-messages"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-messages}{}}}
\subsection{Method \code{messages()}}{
Fetches Geoprocessing Messages
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$messages()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-arc_gp_job-clone"></a>}}
\if{latex}{\out{\hypertarget{method-arc_gp_job-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{arc_gp_job$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharing.R
\name{arc_group}
\alias{arc_group}
\title{Fetch Group Information}
\usage{
arc_group(group_id, host = arc_host(), token = arc_token())
}
\arguments{
\item{group_id}{the unique group identifier. A scalar character.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}
}
\value{
a list with group metadata
}
\description{
Fetches metadata about a group based on a provided \code{group_id}.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
arc_group("2f0ec8cb03574128bd673cefab106f39")
\dontshow{\}) # examplesIf}
}
\seealso{
Other portal organization: 
\code{\link{arc_user}()}
}
\concept{portal organization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/esri-field-mapping.R
\name{as_fields}
\alias{as_fields}
\alias{infer_esri_type}
\alias{fields_as_ptype_df}
\alias{ptype_tbl}
\title{Esri Field Type Mapping}
\usage{
as_fields(.data, arg = rlang::caller_arg(.data), call = rlang::caller_env())

infer_esri_type(
  .data,
  arg = rlang::caller_arg(.data),
  call = rlang::caller_env()
)

fields_as_ptype_df(fields, n = 0, call = rlang::caller_env())

ptype_tbl(fields, n = 0, call = rlang::caller_env())
}
\arguments{
\item{.data}{an object of class \code{data.frame}.}

\item{arg}{An argument name in the current function.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}

\item{fields}{a list or data.frame of field types. Requires the fields \code{type} and \code{name} to be present.}

\item{n}{the number of rows to create in the prototype table}
}
\value{
\itemize{
\item \code{fields_as_ptype_df()} takes a \code{data.frame} with columns \code{name} and \code{type} and creates an empty \code{data.frame} with the corresponding columns and R types
\item \code{as_fields()} returns a \code{data.frame} with columns \code{name}, \code{type}, \code{alias}, \code{nullable}, and \code{editable} columns
\itemize{
\item This resembles that of the \code{fields} returned by a FeatureService
}
}
}
\description{
Infers Esri field types from R objects. Use \code{\link[=as_fields]{as_fields()}} to create a
data.frame of valid \href{https://developers.arcgis.com/web-map-specification/objects/field/}{Esri Field Types}
from an \code{sf} object or \code{data.frame}.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
\itemize{
\item \code{as_fields()} takes a data frame-like object and infers the Esri field type from it.
\item \code{fields_as_pytpe_df()} takes a list with \code{type} and \code{name} and creates an empty \code{data.frame} with the corresponding column names and types.
\item \code{get_ptype()} takes a scalar character containing the Esri field type and returns a prototype of the pertinent R type
}
\subsection{Field type mapping:}{

Esri field types are mapped as
\itemize{
\item \code{esriFieldTypeSmallInteger}: integer
\item \code{esriFieldTypeSingle}: double
\item \code{esriFieldTypeGUID}: integer
\item \code{esriFieldTypeOID}: integer
\item \code{esriFieldTypeInteger}: integer
\item \code{esriFieldTypeBigInteger}: double
\item \code{esriFieldTypeDouble}: double
\item \code{esriFieldTypeString}: character
\item \code{esriFieldTypeDate}: date
}

R types are mapped as
\itemize{
\item \code{double}: esriFieldTypeDouble
\item \code{integer}: esriFieldTypeInteger
\item \code{character}: esriFieldTypeString
\item \code{date}: esriFieldTypeDate
\item \code{raw}: esriFieldTypeBlob
}
}
}
\examples{
inferred <- as_fields(iris)
inferred

fields_as_ptype_df(inferred)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user-group-content.R
\name{content}
\alias{content}
\alias{arc_group_content}
\alias{arc_user_content}
\title{Portal Content Items}
\usage{
arc_group_content(group, host = arc_host(), token = arc_token())

arc_user_content(
  user = arc_user_self(token = token),
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{group}{a scalar character of the group ID or a \code{PortalGroup} object created using \code{\link[=arc_group]{arc_group()}}}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}

\item{user}{a scalar character of the username or a \code{PortalUser} object created using \code{\link[=arc_user]{arc_user()}}}
}
\value{
a \code{data.frame} of content item metadata
}
\description{
For a given user or group, returns a data.frame of all content items owned by them.
}
\examples{
\dontrun{
library(arcgis)

# authenticate
set_arc_token(auth_user())

# get your own content items
self <- arc_user_self()
arc_user_content(self$username)

# get a specific group's items
arc_group_content("2f0ec8cb03574128bd673cefab106f39")
}
}
\references{
\itemize{
\item \href{https://developers.arcgis.com/rest/users-groups-and-items/group-content/}{Group Content API Reference}
\item \href{https://developers.arcgis.com/rest/users-groups-and-items/user-content/}{User Content API Reference}
}
}
\keyword{content}
\keyword{portal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auth-shiny.R
\name{auth_shiny}
\alias{auth_shiny}
\alias{oauth_provider_arcgis}
\title{Authenticate with Shiny}
\usage{
oauth_provider_arcgis(host = arc_host())

auth_shiny(
  client = Sys.getenv("ARCGIS_CLIENT"),
  secret = Sys.getenv("ARCGIS_SECRET"),
  redirect_uri = Sys.getenv("ARCGIS_REDIRECT_URI"),
  host = arc_host(),
  ...
)
}
\arguments{
\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{client}{an OAuth 2.0 developer application client ID. By default uses the
environment variable \code{ARCGIS_CLIENT}.}

\item{secret}{an OAuth 2.0 developer application secret. By default uses the environment
variable \code{ARCGIS_SECRET}.}

\item{redirect_uri}{default \code{Sys.getenv("ARCGIS_REDIRECT_URI")}. The redirect URL after completing authentication flow.}

\item{...}{additional arguments passed to \code{\link[shinyOAuth:oauth_client]{shinyOAuth::oauth_client()}}}
}
\description{
\code{oath_provider_arcgis()} and \code{auth_shiny()} provide an integrated
authentication experience for \code{{shiny}} applications via the package
\href{https://github.com/lukakoning/shinyOAuth/}{shinyOAuth}. Applications that
use \code{auth_shiny()} will enable user-specific behavior and authentications.
}
\details{
Authentication with \code{auth_shiny()} requires the package \code{shinyOAuth}.
When creating an OAuth app in ArcGIS Online / Enterprise, ensure that
the app has a valid redirect URI. This \strong{must be} the same redirect URI
that is used by \code{shinyOAuth} and \strong{must be} the same port as the
shiny application is served on.

\code{oauth_provider_arcgis()} is a low-level provider function that mimics
other \verb{oauth_provider_*()} functions from \code{shinyOAuth}. Use \code{auth_shiny()}
to create an OAuth client.

\strong{Important:} The client object returned by \code{auth_shiny()} must be created
outside of the server function to ensure a shared state across the OAuth flow.
Creating the client inside the server function will result in authentication failures.

The user info returned from \code{auth$token@userinfo} has the same structure
as \code{\link[=arc_portal_self]{arc_portal_self()}}. See that function's documentation for available fields.

The below example is derived from the shinyOAuth documentation.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(shiny)
library(shinyOAuth)
library(arcgisutils)

# Simple UI
ui <- fluidPage(
  use_shinyOAuth(),
  uiOutput("login_information")
)

client <- auth_shiny()

server <- function(input, output, session) \{
  # Set auto_redirect = FALSE for manual login
  auth <- shinyOAuth::oauth_module_server(
    "AGOL",
    client,
    auto_redirect = FALSE
  )

  output$login_information <- renderUI(\{
    if (auth$authenticated) \{
      user_info <- auth$token@userinfo
      tagList(
        tags$p("You are logged in! Your details:"),
        tags$pre(paste(capture.output(str(user_info)), collapse = "\\n"))
      )
    \} else \{
      tagList(
        tags$p("You are not logged in."),
        actionButton("login_btn", "Login with ArcGIS")
      )
    \}
  \})

  # Trigger login when button is clicked
  observeEvent(input$login_btn, \{
    auth$request_login()
  \})
\}

runApp(
  shinyApp(ui, server),
  port = 8100,
  launch.browser = FALSE
)

# Open the app in your regular browser at http://localhost:8100
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-requests.R
\name{detect_errors}
\alias{detect_errors}
\alias{catch_error}
\title{Detect errors in parsed json response}
\usage{
detect_errors(response, error_call = rlang::caller_env())

catch_error(response, error_call = rlang::caller_env())
}
\arguments{
\item{response}{for \code{detect_errors()}, a list typically from \code{RcppSimdJson::fparse(httr2::resp_body_string(resp))}. For \code{catch_error()}, the string from \code{httr2::resp_body_string(resp)}.}

\item{error_call}{default \code{\link[rlang:stack]{rlang::caller_env()}}. The environment from which
to throw the error from.}
}
\value{
Nothing. Used for it's side effect. If an error code is encountered in the
response an error is thrown with the error code and the error message.
}
\description{
The requests responses from ArcGIS don't return the status code
in the response itself but rather from the body in the json.
This function checks for the existence of an error. If an error is found,
the contents of the error message are bubbled up.
}
\examples{
\dontrun{
response <- list(
  error = list(
    code = 400L,
    message = "Unable to generate token.",
    details = "Invalid username or password."
  )
)

detect_errors(response)
}
}
\concept{requests}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-geometry.R
\name{determine_esri_geo_type}
\alias{determine_esri_geo_type}
\title{Determine Esri Geometry type}
\usage{
determine_esri_geo_type(x, call = rlang::caller_env())
}
\arguments{
\item{x}{an object of class \code{data.frame}, \code{sf}, \code{sfc}, or \code{sfg}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
returns a character scalar of the corresponding Esri geometry type
}
\description{
Takes an \code{sf} or \code{sfc} object and returns the appropriate Esri geometry type.
}
\details{
\subsection{Geometry type mapping}{
\itemize{
\item \code{POINT}:  \code{esriGeometryPoint}
\item \code{MULTIPOINT}:  \code{esriGeometryMultipoint}
\item \code{LINESTRING}:  \code{esriGeometryPolyline}
\item \code{MULTILINESTRING}:  \code{esriGeometryPolyline}
\item \code{POLYGON}:  \code{esriGeometryPolygon}
\item \code{MULTIPOLYGON}:  \code{esriGeometryPolygon}
}
}
}
\examples{
determine_esri_geo_type(sf::st_point(c(0, 0)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sharing.R
\name{arc_item}
\alias{arc_item}
\title{Portal Item Metadata}
\usage{
arc_item(item_id, host = arc_host(), token = arc_token())
}
\arguments{
\item{item_id}{the ID of the item to fetch. A scalar character.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}
}
\value{
an object of class \code{PortalItem} which is a list with the item's metadata.
}
\description{
Given the unique ID of a content item, fetches the item's metadata from a portal.
}
\details{
See \href{https://developers.arcgis.com/rest/users-groups-and-items/item/}{API Reference} for more information.

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
arc_item("9df5e769bfe8412b8de36a2e618c7672")
\dontshow{\}) # examplesIf}
}
\seealso{
Other portal item: 
\code{\link{arc_item_data}()}
}
\concept{portal item}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user-self.R
\name{arc_user_self}
\alias{arc_user_self}
\title{Discover Authenticated User Metadata}
\usage{
arc_user_self(
  host = arc_host(),
  token = arc_token(),
  error_call = rlang::caller_call()
)
}
\arguments{
\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}

\item{error_call}{the caller environment to be used when propagating errors.}
}
\value{
a list of the authenticated user's metadata
}
\description{
Given an authentication token, return a list of user-specfic information such as the user ID, username, available credits, email, groups, last login date and more.
}
\examples{
\dontrun{
if (interactive()) {
  arc_user_self(token = auth_user())
}
}
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/self/}{API Reference}
}
\keyword{portal}
\keyword{self}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geoprocessing-class.R
\name{arc_job_status}
\alias{arc_job_status}
\title{Geoprocessing Job Status}
\usage{
arc_job_status(status = character(0))
}
\arguments{
\item{status}{a scalar character. Must be one of \code{"esriJobSubmitted"},
\code{"esriJobWaiting"}, \code{"esriJobExecuting"}, \code{"esriJobSucceeded"}, \code{"esriJobFailed"},
\code{"esriJobTimedOut"}, \code{"esriJobCancelling"}, or \verb{"esriJobCancelled}".}
}
\value{
an object of class \code{arc_job_status}
}
\description{
Represents the status of a geoprocessing job.
}
\examples{
arc_job_status("esriJobSubmitted")
}
\seealso{
Other geoprocessing: 
\code{\link{arc_form_params}()},
\code{\link{arc_gp_job}},
\code{\link{gp_job_from_url}()},
\code{\link{gp_params}}
}
\concept{geoprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geoprocessing-class.R
\name{gp_job_from_url}
\alias{gp_job_from_url}
\title{Create GP Job from existing URL}
\usage{
gp_job_from_url(url, token = arc_token())
}
\arguments{
\item{url}{the url of an existing geoprocessing job}

\item{token}{default \code{\link[=arc_token]{arc_token()}}. The token to be used with the job.}
}
\description{
Create GP Job from existing URL
}
\examples{
if (interactive()) {
job_url <- paste0(
  "https://hydro.arcgis.com/arcgis/rest/services/Tools/Hydrology/",
  "GPServer/TraceDownstream/jobs/jfde67910074649e4a567f0adbb8af870"
)

gp_job_from_url(
  job_url,
  token = auth_user()
)
}
}
\seealso{
Other geoprocessing: 
\code{\link{arc_form_params}()},
\code{\link{arc_gp_job}},
\code{\link{arc_job_status}()},
\code{\link{gp_params}}
}
\concept{geoprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arc-auth.R
\name{auth_code}
\alias{auth_code}
\alias{auth_client}
\alias{auth_binding}
\alias{auth_user}
\alias{auth_key}
\alias{refresh_token}
\alias{validate_or_refresh_token}
\title{Authorization}
\usage{
auth_code(client = Sys.getenv("ARCGIS_CLIENT"), host = arc_host())

auth_client(
  client = Sys.getenv("ARCGIS_CLIENT"),
  secret = Sys.getenv("ARCGIS_SECRET"),
  host = arc_host(),
  expiration = 120
)

auth_binding()

auth_user(
  username = Sys.getenv("ARCGIS_USER"),
  password = Sys.getenv("ARCGIS_PASSWORD"),
  host = arc_host(),
  expiration = 60
)

auth_key(api_key = Sys.getenv("ARCGIS_API_KEY"), host = arc_host())

refresh_token(token, client = Sys.getenv("ARCGIS_CLIENT"), host = arc_host())

validate_or_refresh_token(
  token,
  client = Sys.getenv("ARCGIS_CLIENT"),
  host = arc_host(),
  refresh_threshold = 10,
  call = rlang::caller_env()
)
}
\arguments{
\item{client}{an OAuth 2.0 developer application client ID. By default uses the
environment variable \code{ARCGIS_CLIENT}.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{secret}{an OAuth 2.0 developer application secret. By default uses the environment
variable \code{ARCGIS_SECRET}.}

\item{expiration}{the duration of the token in minutes.}

\item{username}{default \code{Sys.getenv("ARCGIS_USER")}.
Your username to login. \strong{Do not} hard code this value.}

\item{password}{default \code{Sys.getenv("ARCGIS_PASSWORD")}.
Your password to login. \strong{Do not} hard code this value.}

\item{api_key}{default \code{Sys.getenv("ARCGIS_API_KEY")}. A character scalar of an ArcGIS Developer API key.}

\item{token}{an \code{httr2_token} as created by \code{auth_code()} or similar}

\item{refresh_threshold}{default \code{10}. If token expiry is within this
threshold (in seconds) the token will be refreshed only if a
\code{refresh_token} is available. Token refreshing is only possible with
\code{auth_code()} flow.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
an \code{httr2_token}
}
\description{
Authorize your R session to connect to an ArcGIS Portal. See details.
}
\details{
ArcGIS Online and Enterprise Portals utilize OAuth2 authorization via their REST APIs.
\itemize{
\item \code{auth_code()} is the recommend OAuth2 workflow for interactive sessions
\item \code{auth_client()} is the recommended OAuth2 workflow for non-interactive sessions
\item \code{auth_user()} uses legacy username and password authorization using the \code{generateToken} endpoint. It is only recommended for legacy systems that do not implement OAuth2.
\item \code{auth_binding()} fetches a token from the active portal set by \code{arcgisbinding}. Uses \code{arcgisbinding::arc.check_portal()} to extract the authorization token. Recommended if using arcgisbinding.
}
}
\examples{
\dontrun{
auth_code()
auth_client()
auth_user()
auth_key()
auth_binding()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-types.R
\name{portal_types}
\alias{portal_types}
\alias{item_type}
\alias{item_keyword}
\alias{portal_item_keywords}
\alias{portal_item_types}
\title{Portal Item Types}
\usage{
item_type(item_type = character(0))

item_keyword(keyword = character(0))

portal_item_keywords()

portal_item_types()
}
\arguments{
\item{item_type}{a scalar character of the item type. See \code{\link[=portal_item_types]{portal_item_types()}} for valid item types.}

\item{keyword}{a scalar character of the item type keyword. See \code{\link[=portal_item_keywords]{portal_item_keywords()}}.}
}
\description{
Every portal item has an associated item type. Each of those item types have keywords which cna be used to help narrow down search further.
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/items-and-item-types}{REST API Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geoprocessing-types.R, R/utils-envelope.R
\name{gp_params}
\alias{gp_params}
\alias{parse_gp_feature_record_set}
\alias{as_gp_feature_record_set}
\alias{parse_gp_record_set}
\alias{as_record_set}
\alias{as_gp_raster_layer}
\alias{gp_linear_unit}
\alias{as_gp_linear_unit}
\alias{parse_gp_linear_unit}
\alias{gp_areal_unit}
\alias{as_gp_areal_unit}
\alias{parse_gp_areal_unit}
\alias{as_gp_date}
\alias{parse_gp_date}
\alias{as_spatial_reference}
\alias{from_spatial_reference}
\alias{parse_spatial_reference}
\alias{from_envelope}
\title{Geoprocessing Parameter Types}
\usage{
parse_gp_feature_record_set(json)

as_gp_feature_record_set(x)

parse_gp_record_set(json)

as_record_set(x)

as_gp_raster_layer(x)

gp_linear_unit(distance = integer(0), units = character(0))

as_gp_linear_unit(x)

parse_gp_linear_unit(json)

gp_areal_unit(area = integer(0), units = character(0))

as_gp_areal_unit(x)

parse_gp_areal_unit(json)

as_gp_date(x)

parse_gp_date(json)

as_spatial_reference(x)

from_spatial_reference(sr, error_call = rlang::caller_call())

parse_spatial_reference(json)

from_envelope(x, error_call = rlang::caller_call())
}
\arguments{
\item{json}{raw json to parse}

\item{x}{the object to convert into json}

\item{distance}{a scalar number of the distance.}

\item{units}{the unit of the measurement. Must be one of "esriUnknownAreaUnits", "esriSquareInches", "esriSquareFeet", "esriSquareYards", "esriAcres", "esriSquareMiles", "esriSquareMillimeters", "esriSquareCentimeters", "esriSquareDecimeters", "esriSquareMeters", "esriAres", "esriHectares", "esriSquareKilometers", "esriSquareInchesUS", "esriSquareFeetUS", "esriSquareYardsUS", "esriAcresUS", "esriSquareMilesUS".}

\item{area}{a scalar number of the measurement.}

\item{sr}{a list with fields \code{latestWkid}, \code{wkid}, or \code{wkt} representing a spatial reference}

\item{error_call}{the caller environment to be used when propagating errors.}
}
\description{
Functions for converting R objects to and from ArcGIS geoprocessing parameter
types. These functions handle the serialization and parsing of various data
types used in ArcGIS geoprocessing services.
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This package provides support for the following geoprocessing parameter types:
\subsection{Implemented Types}{
\itemize{
\item \strong{GPFeatureRecordSetLayer}: Feature collections with geometry and attributes
\item \strong{GPRecordSet}: Tabular data without geometry
\item \strong{GPRasterDataLayer}: Raster datasets from Portal items, Image Servers, or URLs
\item \strong{GPLinearUnit}: Linear distance measurements with units
\item \strong{GPArealUnit}: Area measurements with units
\item \strong{GPDate}: Date/time values in milliseconds since epoch
\item \strong{GPSpatialReference}: Coordinate reference systems
}
}

\subsection{Not Yet Implemented}{

The following types are planned for future implementation:
\itemize{
\item \strong{GPField}: Field definitions with name, type, and properties
\item \strong{GPMultiValue}: Arrays of values for a single data type
\item \strong{GPValueTable}: Flexible table-like objects with rows and columns
\item \strong{GPComposite}: Parameters that accept multiple data types
\item \strong{GPEnvelope}: Bounding box extents (use \code{as_extent()} for GPExtent)
}
}
}
\section{Usage Patterns}{


Most functions follow a consistent pattern:
\itemize{
\item \verb{as_gp_*()}: Convert R objects to geoprocessing parameter JSON
\item \verb{parse_gp_*()}: Parse geoprocessing response JSON to R objects
\item Constructor functions (e.g., \code{gp_linear_unit()}, \code{gp_areal_unit()}) create typed S7 objects
}
}

\section{Examples}{


\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Create a linear unit
distance <- gp_linear_unit(distance = 100, units = "esriMeters")

# Convert spatial data to feature record set
as_gp_feature_record_set(my_sf_data)

# Parse a geoprocessing response
parse_gp_feature_record_set(response_json)
}\if{html}{\out{</div>}}
}

\examples{
# create a feature record set
fset <- as_gp_feature_record_set(iris[1,])
fset

# create fake gp feature record set to parse
fset_list <- list(
  list(
    dataType = "GPFeatureRecordSetLayer",
    paramName = "example",
    value = as_featureset(iris[1,])
  )
)

# create the json
json <- yyjsonr::write_json_str(fset_list, auto_unbox = TRUE)

# parse the record set json
parse_gp_feature_record_set(json)

# linear units
lu <- gp_linear_unit(10, "esriMeters")
lu
as_gp_linear_unit(lu)

# areal units
au <- gp_areal_unit(10, "esriSquareMeters")
au

as_gp_areal_unit(au)

# dates
json <- r"({
  "paramName": "Output_Date",
  "dataType": "GPDate",
  "value": 1199145600000
})"

parse_gp_date(json)
sr <- list(wkid = 4326L)
from_spatial_reference(sr)
x <- list(
  xmin = -122.4195,
  ymin = 37.330219000000056,
  xmax = -122.030757,
  ymax = 37.77650360000007,
  spatialReference = list(wkid = 4326L, latestWkid = 4326L)
)

from_envelope(x)
}
\references{
\href{https://developers.arcgis.com/rest/services-reference/enterprise/gp-data-types/}{API Documentation}
}
\seealso{
Other geoprocessing: 
\code{\link{arc_form_params}()},
\code{\link{arc_gp_job}},
\code{\link{arc_job_status}()},
\code{\link{gp_job_from_url}()}
}
\concept{geoprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arc-base-req.R
\name{arc_base_req}
\alias{arc_base_req}
\title{Generate base request}
\usage{
arc_base_req(
  url,
  token = NULL,
  path = NULL,
  query = NULL,
  error_call = rlang::caller_env()
)
}
\arguments{
\item{url}{a valid url that is passed to \code{\link[httr2:request]{httr2::request()}}}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}

\item{path}{a character vector of paths to be appended to url using \code{\link[httr2:req_url]{httr2::req_url_path_append()}}}

\item{query}{a named vector or named list of query parameters to be appended to the url using \code{\link[httr2:req_url]{httr2::req_url_query()}}}

\item{error_call}{the caller environment to be used when propagating errors.}
}
\value{
an \code{httr2_request} with the \code{X-Esri-Authorization} header and \code{User-Agent} set.
}
\description{
This function takes a url and creates a basic httr2 request that
adds the user-agent and adds an authorization token to the
\code{X-Esri-Authorization} header.
}
\examples{
arc_base_req("https://arcgis.com")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arc-token.R
\name{arc_token}
\alias{arc_token}
\alias{set_arc_token}
\alias{unset_arc_token}
\alias{obj_check_token}
\alias{check_token_has_user}
\title{Manage authorization tokens}
\usage{
arc_token(token = "ARCGIS_TOKEN")

set_arc_token(token, ...)

unset_arc_token(token = NULL)

obj_check_token(token, call = rlang::caller_env())

check_token_has_user(token, call = rlang::caller_env())
}
\arguments{
\item{token}{for \code{arc_token()}, the name of a token to fetch. For \code{set_arc_token()},
it is an \code{httr2_token} that will be set. For \code{unset_arc_token()}, a character
vector of token names to be unset.}

\item{...}{named arguments to set \code{httr2_token}. Must be valid names and must be an \code{httr2_token}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\description{
These functions are used to set, fetch, and check authorization tokens.
}
\details{
It is possible to have multiple authorization tokens in one session. These
functions assist you in managing them.

\code{arc_token()} is used to fetch tokens by name. The default token is \code{ARCGIS_TOKEN}.
However, they can be any valid character scalar. \code{set_arc_token()} will create
store a token with the name \code{ARCGIS_TOKEN}. However, you can alternatively
set the tokens by name using a key-value pair. The key is what you would pass
to \code{arc_token()} to fetch the \code{httr2_token} object. To remove a token that has
been set, use \code{unset_arc_token()}.

\code{obj_check_token()} is a developer oriented function that can be used to check
if an object is indeed an \code{httr2_token}. To check if a token has expired,
\code{\link[=validate_or_refresh_token]{validate_or_refresh_token()}} will do so.

\code{check_token_has_user()} is a developer oriented function that checks to see
if a token has a \code{username} field associated with it.

For developers:

\code{set_arc_token()} uses a package level environment to store the tokens. The
tokens are fetched from the environment using \code{arc_token()}.
}
\examples{
# create fake tokens
token_a <- httr2::oauth_token("1234", arcgis_host = arc_host())
token_b <- httr2::oauth_token("abcd", arcgis_host = arc_host())

# set token to the default location
set_arc_token(token_a)

# fetch token from the default location
arc_token()

# set token by name
set_arc_token(org_a = token_a, org_b = token_b)

# fetch token by name
arc_token("org_a")
arc_token("org_b")

# unset tokens
unset_arc_token()
unset_arc_token(c("org_a", "org_b"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-dimensions.R
\name{determine_dims}
\alias{determine_dims}
\alias{has_m}
\alias{has_z}
\title{Determine the dimensions of a geometry object}
\usage{
determine_dims(x)

has_m(x)

has_z(x)
}
\arguments{
\item{x}{an object of class \code{sfc} or \code{sfg}}
}
\value{
\code{determine_dims()} returns a scalar character of the value \code{"xy"}, \code{"xyz"}, or \code{"xyzm"} depending
on what dimensions are represented.

\code{has_m()} and \code{has_z()} returns a logical scalar of \code{TRUE} or \code{FALSE} if the
geometry has a Z or M dimension.
}
\description{
Given an sfc or sfg object determine what dimensions are represented.
}
\examples{

geo <- sf::st_read(system.file("shape/nc.shp", package="sf"), quiet = TRUE)[["geometry"]]

determine_dims(geo)
has_z(geo)
has_m(geo)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util-parse-esri-json.R
\name{parse_esri_json}
\alias{parse_esri_json}
\title{Parse Esri JSON}
\usage{
parse_esri_json(string, ..., call = rlang::caller_env())
}
\arguments{
\item{string}{the raw Esri JSON string.}

\item{...}{additional arguments passed to \code{\link[RcppSimdJson:fparse]{RcppSimdJson::fparse}}}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
A data.frame. If geometry is found, returns an sf object.
}
\description{
Parses an Esri FeatureSet JSON object into an R object. If there is no
geometry present, a data.frame is returned. If there is geometry, an sf object is returned.
}
\examples{

esri_json <- '{
    "geometryType": "esriGeometryPolygon",
    "spatialReference": {
        "wkid": 4326
    },
    "hasZ": false,
    "hasM": false,
    "features": [
        {
            "attributes": {
                "id": 1
            },
            "geometry": {
                "rings": [
                    [
                        [0.0, 0.0],
                        [1.0, 0.0],
                        [1.0, 1.0],
                        [0.0, 1.0],
                        [0.0, 0.0]
                    ]
                ]
            }
        }
    ]
}'

parse_esri_json(esri_json)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-requests.R
\name{arc_agent}
\alias{arc_agent}
\title{Set user-agent for arcgisutils}
\usage{
arc_agent(req)
}
\arguments{
\item{req}{an httr2 request}
}
\value{
an httr2 request object
}
\description{
Override the default user-agent set by httr2 to indicate that a request
came from arcgisutils.
}
\examples{
req <- httr2::request("http://example.com")
arc_agent(req)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind-results.R
\name{rbind_results}
\alias{rbind_results}
\title{Combine multiple data.frames}
\usage{
rbind_results(x, call = rlang::current_env(), .ptype = data.frame())
}
\arguments{
\item{x}{a list where each element is a \code{data.frame} or \code{NULL}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}

\item{.ptype}{currently unused. Reserved for a future release.}
}
\value{
see details.
}
\description{
A general function that takes a list of \code{data.frame}s and returns a single
and combines them into a single object. It will use the fastest method
available. In order this is \code{\link[collapse:rowbind]{collapse::rowbind()}}, \code{\link[data.table:rbindlist]{data.table::rbindlist()}},
\code{\link[vctrs:list_unchop]{vctrs::list_unchop()}}, then \code{do.call(rbind.data.frame, x)}.
}
\details{
If all items in the list are \code{data.frame}s, then the result will be a \code{data.frame}.
If all elements are an \code{sf} object, then the result will be an \code{sf} object.
If the items are mixed, the result will be a \code{data.frame}.

If any items are \code{NULL}, then an attribute \code{null_elements} will be attached
to the result. The attribute is an integer vector of the indices that
were \code{NULL}.
}
\examples{

x <- head(iris)
res <- rbind_results(list(x, NULL, x))
attr(res, "null_elements")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arc-paginate-req.R
\name{arc_paginate_req}
\alias{arc_paginate_req}
\title{Paginate ArcGIS Requests}
\usage{
arc_paginate_req(req, page_size = 10, max_pages = Inf, .progress = TRUE)
}
\arguments{
\item{req}{an \code{httr2_request} ideally created with \code{arc_base_req}}

\item{page_size}{a scalar integer between 1 and 100 indicating the number of responses per page.}

\item{max_pages}{the maximum number of pages to fetch. By default fetches all pages.}

\item{.progress}{default \code{TRUE}. Whether to display a progress bar for requests.}
}
\value{
a list of \code{httr2_response}.
}
\description{
Many API endpoints provide common \href{https://developers.arcgis.com/rest/users-groups-and-items/common-parameters/#paging-properties}{pagination properties}. \code{arc_paginate_request()} automatically applies pagination to an input request.
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/common-parameters/#paging-properties}{API Documentation}
}
\seealso{
\code{\link[=arc_base_req]{arc_base_req()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-servers.R
\name{arc_portal_servers}
\alias{arc_portal_servers}
\title{List ArcGIS Enterprise Servers}
\usage{
arc_portal_servers(
  id = arc_portal_self(token)[["id"]],
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{id}{the portal ID. By default it fetches the \code{id} from \code{\link[=arc_portal_self]{arc_portal_self()}}.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}
}
\value{
a data.frame of servers
}
\description{
The servers resource lists the ArcGIS Server sites that have been federated with the portal.
}
\examples{
\dontrun{
set_arc_token(auth_user())
arc_portal_servers()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-resources.R
\name{arc_portal_resources}
\alias{arc_portal_resources}
\title{Portal File Resources}
\usage{
arc_portal_resources(
  id = arc_portal_self(token)[["id"]],
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{id}{the portal ID. By default it fetches the \code{id} from \code{\link[=arc_portal_self]{arc_portal_self()}}.}

\item{page_size}{a scalar integer between 1 and 100 indicating the number of responses per page.}

\item{max_pages}{the maximum number of pages to fetch. By default fetches all pages.}

\item{.progress}{default \code{TRUE}. Whether to display a progress bar for requests.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}
}
\value{
a data.frame of resources available to your portal.
}
\description{
The resources endpoint lists all file resources for the organization.
}
\examples{
\dontrun{
  set_arc_token(auth_user())
  arc_portal_resources()
}

}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/resources-portal/}{API Reference}
}
\seealso{
Other portal: 
\code{\link{arc_portal_urls}()},
\code{\link{arc_portal_users}()},
\code{\link{self}}
}
\concept{portal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/esri-featureset-list.R,
%   R/esri-featureset-string.R
\name{as_featureset}
\alias{as_featureset}
\alias{as_esri_featureset}
\title{Create Esri FeatureSet Objects}
\usage{
as_featureset(x, crs = sf::st_crs(x), call = rlang::caller_env())

as_esri_featureset(x, crs = sf::st_crs(x), call = rlang::caller_env())
}
\arguments{
\item{x}{an object of class \code{sf}, \code{data.frame}, or \code{sfc}.}

\item{crs}{the coordinate reference system. It must be interpretable by \code{\link[sf:st_crs]{sf::st_crs()}}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
a list or a json string
}
\description{
These functions create an Esri FeatureSet object. A FeatureSet contains an inner array
of features as well as additional metadata about the the collection such as the
geometry type, spatial reference, and object ID field.
}
\examples{

library(sf)
# POINT
# create sfg points
xy <- st_sfc(st_point(c(1, 2)))
xyz <- st_sfc(st_point(c(1, 2, 3)))
xym <- st_sfc(st_point(c(1, 2, 3), dim = "XYM"))

as_esri_featureset(xy)
as_esri_featureset(xyz)
as_esri_featureset(xym)

# MULTIPOINT
# vector to create matrix points
set.seed(0)
x <- rnorm(12)

xy <- st_sfc(st_multipoint(matrix(x, ncol = 2)))
xyz <- st_sfc(st_multipoint(matrix(x, ncol = 3)))
xym <- st_sfc(st_multipoint(matrix(x, ncol = 3), dim = "XYM"))

as_esri_featureset(xy)
as_esri_featureset(xyz)
as_esri_featureset(xym)

# LINESTRING
xy <- st_sfc(st_linestring(matrix(x, ncol = 2)))
xyz <- st_sfc(st_linestring(matrix(x, ncol = 3)))
xym <- st_sfc(st_linestring(matrix(x, ncol = 3), dim = "XYM"))

as_esri_featureset(xy)
as_esri_featureset(xyz)
as_esri_featureset(xym)

# MULTILINESTRING
as_esri_featureset(st_sfc(st_multilinestring(list(xy[[1]], xy[[1]]))))
as_esri_featureset(st_sfc(st_multilinestring(list(xyz[[1]], xyz[[1]]))))
as_esri_featureset(st_sfc(st_multilinestring(list(xym[[1]], xym[[1]]))))

# POLYGON
coords <- rbind(
  c(0, 0, 0, 1),
  c(0, 1, 0, 1),
  c(1, 1, 1, 1),
  c(1, 0, 1, 1),
  c(0, 0, 0, 1)
)

xy <- st_sfc(st_polygon(list(coords[, 1:2])))
xyz <- st_sfc(st_polygon(list(coords[, 1:3])))
xym <- st_sfc(st_polygon(list(coords[, 1:3]), dim = "XYM"))

as_esri_featureset(xy)
as_esri_featureset(xyz)
as_esri_featureset(xym)

# MULTIPOLYGON
as_esri_featureset(st_sfc(st_multipolygon(list(xy[[1]], xy[[1]]))))
as_esri_featureset(st_sfc(st_multipolygon(list(xyz[[1]], xyz[[1]]))))
as_esri_featureset(st_sfc(st_multipolygon(list(xym[[1]], xym[[1]]))))
}
\references{
\href{https://developers.arcgis.com/documentation/common-data-types/featureset-object.htm}{API Reference}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portal-group-user-list.R
\name{arc_group_users}
\alias{arc_group_users}
\title{List users in a group}
\usage{
arc_group_users(
  group,
  name = NULL,
  member_type = NULL,
  joined_before = NULL,
  joined_after = NULL,
  sort_field = c("username", "membertype", "joined"),
  sort_order = c("asc", "desc"),
  page_size = 50,
  max_pages = Inf,
  .progress = TRUE,
  host = arc_host(),
  token = arc_token()
)
}
\arguments{
\item{group}{a group ID as a scalar string or \code{PortalGroup} created via \code{\link[=arc_group]{arc_group()}}}

\item{name}{a scalar string of a user to search for.}

\item{member_type}{default \code{NULL}. If provided must be one of \code{"admin"} or \code{"member"}.}

\item{joined_before}{default \code{NULL}. A scalar date to search for users who joined before this date.}

\item{joined_after}{default \code{NULL}. A scalar date to search for users who joined after this date.}

\item{sort_field}{default \code{"username"}. The field to sort by. Must be one of \code{"username"}, \code{"membertype"}, or \code{"joined"}.}

\item{sort_order}{optional string. One of either \code{asc} or \code{desc} for ascending or descending order respectively.}

\item{page_size}{a scalar integer between 1 and 100 indicating the number of responses per page.}

\item{max_pages}{the maximum number of pages to fetch. By default fetches all pages.}

\item{.progress}{default \code{TRUE}. Whether to display a progress bar for requests.}

\item{host}{default \code{"https://www.arcgis.com"}. The host of your ArcGIS Portal.}

\item{token}{an object of class \code{httr2_token} as generated by \code{\link[=auth_code]{auth_code()}}
or related function}
}
\value{
a data.frame
}
\description{
List all users in a given group.
}
\examples{
\dontrun{
set_arc_token(auth_user())
groups <- arc_user("r-bridge-docs")$groups
arc_group_users(groups$id[1])
}
}
\references{
\href{https://developers.arcgis.com/rest/users-groups-and-items/group-users-list/}{API Reference}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-dates.R
\name{is_date}
\alias{is_date}
\alias{date_to_ms}
\alias{from_esri_date}
\title{Date handling}
\usage{
is_date(x, tz)

date_to_ms(x, tz = "UTC")

from_esri_date(x)
}
\arguments{
\item{x}{an object of class \code{Date} or \code{POSIXt}. In the case of \code{is_date()}, any R object.}

\item{tz}{a character string.  The time zone specification to be used
    for the conversion, \emph{if one is required}.  System-specific (see
    \link[base]{time zones}), but \code{""} is the current time zone, and
    \code{"GMT"} is UTC (Universal Time, Coordinated).  Invalid values
    are most commonly treated as UTC, on some platforms with a warning.}
}
\value{
\itemize{
\item \code{is_date()} returns a logical scalar
\item \code{date_to_ms()} returns a numeric vector of times in milliseconds from the Unix Epoch in the specified time zone.
}
}
\description{
Esri date fields are represented as milliseconds from the Unix Epoch.
}
\details{
\itemize{
\item \code{is_date()}: checks if an object is a \code{Date} or \code{POSIXt} class object.
\item \code{date_to_ms()} converts a date object to milliseconds from the Unix Epoch in the specified time zone.
}
}
\examples{

today <- Sys.Date()

is_date(today)

date_to_ms(today)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-requests.R
\name{fetch_layer_metadata}
\alias{fetch_layer_metadata}
\title{Retrieve metadata}
\usage{
fetch_layer_metadata(url, token = NULL, call = rlang::caller_env())
}
\arguments{
\item{url}{the url of the item.}

\item{token}{an \code{httr2_token} from one of the provided \code{auth_} functions}

\item{call}{default \code{\link[rlang:stack]{rlang::caller_env()}}. The calling environment passed to \code{detect_errors()}.}
}
\value{
returns a list object
}
\description{
Utility functions for feature service metadata.
}
\details{
\itemize{
\item \code{fetch_layer_metadata()} given a request, fetches the metadata by setting
the query parameter \code{f=json}
}
}
\examples{
# url is broken into parts to fit within 100 characters to avoid CRAN notes
url_parts <- c(
  "https://services.arcgis.com/P3ePLMYs2RVChkJx/ArcGIS/rest/services",
  "/USA_Counties_Generalized_Boundaries/FeatureServer/0"
)

furl <- paste0(url_parts, collapse = "")
meta <- fetch_layer_metadata(furl)
head(names(meta))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compact}
\alias{compact}
\alias{\%||\%}
\alias{check_dots_named}
\alias{data_frame}
\title{General utility functions}
\usage{
compact(.x)

a \%||\% b

check_dots_named(dots, call = rlang::caller_env())

data_frame(x, call = rlang::caller_call())
}
\arguments{
\item{.x}{a list}

\item{a}{an R object}

\item{b}{an R object}

\item{dots}{a list collected from dots via \code{rlang::list2(...)}}

\item{call}{default \code{\link[rlang:stack]{rlang::caller_call()}}.}

\item{x}{a data.frame}
}
\value{
\itemize{
\item \code{compact()} a list
\item \verb{\%||\%} the first non-null item or \code{NULL} if both are \code{NULL}
}
}
\description{
General utility functions
}
\details{
\itemize{
\item \code{compact()} removes any \code{NULL} list elements
\item \verb{\%||\%} is a special pipe operator that returns \code{b} if \code{a} is \code{NULL}
}
}
\examples{

# remove null elements
compact(list(a = NULL, b = 1))

# if NULL return rhs
NULL \%||\% 123

# if not NULL return lhs
123 \%||\% NULL

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arc-auth.R
\name{arc_host}
\alias{arc_host}
\title{Determines Portal Host}
\usage{
arc_host()
}
\value{
A scalar character, \code{"https://www.arcgis.com"} by default.
}
\description{
Returns a scalar character indicating the host to make requests to.
}
\details{
By default, the host is ArcGIS Online <\verb{https://www.arcgis.com}>. If the
environment variable \code{ARCGIS_HOST} is set, it will be returned.
}
\examples{
arc_host()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-geometry.R
\name{as_extent}
\alias{as_extent}
\title{Convert an object to an extent}
\usage{
as_extent(x, crs = sf::st_crs(x), call = rlang::caller_env())
}
\arguments{
\item{x}{an sf or sfc object}

\item{crs}{the CRS of the object. Must be parsable by \code{sf::st_crs()}}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}
}
\value{
An extent json object. Use \code{jsonify::to_json(x, unbox = TRUE)} to convert
to json.
}
\description{
Given an sf or sfc object create a list that represents the extent of the
object. The result of this function can be parsed directly into json using
\code{jsonify::to_json(x, unbox = TRUE)} or included into a list as the extent
component that will be eventually converted into json using the above function.
}
\examples{
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
as_extent(nc)
}
